

{-----------------------------------------------------------------------------
 Unit Name: BasicRiemUtils
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Routines for computations of a PseudoRiemanian Plane
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}



unit BasicRiemUtils;
interface
uses
{$IFDEF MSWINDOWS}
  Windows, Graphics,
{$ENDIF}
{$IFDEF LINUX}
  Types, QGraphics,
{$ENDIF}
  Globals;

function DrawRiemannianCurvature: Boolean;
function ComputePseudoRiemCurvature(p: Vector2; var K: double):
  Integer;
procedure DrawXPseudoRiemPlane;
function PickedPtPseudoRiem(pixel: TPoint; var param: Vector2):
  Boolean;
procedure RecordAndDrawSegmentPseudoRiemPlane(x1, x2: Vector3; toSave:
  Boolean);

function CheckRangePseudoRiem(lastX, nextX: Vector3; findProlong: Boolean;
  var newX, newVel: Vector3): Integer;
function DirectionsPseudoRiemPlane(var deriv: Vector3; prevDeriv, here:
  Vector3;
  axis: Integer): Boolean;
function GeodAccelPseudoRiemPlane(var accel: Vector3; vel, x: vector3;
  toRight: Integer):
  Boolean;
function CameBackPseudoRiemPlane(firstX, prevX, curX: Vector3; counter:
  Integer): Boolean;
procedure DrawCurvesPseudoRiemPlane(fromHere: vector2);
function ComputeChristSymbPseudoRiem(p: vector3; var metric, gammaX, gammaY:
  Vector3):
  BOOLEAN;
function ComputeFramePseudoRiemPlane(var u1, u2: Vector3; var phi, topT:
  Double; x: Vector3): double;

implementation
uses
  SysUtils, Math,
{$IFDEF MSWINDOWS}
  Forms, Controls,
{$ENDIF}
{$IFDEF LINUX}
  QForms, QControls,
{$ENDIF}
  OcFunct, Utilities, FSurface, FRunKut;


function PickedPtPseudoRiem(pixel: TPoint; var param: Vector2):
  Boolean;
var
  a, b: double;
begin
  FromDrawingToParameters(pixel, param);
  a := abs(param[1] - gCenter2D[1]) - gHalfEdges2D[1];
  b := abs(param[2] - gCenter2D[2]) - gHalfEdges2D[2];
  Result := (a <= 0) and (b <= 0);
end;


function ComputePseudoRiemCurvature(p: Vector2; var K: double):
  Integer;
{Computes scalar curvature at p. If (EG-F*F)>1.0e-6  }
{at p, result is 1. If (EG-F*F)<-1.0e-6 , result is -1. }
{Otherwise, result is 0.}
var
  Eyy, Fxy, Gxx, Ex, Ey, Fx, Fy, Gx, Gy, sqrDet, a0, a1, a3, a4, aux:
  double;
  E, F, G, det: double;
begin
  E := gTheF1.V(p);
  F := gTheF2.V(p);
  G := gTheF3.V(p);
  det := E * G - sqr(F);
  if det > 1.0E-6 then
    result := 1
  else if det < -1.0E-6 then
    result := -1
  else
    result := 0;
  sqrDet := sqr(det);
  if sqrDet > 1.0E-12 then begin
    a0 := gTheF1.V([p[1], p[2] - g2Incr2D[2]]);
    a1 := gTheF1.V([p[1], p[2] - gIncr2D[2]]);
    a3 := gTheF1.V([p[1], p[2] + gIncr2D[2]]);
    a4 := gTheF1.V([p[1], p[2] + g2Incr2D[2]]);
    Ey := (a0 - a4 + 8 * (a3 - a1)) / g24Incr2D[2];
    Eyy := (16 * (a1 + a3) - 30 * E - a0 - a4) / g24SqrIncr2D[2];
    Ex := (gTheF1.V([p[1] + gIncr2D[1], p[2]]) - gTheF1.V([p[1] - gIncr2D[1],
      p[2]])) /
      g4Incr2D[1];

    a0 := gTheF3.V([p[1] - g2Incr2D[1], p[2]]);
    a1 := gTheF3.V([p[1] - gIncr2D[1], p[2]]);
    a3 := gTheF3.V([p[1] + gIncr2D[1], p[2]]);
    a4 := gTheF3.V([p[1] + g2Incr2D[1], p[2]]);
    Gx := (a0 - a4 + 8 * (a3 - a1)) / g24Incr2D[1];
    Gxx := (16 * (a1 + a3) - 30 * G - a0 - a4) / g24SqrIncr2D[1];
    Gy := (gTheF3.V([p[1], p[2] + gIncr2D[2]]) - gTheF3.V([p[1], p[2] -
      gIncr2D[2]])) /
      g4Incr2D[2];

    a0 := gTheF2.V([p[1] - gIncr2D[1], p[2] - gIncr2D[2]]);
    a1 := gTheF2.V([p[1] - gIncr2D[1], p[2] + gIncr2D[2]]);
    a3 := gTheF2.V([p[1] + gIncr2D[1], p[2] + gIncr2D[2]]);
    a4 := gTheF2.V([p[1] + gIncr2D[1], p[2] - gIncr2D[2]]);
    Fx := (a4 - a0 + a3 - a1) / g4Incr2D[1];
    Fy := (a1 - a0 + a3 - a4) / g4Incr2D[2];
    Fxy := (a3 - a4 - a1 + a0) / g4Incr1Incr2;

    a1 := Fxy - Eyy - Gxx;
    a0 := Fy - Gx;
    a3 := Fx - Ey;

    aux := a1 * det + Ex * (F * Gy - a0 * G) + a3 * (a0 * f - E * Gy);
    aux := aux + sqr(Ey) * G + sqr(Gx) * E - 2 * Ey * Gx * F;
    K := aux / sqrDet;
    if result = -1 then K := -K;
    gKStr := FloatToStrF(K, ffFixed, 7, 4);
    if result = -1 then
      gKStr := kBullet +' ' + gKStr;
  end
end; {CalculaCurvatura}

function DrawRiemannianCurvature: Boolean;
var
  i, j: Integer;
  x: vector2;
  K: double;
  index: Byte;
  metricType: Integer;
  goodPt: Boolean;
  line: pTColorArray;
begin
  Result := False;
  if gMap <> nil then
    FillChar(gMap^, kWidth * kHeight, 0)
  else
    Exit;
  for j := 0 to kHeightMinus1 do begin
    if (j mod 30 = 0) then Application.ProcessMessages;
    if gStop then
      Exit;
    line := gStdBmp.ScanLine[j];
    x[2] := dp3 * j + dp4;
    if (abs(x[2] - gCenter2D[2]) > gHalfEdges2D[2]) then
      continue;
    for i := 0 to kWidthMinus1 do begin
      x[1] := dp1 * i + dp2;
      goodPt := (abs(x[1] - gCenter2D[1]) <=
        gHalfEdges2D[1]);
      if goodPt then
        metricType := ComputePseudoRiemCurvature(x, K);
      if goodPt and (metricType <> 0) then begin
        if metricType = kRiemMetric then
          index := trunc(1 + 50 * (1 - K / (1 + abs(K))))
        else
          index := trunc(101 + 50 * (1 - K / (1 + abs(K))));
        line^[i] := gPalette2D^[index];
        gMap^[i, j] := index;
      end;
    end;
  end;
  Surface.Drawing2D.Invalidate;
  Result := True;
end;


function XRiem(place: vector2; var p1, p2, p3, p4: TPoint): Boolean;
var
  i: integer;
  dir1, dir2, pos1, pos2, pos3, pos4: vector2;
  aux, eLow, fLow, gLow, l1, l2, p, q: double;
begin
  XRiem := FALSE;
  elow := gTheF1.V(place);
  flow := gTheF2.V(place);
  glow := gTheF3.V(place);

  aux := sqr(fLow) - eLow * gLow;
  if aux < 1.0E-10 then
    Exit;

  aux := sqrt(aux);
  l1 := sqr(-fLow + aux) + sqr(eLow);
  l2 := sqr(fLow + aux) + sqr(gLow);
  if l1 < l2 then begin
    l2 := 1 / sqrt(l2);
    dir1[1] := l2 * gLow;
    dir1[2] := -(fLow + aux) * l2;
  end
  else begin
    l1 := 1 / sqrt(l1);
    dir1[1] := (aux - fLow) * l1;
    dir1[2] := eLow * l1;
  end;

  aux := -aux;
  l1 := sqr(-fLow + aux) + sqr(eLow);
  l2 := sqr(fLow + aux) + sqr(gLow);
  if l1 < l2 then begin
    l2 := 1 / sqrt(l2);
    dir2[1] := l2 * gLow;
    dir2[2] := -(fLow + aux) * l2;
  end
  else begin
    l1 := 1 / sqrt(l1);
    dir2[1] := (aux - fLow) * l1;
    dir2[2] := eLow * l1;
  end;

  p := 0.025 * gSpan2D / hypot(dir1[1], dir1[2]);
  q := 0.025 * gSpan2D / hypot(dir2[1], dir2[2]);
  for i := 1 to 2 do begin
    dir1[i] := p * dir1[i];
    dir2[i] := q * dir2[i];
    pos1[i] := place[i] - dir1[i];
    pos2[i] := place[i] + dir1[i];
    pos3[i] := place[i] - dir2[i];
    pos4[i] := place[i] + dir2[i];
  end;
  FromParametersToDrawing(pos1, p1);
  FromParametersToDrawing(pos2, p2);
  FromParametersToDrawing(pos3, p3);
  FromParametersToDrawing(pos4, p4);
  XRiem := TRUE;
end; {XRiem}

procedure DrawXPseudoRiemPlane;
var
  aux, eLow, fLow, gLow, det: double;
  p: Segment2;
  can: TCanvas;

  procedure Normalize;
  var
    i, h, v: Integer;
  begin
    gX2D[0] := gCurPtDrawing2D;
    for i := 1 to 2 do begin
      aux := max(0.01, hypot(p[i, 1], p[i, 2]));
      p[i, 1] := p[i, 1] / aux;
      p[i, 2] := p[i, 2] / aux;
      h := round(p[i, 1] * kWidth / gEdges2D[1]);
      v := round(p[i, 2] * kHeight / gEdges2D[2]);
      gX2D[2 * i].x := gCurPtDrawing2D.x + h;
      gX2D[2 * i].y := gCurPtDrawing2D.y - v;
      gX2D[2 * i - 1].x := gCurPtDrawing2D.x - h;
      gX2D[2 * i - 1].y := gCurPtDrawing2D.y + v;
    end;
  end;

begin
  EraseX;
  eLow := gTheF1.V(gCur2DPos);
  fLow := gTheF2.V(gCur2DPos);
  gLow := gTheF3.V(gCur2DPos);
  det := sqr(fLow) - eLow * gLow;
  if det < 1.0E-14 then
    Exit;
  det := sqrt(det);
  if (abs(eLow) > 1.0E-8) and (abs(gLow) > 1.0E-8) then begin
    p[1, 1] := -gLow;
    p[1, 2] := det + fLow;
    p[2, 1] := det + fLow;
    p[2, 2] := -eLow;
  end
  else if abs(eLow) > 1.0E-8 then begin
    p[1, 1] := 0;
    p[1, 2] := 1;
    p[2, 1] := 2 * fLow;
    p[2, 2] := -eLow;
  end
  else if abs(gLow) > 1.0E-8 then begin
    p[1, 1] := 1;
    p[1, 2] := 0;
    p[2, 1] := gLow;
    p[2, 2] := -2 * fLow;
  end
  else begin
    p[1, 1] := 1;
    p[1, 2] := 0;
    p[2, 1] := 0;
    p[2, 2] := 1;
  end;
  Normalize;
  can := surface.Drawing2D.canvas;
  can.Pen.Color := clTeal;
  can.Pen.Mode := pmNotXoR;
  can.moveto(gX2D[1].x, gX2D[1].y);
  can.lineto(gX2D[2].x, gX2D[2].y);
  can.moveto(gX2D[3].x, gX2D[3].y);
  can.lineto(gX2D[4].x, gX2D[4].y);
  gThereIsX2D := True;
  ComputePseudoRiemCurvature(gCur2DPos, aux);
  surface.ShowCurvatures(false);
end;

function ComputeChristSymbPseudoRiem(p: vector3; var metric, gammaX, gammaY:
  Vector3):
  BOOLEAN;
var
  x1, x2, Ex, Ey, Fx, Fy, Gx, Gy, E, F, G, det: double;
begin
  result := FALSE;
  E := gTheF1.V(p);
  metric[1] := E;
  x1 := gTheF1.V([p[1] + gIncr2D[1], p[2]]);
  x2 := gTheF1.V([p[1] - gIncr2D[1], p[2]]);
  Ex := (x1 - x2) / (2 * gIncr2D[1]);
  x1 := gTheF1.V([p[1], p[2] + gIncr2D[2]]);
  x2 := gTheF1.V([p[1], p[2] - gIncr2D[2]]);
  Ey := (x1 - x2) / (2 * gIncr2D[2]);
  F := gTheF2.V(p);
  metric[2] := F;
  x1 := gTheF2.V([p[1] + gIncr2D[1], p[2]]);
  x2 := gTheF2.V([p[1] - gIncr2D[1], p[2]]);
  Fx := (x1 - x2) / (2 * gIncr2D[1]);
  x1 := gTheF2.V([p[1], p[2] + gIncr2D[2]]);
  x2 := gTheF2.V([p[1], p[2] - gIncr2D[2]]);
  Fy := (x1 - x2) / (2 * gIncr2D[2]);
  G := gTheF3.V(p);
  metric[3] := G;
  x1 := gTheF3.V([p[1] + gIncr2D[1], p[2]]);
  x2 := gTheF3.V([p[1] - gIncr2D[1], p[2]]);
  Gx := (x1 - x2) / (2 * gIncr2D[1]);
  x1 := gTheF3.V([p[1], p[2] + gIncr2D[2]]);
  x2 := gTheF3.V([p[1], p[2] - gIncr2D[2]]);
  Gy := (x1 - x2) / (2 * gIncr2D[2]);
  det := 2 * (E * G - sqr(F));
  if abs(det) > 1.0E-8 then begin
    result := True;
    gammaX[1] := (G * Ex - 2 * F * Fx + F * Ey) / det;
    gammaX[2] := (G * Ey - F * Gx) / det;
    gammaX[3] := (2 * G * Fy - G * Gx - F * Gy) / det;
    gammaY[1] := (2 * E * Fx - E * Ey - F * Ex) / det;
    gammaY[2] := (E * Gx - F * Ey) / det;
    gammaY[3] := (E * Gy - 2 * F * Fy + F * Gx) / det;
  end;
end; {ComputeChristSymb}



function CheckRangePseudoRiem(lastX, nextX: Vector3; findProlong: Boolean;
  var newX, newVel: Vector3): Integer;
var
  inside: Boolean;
begin
  inside := Min(gHalfEdges2D[1] - abs(nextX[1] - gCenter2D[1]),
    gHalfEdges2D[2] - abs(nextX[2] - gCenter2D[2])) >= 0;
  result := ord(inside);
end;




function DirectionsPseudoRiemPlane(var deriv: Vector3; prevDeriv, here:
  Vector3;
  axis: Integer): Boolean;
var
  i: integer;
  aux, eLow, fLow, gLow, l1, l2, p, q: double;
begin
  Result := False;
  eLow := gTheF1.V(here);
  fLow := gTheF2.V(here);
  gLow := gTheF3.V(here);
  aux := sqr(fLow) - eLow * gLow;
  if aux < 1.0E-7 then
    Exit;
  aux := (1 - 2 * axis) * sqrt(aux);
  l1 := sqr(-fLow + aux) + sqr(eLow);
  l2 := sqr(fLow + aux) + sqr(gLow);
  if l1 < l2 then begin
    l2 := 1 / sqrt(l2);
    p := l2 * gLow;
    q := -(fLow + aux) * l2;
  end
  else begin
    l1 := 1 / sqrt(l1);
    p := (aux - fLow) * l1;
    q := eLow * l1;
  end;
  deriv[1] := p;
  deriv[2] := q;
  aux := deriv[1] * prevDeriv[1] + deriv[2] * prevDeriv[2];
  if aux < 0 then
    for i := 1 to 2 do
      deriv[i] := -deriv[i];
  Result := True;
end;

function GeodAccelPseudoRiemPlane(var accel: Vector3; vel, x: vector3;
  toRight: Integer):
  Boolean;
var
  metric, gammaX, gammaY: Vector3;
  externalAccel: vector2;
  det: double;
begin
  result := FALSE;
  if ComputeChristSymbPseudoRiem(x, metric, gammaX, gammaY) then begin
    result := TRUE;
    accel[1] := -(gammaX[1] * sqr(vel[1]) + 2 * gammaX[2] * vel[1] *
      vel[2] + gammaX[3] * sqr(vel[2]));
    accel[2] := -(gammaY[1] * sqr(vel[1]) + 2 * gammaY[2] * vel[1] *
      vel[2] + gammaY[3] * sqr(vel[2]));
  end;
  if toRight <> 0 then begin
    det := sqrt(abs(metric[1] * metric[3] - sqr(metric[2])));
    externalAccel[1] := toRight * (vel[1] * metric[2] + vel[2] * metric[3])
      / det;
    externalAccel[2] := -toRight * (vel[1] * metric[1] + vel[2] * metric[2])
      / det;
    accel[1] := accel[1] + externalAccel[1];
    accel[2] := accel[2] + externalAccel[2];
  end;
end;

procedure RecordAndDrawSegmentPseudoRiemPlane(x1, x2: Vector3; toSave:
  Boolean);
var
  p1Window, p2Window: Vector2;
  can: TCanvas;
begin
  can := surface.Drawing2D.canvas;
  can.pen.mode := pmCopy;
  can.Pen.Color := gLinesData[gActiveLine].pen2DColor;
  FromParametersToDrawingReal(x1, p1Window);
  FromParametersToDrawingReal(x2, p2Window);
  AddSegment(p1Window, p2Window, 2, 0);
  can.moveto(round(p1Window[1]), round(p1Window[2]));
  can.lineto(round(p2Window[1]), round(p2Window[2]));
end;

function CameBackPseudoRiemPlane(firstX, prevX, curX: Vector3; counter:
  Integer): Boolean;
var
  bb, bc, cc, aux: Double;
begin
  result := False;
  if counter < 25 then
    Exit;
  bb := sqr(firstX[1] - prevX[1]) + sqr(firstX[2] - prevX[2]);
  if sqrt(bb) < gMaxSegmentLength2D then begin
    bc := (firstX[1] - prevX[1]) * (curX[1] - prevX[1]) + (firstX[2] -
      prevX[2]) * (curX[2] - prevX[2]);
    if bc > 0 then begin
      cc := sqr(curX[1] - prevX[1]) + sqr(curX[2] - prevX[2]);
      if cc > bc then begin
        aux := sqrt(abs(bb - sqr(bc) / cc));
        result := aux < 2 * gRadius2D;
      end;
    end;
  end;
end;

procedure DrawFirstOrderLines(fromHere: Vector2);
var
  axis, j: Integer;
  initvel: array[0..1] of Vector3;
  vel, here: Vector3;
  key: Integer;
begin
  Randomize;
  surface.Drawing2D.canvas.Pen.Mode := pmCopy;
  repeat
    for j := 1 to 2 do
      initVel[0, j] := Random;
  until (initVel[0, 1] <> 0) or (initVel[0, 2] <> 0);
  initVel[0, 3] := 0;
  here[1] := fromHere[1]; here[2] := fromHere[2];
  if not DirectionsPseudoRiemPlane(vel, initVel[0], here, 0) then
    Exit;
  DirectionsPseudoRiemPlane(initVel[1], initVel[0], here, 1);
  initVel[0] := vel;
  theRunKut.SetEnvironment(2, 1, kTol, DirectionsPseudoRiemPlane,
    CheckRangePseudoRiem,
    RecordAndDrawSegmentPseudoRiemPlane, CameBackPseudoRiemPlane, nil);
  for axis := 0 to 1 do begin
    vel := initVel[axis];
    here[1] := fromHere[1]; here[2] := fromHere[2];
    key := theRunKut.ComputeIntegralCurve(here, vel, axis);
    if ((key and kDoubled) = 0) then begin
      here[1] := fromHere[1]; here[2] := fromHere[2];
      vel[1] := -initVel[axis, 1];
      vel[2] := -initVel[axis, 2];
      theRunKut.ComputeIntegralCurve(here, vel, axis);
    end;
  end;
end;

procedure DrawCoordLinesPseudoRiemPlane(here: Vector2);
var
  pt: Vector2Single;
  par1, par2: vector2;
  can: TCanvas;
begin
  FromParametersToDrawingSingle(here, pt);
  can := Surface.Drawing2D.canvas;
  can.Pen.Color := gLinesData[kCoordinates].pen2DColor;
  can.Pen.Width := 1;
  can.Pen.Mode := pmCopy;

  par1[1] := 0;
  par1[2] := pt[2];
  can.moveto(0, round(pt[2]));
  par2[1] := kWidth;
  par2[2] := pt[2];
  can.lineto(kWidth, round(pt[2]));
  AddSegment(par1, par2, 2, 0);

  par1[1] := pt[1];
  par1[2] := 0;
  can.moveto(round(pt[1]), 0);
  par2[1] := pt[1];
  par2[2] := kHeight;
  can.lineto(round(pt[1]), kHeight);
  AddSegment(par1, par2, 2, 0);
end;

procedure DrawCurvesPseudoRiemPlane(fromHere: vector2);
begin
  if (not gAlreadyDrawn) then
    Exit;
  surface.cursor := crHourGlass;
  if gGeomOption = CoordLines then
    DrawCoordLinesPseudoRiemPlane(fromHere)
  else begin
    EraseX;
    DrawFirstOrderLines(fromHere);
  end;
  surface.cursor := crDefault;
end;

function ComputeFramePseudoRiemPlane(var u1, u2: Vector3; var phi, topT:
  Double; x: Vector3): double;
var
  E, F, G, a: Double;
  e1, e2, p: Vector2;
begin
  p[1] := x[1]; p[2] := x[2]; u1[3] := 0; u2[3] := 0;
  E := gTheF1.V(p);
  F := gTheF2.V(p);
  G := gTheF3.V(p);
  result := E * G - sqr(F);
  topT := sqrt(abs(E * sqr(u1[1]) + 2 * F * u1[1] * u1[2] + G * sqr(u1[2])));
  if result > 1.0E-12 then begin
    u2[1] := -(F * u1[1] + G * u1[2]);
    u2[2] := E * u1[1] + F * u1[2];
    a := sqrt(abs(E * sqr(u2[1]) + 2 * F * u2[1] * u2[2] + G
      * sqr(u2[2])));
    if a < 1.0E-7 then
      result := 0
    else begin
      a := topT / a;
      u2[1] := a * u2[1];
      u2[2] := a * u2[2];
      u2[3] := 0;
    end;
  end
  else if result < -1.0E-7 then begin
    a := sqrt(-result);
    if abs(E) > 1.0E-7 then begin
      e1[1] := -F + a; e1[2] := E;
      e2[1] := -F - a; e2[2] := E
    end
    else begin
      e1[1] := 1; e1[2] := 0;
      e2[1] := -G; e2[2] := 2 * F;
    end;
    a := hypot(e1[1], e1[2]);
    e1[1] := e1[1] / a; e1[2] := e1[2] / a;
    a := hypot(e2[1], e2[2]);
    e2[1] := e2[1] / a; e2[2] := e2[2] / a;
    a := e1[1] * e2[1] + e1[2] * e2[2];
    if a < 0 then begin
      e2[1] := -e2[1]; e2[2] := -e2[2];
    end;
    a := abs(a);
    phi := arccos(a) / 2;
    u1[1] := (e1[1] + e2[1]) / sqrt(2 * (1 + a));
    u1[2] := (e1[2] + e2[2]) / sqrt(2 * (1 + a));
    u2[1] := (e1[1] - e2[1]) / sqrt(2 * (1 - a));
    u2[2] := (e1[2] - e2[2]) / sqrt(2 * (1 - a));
  end
  else
    result := 0;
end;


end.
