

{-----------------------------------------------------------------------------
 Unit Name: DrawParSurfNoShadows
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Drawing of a surface in parametric form without shadows
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}



unit DrawParSurfNoShadows;

interface
uses
 Classes,
{$IFDEF MSWINDOWS}
  Windows, Graphics,
{$ENDIF}
{$IFDEF LINUX}
 Types, QGraphics,
{$ENDIF}
  FSurface;

procedure DrawSurfaceWithoutShadows;

implementation

uses
Math, SysUtils,
{$IFDEF MSWINDOWS}
 MMSystem,
{$ENDIF}
{$IFDEF LINUX}
  DateUtils,
{$ENDIF}
  Globals, OcFunct, Utilities;

type
  triangCol = array[1..3] of PointData;
  threeBytes = array[1..3] of Byte;


var
  gId: threeBytes;
  gInitVal: array[1..3] of array of single;
  cVal: array of Single;
  cValDerivs: array of Vector2Single;

procedure InitializeZBuffer;
begin
  FillChar(gZBuffer^, SizeOf(gZBuffer^), #$7F);
end;



function ComputeSpecialNormal(i, j: Integer; var n: vector3Single; var norm:
  single): Boolean;
var
  k: Integer;
  ph, pv: Vector3Single;
begin
  if i < gGridU then begin
    if j < gGridV then
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i + 1, j + 1].pos[k] - gPointsData[i,
          j].pos[k];
        pv[k] := gPointsData[i, j + 1].pos[k] - gPointsData[i + 1,
          j].pos[k];
      end
    else
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i, j - 1].pos[k] - gPointsData[i + 1,
          j].pos[k];
        pv[k] := gPointsData[i + 1,
          j - 1].pos[k] - gPointsData[i, j].pos[k];
      end
  end
  else begin
    if j < gGridV then
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i - 1, j].pos[k] - gPointsData[i,
          j + 1].pos[k];
        pv[k] := gPointsData[i, j].pos[k] - gPointsData[i - 1,
          j + 1].pos[k];
      end
    else
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i, j].pos[k] - gPointsData[i - 1,
          j - 1].pos[k];
        pv[k] := gPointsData[i - 1, j].pos[k] - gPointsData[i,
          j - 1].pos[k];
      end
  end;
  n[1] := ph[2] * pv[3] - ph[3] * pv[2];
  n[2] := ph[3] * pv[1] - ph[1] * pv[3];
  n[3] := ph[1] * pv[2] - ph[2] * pv[1];
  norm := sqrt(sqr(n[1]) + sqr(n[2]) + sqr(n[3]));
  Result := norm > 1.0E-11;
end;

procedure AssignCol(i, j: Integer; ph, pv:
  Vector3);
{It is assumed that at the TPoint (i,j) of the grid the surface}
{TPoint is 'pos', and 'ph', 'pv' are two vectors tangent to the surface}
{at 'pos'. The brightness of that TPoint is computed and assigned}
{to gPointsData[i, j].br.}
const
  intAmb: double = 96;
var
  ii: Integer;
  ray, normal: Vector3Single;
  rayNorm, aux, sigma: single;
begin
  for ii := 1 to 3 do
    ray[ii] := gViewpoint[ii] - gPointsData[i, j].pos[ii];
  rayNorm := sqrt(sqr(ray[1]) + sqr(ray[2]) + sqr(ray[3]));
  for ii := 1 to 3 do
    ray[ii] := ray[ii] / rayNorm;
  normal[1] := ph[2] * pv[3] - ph[3] * pv[2];
  normal[2] := ph[3] * pv[1] - ph[1] * pv[3];
  normal[3] := ph[1] * pv[2] - ph[2] * pv[1];
  aux := sqrt(sqr(normal[1]) + sqr(normal[2]) + sqr(normal[3]));
  if (aux > 1.0E-11) or ComputeSpecialNormal(i, j, normal, aux) then begin
    for ii := 1 to 3 do
      normal[ii] := normal[ii] / aux;
    sigma := ray[1] * normal[1] + ray[2] * normal[2] + ray[3] * normal[3];
    gPointsData[i, j].ambLight := intAmb * abs(sigma);
    if sigma < 0 then
      gPointsData[i, j].ambLight := -gPointsData[i, j].ambLight;
  end
  else
    gPointsData[i, j].ambLight := 100;
end;

procedure FillDataPoints(di, dj: Integer;
  ph, pv: Vector3);
var
  i: Integer;
  p: Vector3Single;
  a: Single;
begin
  p := gPointsData[di, dj].pos;
  gPointsData[di, dj].outside := False;
  gPointsData[di, dj].inside := True;
  gPointsData[di, dj].proj := ProjectSToPaneRealAndDist(p);
  AssignCol(di, dj, ph, pv);
  if gClips <> nil then
    for i := 0 to High(gClips) do begin
      a := HeightOverClipS(gClips[i], p);
      if a > 0 then begin
        gPointsData[di, dj].inside := False;
        a := a - (abs(gClips[i].u[1]) * (abs(ph[1]) + abs(pv[1])) +
          abs(gClips[i].u[2]) * (abs(ph[2]) + abs(pv[2])) +
          abs(gClips[i].u[3]) * (abs(ph[3]) + abs(pv[3])));
        gPointsData[di, dj].outside := (a > 0);
        if gPointsData[di, dj].outside then Exit;
      end
    end;
end;




procedure ComputeDataGrid;
 {We compute the illumination that would reach the}
 {eye from each point of the grid as if it were visible from the eye. }
 {We take advantage of the stored data as far as possible.}
const
  cc = 16;
var
  i, j, k: Integer;
  u, w, hu, hv: double;
  p, pv, ph: Vector3;
begin
  hu := gStep2D[1] / cc;
  hv := gStep2D[2] / cc;
  for i := 0 to gGridU do
    for j := 0 to gGridV do begin
      theFPar(p, gMin2D[1] + i * gStep2D[1], gMin2D[2] + j * gStep2D[2]);
      gPointsData[i, j].pos[1] := p[1];
      gPointsData[i, j].pos[2] := p[2];
      gPointsData[i, j].pos[3] := p[3];
    end;

 {First, the corners}
  u := gMin2D[1];
  w := gMin2D[2];
  theFPar(ph, u + hu, w);
  theFPar(pv, u, w + hv);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[0, 0].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[0, 0].pos[k]);
  end;
  FillDataPoints(0, 0, ph, pv);

  u := gMax2D[1];
  theFPar(ph, u, w + hv);
  theFPar(pv, u - hu, w);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[gGridU, 0].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[gGridU, 0].pos[k]);
  end;
  FillDataPoints(gGridU, 0, ph, pv);

  w := gMax2D[2];
  theFPar(ph, u - hu, w);
  theFPar(pv, u, w - hv);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[gGridU, gGridV].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[gGridU, gGridV].pos[k]);
  end;
  FillDataPoints(gGridU, gGridV, ph, pv);

  u := gMin2D[1];
  theFPar(ph, u, w - hv);
  theFPar(pv, u + hu, w);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[0, gGridV].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[0, gGridV].pos[k]);
  end;
  FillDataPoints(0, gGridV, ph, pv);

 {now the limiting rows}
  for i := 1 to gGridU - 1 do begin
    u := gMin2D[1] + i * gStep2D[1];
    w := gMin2D[2];
    theFPar(pv, u, w + hv);
    theFPar(ph, u + hu, w);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[i, 0].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[i, 0].pos[k]);
    end;
    FillDataPoints(i, 0, ph, pv);

    w := gMax2D[2];
    theFPar(pv, u, w - hv);
    theFPar(ph, u - hu, w);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[i, gGridV].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[i, gGridV].pos[k]);
    end;
    FillDataPoints(i, gGridV, ph, pv);
  end;

  for i := 1 to gGridV - 1 do begin
    w := gMin2D[2] + i * gStep2D[2];
    u := gMax2D[1];
    theFPar(pv, u - hu, w);
    theFPar(ph, u, w + hv);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[gGridU, i].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[gGridU, i].pos[k]);
    end;
    FillDataPoints(gGridU, i, ph, pv);

    u := gMin2D[1];
    theFPar(ph, u + hu, w);
    theFPar(pv, u, w + hv);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[0, i].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[0, i].pos[k]);
    end;
    FillDataPoints(0, i, ph, pv);

  end;

  {And, finally, the interior of the domain}
  for i := 1 to gGridU - 1 do
    for j := 1 to gGridV - 1 do begin
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i + 1, j].pos[k] - gPointsData[i,
          j].pos[k];
        pv[k] := gPointsData[i, j + 1].pos[k] - gPointsData[i,
          j].pos[k];
      end;
      FillDataPoints(i, j, ph, pv);
    end;
end;


procedure FillTriangle(triCol: triangCol);
var
{x,y are window coordinates; u,v parameters; }
  permu: threeBytes;
  virt: TPoint;
  A, B, C: PointData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  val, initialVal: array[1..2] of single;
  derivs: array[1..2, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, fromHere, toThere, pm, cc: Integer;
  auxab, auxac, auxbc: double;
  line: pTColorArray;
  gotDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    derivs[1, 1] := A.ambLight * ax + B.ambLight * bx + C.ambLight * cx;
    derivs[1, 2] := A.ambLight * ay + B.ambLight * by + C.ambLight * cy;
    derivs[2, 1] := A.proj[3] * ax + B.proj[3] * bx + C.proj[3] * cx;
    derivs[2, 2] := A.proj[3] * ay + B.proj[3] * by + C.proj[3] * cy;
    gotDerivs := True;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not gotDerivs then
      ComputeDerivatives;
    for ii := 1 to 2 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 2 do
      val[ii] := val[ii] + derivs[ii, 1];
  end;

begin
  gotDerivs := False;
  permu := gId;
  if triCol[2].proj[2] < triCol[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triCol[3].proj[2] < triCol[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triCol[3].proj[2] < triCol[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triCol[permu[1]];
  B := triCol[permu[2]];
  C := triCol[permu[3]];

  ACx := A.proj[1] - C.proj[1];
  ACy := A.proj[2] - C.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.proj[1] - B.proj[1];
  ABy := A.proj[2] - B.proj[2];
  BCx := B.proj[1] - C.proj[1];
  BCy := B.proj[2] - C.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;

  cc := ord(A.ambLight < 0) + ord(B.ambLight < 0) + ord(C.ambLight < 0);
  if cc > 1 then begin
    A.ambLight := abs(A.ambLight) + 101;
    B.ambLight := abs(B.ambLight) + 101;
    C.ambLight := abs(C.ambLight) + 101;
  end
  else begin
    A.ambLight := abs(A.ambLight);
    B.ambLight := abs(B.ambLight);
    C.ambLight := abs(C.ambLight);
  end;

  coefAC := ACx / ACy;
  initialVal[1] := A.ambLight;
  initialVal[2] := A.proj[3];

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(kHeightMinus1, floor(B.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.proj[2];
      auxac := A.proj[1] + difj * coefAC;
      auxab := A.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        line := gMovingBmp.ScanLine[j];
        if CAtLeftOfB then begin
          fromHere := Max(0, ceil(auxac));
          toThere := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          fromHere := Max(0, ceil(auxab));
          toThere := Min(floor(auxac), kWidthMinus1);
        end;

        if fromHere <= toThere then begin
          difi := fromHere - A.proj[1] - 1;
          ComputeData;
          for i := fromHere to toThere do begin
            difi := difi + 1;
            IncrementI;
            if (val[2] < gZBuffer^[i, j]) then begin
              gZBuffer^[i, j] := val[2];
              pm := floor(val[1]);
              line[i] := gPalette3D^[pm];
            end
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(kHeightMinus1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.proj[2];
    difb := jInf - 1 - B.proj[2];
    auxac := A.proj[1] + difj * coefAC;
    auxbc := B.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      line := gMovingBmp.ScanLine[j];
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        fromHere := Max(0, ceil(auxac));
        toThere := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        fromHere := Max(0, ceil(auxbc));
        toThere := Min(floor(auxac), kWidthMinus1);
      end;
      if fromHere <= toThere then begin
        difi := fromHere - A.proj[1] - 1;
        ComputeData;
        for i := fromHere to toThere do begin
          difi := difi + 1;
          IncrementI;
          if (val[2] < gZBuffer^[i, j]) then begin
            gZBuffer^[i, j] := val[2];
            pm := floor(val[1]);
            line[i] := gPalette3D^[pm];
          end
        end;
      end;
    end;
  end;
end;

procedure FillTriangleWithClips(triCol: triangCol);
var
{x,y are window coordinates; u,v parameters; }
  permu: threeBytes;
  virt: TPoint;
  A, B, C: PointData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  val, initialVal: array[1..2] of single;
  derivs: array[1..2, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, fromHere, toThere, pm, cc, k: Integer;
  auxab, auxac, auxbc: double;
  line: pTColorArray;
  gotDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    derivs[1, 1] := A.ambLight * ax + B.ambLight * bx + C.ambLight * cx;
    derivs[1, 2] := A.ambLight * ay + B.ambLight * by + C.ambLight * cy;
    derivs[2, 1] := A.proj[3] * ax + B.proj[3] * bx + C.proj[3] * cx;
    derivs[2, 2] := A.proj[3] * ay + B.proj[3] * by + C.proj[3] * cy;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := gInitVal[1, k] * ax + gInitVal[2, k] * bx +
        gInitVal[3, k] * cx;
      cValDerivs[k, 2] := gInitVal[1, k] * ay + gInitVal[2, k] * by +
        gInitVal[3, k] * cy;
    end;

    gotDerivs := True;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not gotDerivs then
      ComputeDerivatives;
    for ii := 1 to 2 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
    for ii := 0 to gClipTop do
      cVal[ii] := gInitVal[1, ii] + cValDerivs[ii, 1] * difi + cValDerivs[ii,
        2]
        *
        difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 2 do
      val[ii] := val[ii] + derivs[ii, 1];
    for ii := 0 to gClipTop do
      cVal[ii] := cVal[ii] + cValDerivs[ii, 1];
  end;

begin
  gotDerivs := False;
  permu := gId;
  if triCol[2].proj[2] < triCol[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triCol[3].proj[2] < triCol[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triCol[3].proj[2] < triCol[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triCol[permu[1]];
  B := triCol[permu[2]];
  C := triCol[permu[3]];

  ACx := A.proj[1] - C.proj[1];
  ACy := A.proj[2] - C.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.proj[1] - B.proj[1];
  ABy := A.proj[2] - B.proj[2];
  BCx := B.proj[1] - C.proj[1];
  BCy := B.proj[2] - C.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;

  cc := ord(A.ambLight < 0) + ord(B.ambLight < 0) + ord(C.ambLight < 0);
  if cc > 1 then begin
    A.ambLight := abs(A.ambLight) + 101;
    B.ambLight := abs(B.ambLight) + 101;
    C.ambLight := abs(C.ambLight) + 101;
  end
  else begin
    A.ambLight := abs(A.ambLight);
    B.ambLight := abs(B.ambLight);
    C.ambLight := abs(C.ambLight);
  end;

  coefAC := ACx / ACy;
  initialVal[1] := A.ambLight;
  initialVal[2] := A.proj[3];
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      gInitVal[i, j] := HeightOverClipS(gClips[j], tricol[permu[i]].pos);

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(kHeightMinus1, floor(B.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.proj[2];
      auxac := A.proj[1] + difj * coefAC;
      auxab := A.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        line := gMovingBmp.ScanLine[j];
        if CAtLeftOfB then begin
          fromHere := Max(0, ceil(auxac));
          toThere := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          fromHere := Max(0, ceil(auxab));
          toThere := Min(floor(auxac), kWidthMinus1);
        end;

        if fromHere <= toThere then begin
          difi := fromHere - A.proj[1] - 1;
          ComputeData;
          for i := fromHere to toThere do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);

            if good and (val[2] < gZBuffer^[i, j]) then begin
              gZBuffer^[i, j] := val[2];
              pm := floor(val[1]);
              line[i] := gPalette3D^[pm];
            end
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(kHeightMinus1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.proj[2];
    difb := jInf - 1 - B.proj[2];
    auxac := A.proj[1] + difj * coefAC;
    auxbc := B.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      line := gMovingBmp.ScanLine[j];
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        fromHere := Max(0, ceil(auxac));
        toThere := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        fromHere := Max(0, ceil(auxbc));
        toThere := Min(floor(auxac), kWidthMinus1);
      end;
      if fromHere <= toThere then begin
        difi := fromHere - A.proj[1] - 1;
        ComputeData;
        for i := fromHere to toThere do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good and (val[2] < gZBuffer^[i, j]) then begin
            gZBuffer^[i, j] := val[2];
            pm := floor(val[1]);
            line[i] := gPalette3D^[pm];
          end
        end;
      end;
    end;
  end;
end;


procedure DrawSurfaceWithoutShadows;
var
  i, j: Integer;
  triCol: triangCol;
begin
  ComputeDataGrid;
  InitializeZBuffer;
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  if gClips <> nil then begin
    for i := 1 to 3 do begin
      gInitVal[i] := nil;
      SetLength(gInitVal[i], gClipTop + 1);
    end;
    cVal := nil;
    SetLength(cVal, gClipTop + 1);
    cValDerivs := nil;
    SetLength(cValDerivs, gClipTop + 1);
  end;
  for i := 1 to gGridU do
    for j := 1 to gGridV do begin
      triCol[1] := gPointsData[i, j];
      triCol[2] := gPointsData[i - 1, j - 1];
      triCol[3] := gPointsData[i - 1, j];
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and gPointsData[i - 1, j].inside then
        FillTriangle(triCol)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i - 1, j].outside) then
        FillTriangleWithClips(triCol);

      triCol[3] := gPointsData[i, j - 1];
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and gPointsData[i, j - 1].inside then
        FillTriangle(triCol)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i, j - 1].outside) then
        FillTriangleWithClips(triCol);
    end;
end;


procedure WashOut;
var
  i: Integer;
begin
  for i := 1 to 3 do
    gInitVal[i] := nil;
  cVal := nil;
  cValDerivs := nil;
end;

initialization
  ;
finalization
  WashOut;

end.
