

{-----------------------------------------------------------------------------
 Unit Name: DrawParamSurf
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Drawing of a surface given in parametric form (x(u,v), y(u,v), z(u,v)
    the surface is illuminated by a pointlike focus that casts shadows
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit DrawParamSurf;
interface
uses
{$IFDEF MSWINDOWS}
  Windows, Graphics,
{$ENDIF}
{$IFDEF LINUX}
  Types, QGraphics,
{$ENDIF}
Globals;

procedure ComputeParamSurface(bm: TBitmap; var map: ColorMapPtr);
procedure ComputeZBuffer;
procedure ComputeZBufferWithClips;
procedure ComputeNormalPairParam;
function DrawParametricSurface: Boolean;


implementation
uses
 SysUtils, Math,
{$IFDEF MSWINDOWS}
  Forms,
{$ENDIF}
{$IFDEF LINUX}
  QForms,
{$ENDIF}
  Utilities, FSurface, OcFunct, Antialiasing, ParallelFrame,
  InitParams, ResStrings, Manageclips;
type
  ptDataAndParam = record
    c: PointData;
    u, v: Integer;
  end;

  threeBytes = array[1..3] of Byte;
  triang = array[1..3] of ptDataAndParam;
  triangFoc = array[1..3] of FocusSpacePtData;
var
  gId: threeBytes;
  theMap: ColorMapPtr;
  cInitVal: array[1..3] of array of single;
  cVal: array of Single;
  cValDerivs: array of Vector2Single;

procedure InitializeZBuffer;
begin
  FillChar(gZBuffer^, SizeOf(gZBuffer^), #$7F);
end;

procedure InitializeParamBuffer;
begin
  FillChar(gParamBuffer^, SizeOf(gParamBuffer^), #$80);
end;

procedure InitializeFBuffer;
begin
  FillChar(gFBuffer^, SizeOf(gFBuffer^), #$7F);
end;

function ComputeSpecialNormal(i, j: Integer; var n: Vector3Single; var norm:
  single):
  Boolean;
var
  k: Integer;
  ph, pv: Vector3Single;
begin
  if i < gGridU then begin
    if j < gGridV then
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i + 1, j + 1].pos[k] - gPointsData[i,
          j].pos[k];
        pv[k] := gPointsData[i, j + 1].pos[k] - gPointsData[i + 1,
          j].pos[k];
      end
    else
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i, j - 1].pos[k] - gPointsData[i + 1,
          j].pos[k];
        pv[k] := gPointsData[i + 1,
          j - 1].pos[k] - gPointsData[i, j].pos[k];
      end
  end
  else begin
    if j < gGridV then
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i - 1, j].pos[k] - gPointsData[i,
          j + 1].pos[k];
        pv[k] := gPointsData[i, j].pos[k] - gPointsData[i - 1,
          j + 1].pos[k];
      end
    else
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i, j].pos[k] - gPointsData[i - 1,
          j - 1].pos[k];
        pv[k] := gPointsData[i - 1, j].pos[k] - gPointsData[i,
          j - 1].pos[k];
      end
  end;
  n[1] := ph[2] * pv[3] - ph[3] * pv[2];
  n[2] := ph[3] * pv[1] - ph[1] * pv[3];
  n[3] := ph[1] * pv[2] - ph[2] * pv[1];
  norm := EuclNorm3S(n);
  Result := norm > 1.0E-11;
end;

procedure AssignCol(i, j: Integer; ph, pv:
  Vector3);
const
  dullness: double = 0.3;
  OneMinusDullness: double = 1 - 0.3 {1 - dullness};
  intFocus: double = 60;
  intAmb: double = 40;
var
  ii: Integer;
  ray, normal, fRay: Vector3Single;
  rayNorm, fRayNorm, aux, sigma, lambda, omega, fComp: single;
begin
  for ii := 1 to 3 do begin
    ray[ii] := gViewpoint[ii] - gPointsData[i, j].pos[ii];
    fRay[ii] := gFocus[ii] - gPointsData[i, j].pos[ii];
  end;
  rayNorm := EuclNorm3S(ray);
  fRayNorm := EuclNorm3S(fRay);
  for ii := 1 to 3 do begin
    ray[ii] := ray[ii] / rayNorm;
    fRay[ii] := fRay[ii] / fRayNorm;
  end;
  normal[1] := ph[2] * pv[3] - ph[3] * pv[2];
  normal[2] := ph[3] * pv[1] - ph[1] * pv[3];
  normal[3] := ph[1] * pv[2] - ph[2] * pv[1];
  aux := EuclNorm3S(normal);
  if (aux > 1.0E-11) or ComputeSpecialNormal(i, j, normal, aux) then begin
    for ii := 1 to 3 do
      normal[ii] := normal[ii] / aux;
    sigma := ray[1] * normal[1] + ray[2] * normal[2] + ray[3] * normal[3];
    lambda := fRay[1] * normal[1] + fRay[2] * normal[2] + fRay[3] *
      normal[3];
    omega := 2 * lambda * sigma;
    gPointsData[i, j].ambLight := intAmb * (1 + abs(sigma)) / 2;
    if sigma < 0 then
      gPointsData[i, j].ambLight := -gPointsData[i, j].ambLight;
    if omega > 0 then begin
      lambda := abs(lambda);
      omega := (1 + omega - (fRay[1] * ray[1] + fRay[2] * ray[2] +
        fRay[3] * ray[3])) / 2;
      fComp := lambda * (OneMinusDullness * sqr(omega) + dullness * (1 +
        abs(sigma)) / 2);
      gFocusPtData[i, j].focLight := Min(99 - abs(gPointsData[i,
        j].ambLight), intFocus * fComp);
    end
    else
      gFocusPtData[i, j].focLight := 0;
  end
  else begin
    gPointsData[i, j].ambLight := 100;
    gFocusPtData[i, j].focLight := 0;
  end;

end;



procedure FillPointData(di, dj: Integer;
  ph, pv: Vector3; allData: Boolean);
var
  i: Integer;
  p: Vector3Single;
  a: Single;
begin
  p := gPointsData[di, dj].pos;
  gPointsData[di, dj].outside := False;
  gPointsData[di, dj].inside := True;
  gPointsData[di, dj].proj := ProjectSToPaneRealAndDist(p);
  if allData then begin
    gFocusPtData[di, dj].pt := ProjectToFocusPlaneRealAndDist(p);
    AssignCol(di, dj, ph, pv);
  end;
  if gClips <> nil then
    for i := 0 to High(gClips) do begin
      a := HeightOverClipS(gClips[i], p);
      if a > 0 then begin
        gPointsData[di, dj].inside := False;
        a := a - (abs(gClips[i].u[1]) * (abs(ph[1]) + abs(pv[1])) +
          abs(gClips[i].u[2]) * (abs(ph[2]) + abs(pv[2])) +
          abs(gClips[i].u[3]) * (abs(ph[3]) + abs(pv[3])));
        gPointsData[di, dj].outside := (a > 0);
        if gPointsData[di, dj].outside then Exit;
      end
    end;
end;



procedure ComputeGridData(allData: Boolean);
 {We compute the illumination that would reach the}
 {eye from each point of the grid as if it were visible from the eye. }
 {We take advantage of the stored data as far as possible.}
const
  cc = 16;
var
  i, j, k: Integer;
  u, w, hu, hv: double;
  p, pv, ph: Vector3;
  aPointData: PointData;
  aFocusSpaceData: FocusSpacePtData;
begin
  hu := gStep2D[1] / 10;
  hv := gStep2D[2] / 10;
  for i := 0 to gGridU do
    for j := 0 to gGridV do begin
      theFPar(p, gMin2D[1] + i * gStep2D[1], gMin2D[2] + j * gStep2D[2]);
      gPointsData[i, j].pos[1] := p[1];
      gPointsData[i, j].pos[2] := p[2];
      gPointsData[i, j].pos[3] := p[3];
    end;

 {First, the corners}
  u := gMin2D[1];
  w := gMin2D[2];
  theFPar(ph, u + hu, w);
  theFPar(pv, u, w + hv);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[0, 0].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[0, 0].pos[k]);
  end;
  FillPointData(0, 0, ph, pv, allData);

  u := gMax2D[1];
  theFPar(ph, u, w + hv);
  theFPar(pv, u - hu, w);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[gGridU, 0].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[gGridU, 0].pos[k]);
  end;
  FillPointData(gGridU, 0, ph, pv, allData);

  w := gMax2D[2];
  theFPar(ph, u - hu, w);
  theFPar(pv, u, w - hv);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[gGridU, gGridV].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[gGridU, gGridV].pos[k]);
  end;
  FillPointData(gGridU, gGridV, ph, pv, allData);

  u := gMin2D[1];
  theFPar(ph, u, w - hv);
  theFPar(pv, u + hu, w);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - gPointsData[0, gGridV].pos[k]);
    pv[k] := cc * (pv[k] - gPointsData[0, gGridV].pos[k]);
  end;
  FillPointData(0, gGridV, ph, pv, allData);

 {now the limiting rows}
  for i := 1 to gGridU - 1 do begin
    u := gMin2D[1] + i * gStep2D[1];
    w := gMin2D[2];
    theFPar(pv, u, w + hv);
    theFPar(ph, u + hu, w);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[i, 0].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[i, 0].pos[k]);
    end;
    FillPointData(i, 0, ph, pv, allData);

    w := gMax2D[2];
    theFPar(pv, u, w - hv);
    theFPar(ph, u - hu, w);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[i, gGridV].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[i, gGridV].pos[k]);
    end;
    FillPointData(i, gGridV, ph, pv, allData);
  end;

  for i := 1 to gGridV - 1 do begin
    w := gMin2D[2] + i * gStep2D[2];
    u := gMax2D[1];
    theFPar(pv, u - hu, w);
    theFPar(ph, u, w + hv);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[gGridU, i].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[gGridU, i].pos[k]);
    end;
    FillPointData(gGridU, i, ph, pv, allData);

    u := gMin2D[1];
    theFPar(ph, u + hu, w);
    theFPar(pv, u, w + hv);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - gPointsData[0, i].pos[k]);
      pv[k] := cc * (pv[k] - gPointsData[0, i].pos[k]);
    end;
    FillPointData(0, i, ph, pv, allData);

  end;

  {And, finally, the interior of the domain}
  for i := 1 to gGridU - 1 do
    for j := 1 to gGridV - 1 do begin
      for k := 1 to 3 do begin
        ph[k] := gPointsData[i + 1, j].pos[k] - gPointsData[i,
          j].pos[k];
        pv[k] := gPointsData[i, j + 1].pos[k] - gPointsData[i,
          j].pos[k];
      end;
      FillPointData(i, j, ph, pv, allData);
    end;
end;


procedure FillTriangle(bm: TBitmap; vertices: triang; triangleF:
  triangFoc);
var
{x,y Drawing3D coordinates; u,v parameters of the point}
{p,q coordinates in the virtual screen used for the Focus}
  permu: threeBytes;
  virt: TPoint;
  A, B, C: ptDataAndParam;
  AE, BE, CE: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy, light: single;
  dfp, dfq, dfExtrapol: single;
  val, initialVal: array[1..8] of single;
  derivs: array[1..8, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, pm: Integer;
  auxab, auxac, auxbc: double;
  line: pTColorArray;
  noFBorder, haveDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    derivs[1, 1] := A.c.ambLight * ax + B.c.ambLight * bx + C.c.ambLight * cx;
    derivs[1, 2] := A.c.ambLight * ay + B.c.ambLight * by + C.c.ambLight * cy;
    derivs[2, 1] := A.u * ax + B.u * bx + C.u * cx;
    derivs[2, 2] := A.u * ay + B.u * by + C.u * cy;
    derivs[3, 1] := A.v * ax + B.v * bx + C.v * cx;
    derivs[3, 2] := A.v * ay + B.v * by + C.v * cy;
    derivs[4, 1] := AE.focLight * ax + BE.focLight * bx + CE.focLight * cx;
    derivs[4, 2] := AE.focLight * ay + BE.focLight * by + CE.focLight * cy;
    derivs[5, 1] := AE.pt[1] * ax + BE.pt[1] * bx +
      CE.pt[1] * cx;
    derivs[5, 2] := AE.pt[1] * ay + BE.pt[1] * by +
      CE.pt[1] * cy;
    derivs[6, 1] := AE.pt[2] * ax + BE.pt[2] * bx +
      CE.pt[2] * cx;
    derivs[6, 2] := AE.pt[2] * ay + BE.pt[2] * by +
      CE.pt[2] * cy;
    derivs[7, 1] := A.c.proj[3] * ax + B.c.proj[3] * bx + C.c.proj[3] * cx;
    derivs[7, 2] := A.c.proj[3] * ay + B.c.proj[3] * by + C.c.proj[3] * cy;
    derivs[8, 1] := AE.pt[3] * ax + BE.pt[3] * bx +
      CE.pt[3] * cx;
    derivs[8, 2] := AE.pt[3] * ay + BE.pt[3] * by +
      CE.pt[3] * cy;
    det := derivs[6, 2] * derivs[5, 1] - derivs[6, 1] * derivs[5, 2];
    noFBorder := abs(det) > 1.0E-7;
    if noFBorder then begin
      dfp := (derivs[7, 1] * derivs[6, 2] - derivs[7, 2] * derivs[6, 1]) /
        det;
      dfq := (derivs[7, 2] * derivs[5, 1] - derivs[7, 1] * derivs[5, 2]) /
        det;
    end;
    haveDerivs := True;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    for ii := 1 to 8 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 8 do
      val[ii] := val[ii] + derivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].c.proj[2] < vertices[1].c.proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].c.proj[2] < vertices[permu[1]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].c.proj[2] < vertices[permu[2]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  AE := triangleF[permu[1]];
  BE := triangleF[permu[2]];
  CE := triangleF[permu[3]];

  ACx := A.c.proj[1] - C.c.proj[1];
  ACy := A.c.proj[2] - C.c.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.c.proj[1] - B.c.proj[1];
  ABy := A.c.proj[2] - B.c.proj[2];
  BCx := B.c.proj[1] - C.c.proj[1];
  BCy := B.c.proj[2] - C.c.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;

  auxab := A.c.ambLight + B.c.ambLight + C.c.ambLight;
  if auxab < 0 then begin
    A.c.ambLight := abs(A.c.ambLight) + 101;
    B.c.ambLight := abs(B.c.ambLight) + 101;
    C.c.ambLight := abs(C.c.ambLight) + 101;
  end
  else begin
    A.c.ambLight := abs(A.c.ambLight);
    B.c.ambLight := abs(B.c.ambLight);
    C.c.ambLight := abs(C.c.ambLight);
  end;

  coefAC := ACx / ACy;
  initialVal[1] := A.c.ambLight;
  initialVal[2] := A.u;
  initialVal[3] := A.v;
  initialVal[4] := AE.FocLight;
  initialVal[5] := AE.pt[1];
  initialVal[6] := AE.pt[2];
  initialVal[7] := A.c.proj[3];
  initialVal[8] := AE.pt[3];



  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.c.proj[2]));
    jSup := min(kHeightMinus1, floor(B.c.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.c.proj[2];
      auxac := A.c.proj[1] + difj * coefAC;
      auxab := A.c.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        line := bm.ScanLine[j];
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), kWidthMinus1);
        end;

        if here <= there then begin
          difi := here - A.c.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            virt.x := Floor(val[5]);
            virt.y := Floor(val[6]);
            light := 0;
            if (virt.x >= 0) and (virt.x < kWidth) and (virt.y
              >= 0) and (virt.y < kHeight) and noFBorder then begin
              dfExtrapol := val[8] + dfp * (virt.x -
                val[5]) + dfq * (virt.y - val[6]);
              if dfExtrapol < 1.002 * gfBuffer^[virt.x, virt.y] then
                light := val[4];
            end;
            if (val[7] < gZBuffer^[i, j]) then begin
              gZBuffer^[i, j] := val[7];
              gParamBuffer^[i, j, 1] := val[2];
              gParamBuffer^[i, j, 2] := val[3];
              light := light + val[1];
              pm := floor(light);
              if theMap <> nil then
                theMap^[i, j] := pm;
              line^[i] := gPalette3D^[pm];
            end
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.c.proj[2]));
    jSup := min(kHeightMinus1, floor(C.c.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.c.proj[2];
    difb := jInf - 1 - B.c.proj[2];
    auxac := A.c.proj[1] + difj * coefAC;
    auxbc := B.c.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      line := bm.ScanLine[j];
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), kWidthMinus1);
      end;
      if here <= there then begin
        difi := here - A.c.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          virt.x := Floor(val[5]);
          virt.y := Floor(val[6]);
          light := 0;
          if (virt.x >= 0) and (virt.x < kWidth) and (virt.y
            >= 0) and (virt.y < kHeight) and noFBorder then begin
            dfExtrapol := val[8] + dfp * (virt.x -
              val[5]) + dfq * (virt.y - val[6]);
            if dfExtrapol < 1.002 * gfBuffer^[virt.x, virt.y] then
              light := val[4];
          end;
          if (val[7] < gZBuffer^[i, j]) then begin
            gZBuffer^[i, j] := val[7];
            gParamBuffer^[i, j, 1] := val[2];
            gParamBuffer^[i, j, 2] := val[3];
            light := light + val[1];
            pm := floor(light);
            if theMap <> nil then
              theMap^[i, j] := pm;
            line^[i] := gPalette3D^[pm];
          end
        end;
      end;
    end;
  end;
end;

procedure FillTriangleWithClips(bm: TBitmap; vertices: triang; triangleF:
  triangFoc);
var
{x,y Drawing3D coordinates; u,v parameters of the point}
{p,q coordinates in the virtual screen used for the Focus}
  permu: threeBytes;
  virt: TPoint;
  A, B, C: ptDataAndParam;
  AE, BE, CE: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy, light: single;
  dfp, dfq, dfExtrapol: single;
  val, initialVal: array[1..8] of single;
  derivs: array[1..8, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, pm, k: Integer;
  auxab, auxac, auxbc: double;
  line: pTColorArray;
  noFBorder, haveDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    derivs[1, 1] := A.c.ambLight * ax + B.c.ambLight * bx + C.c.ambLight * cx;
    derivs[1, 2] := A.c.ambLight * ay + B.c.ambLight * by + C.c.ambLight * cy;
    derivs[2, 1] := A.u * ax + B.u * bx + C.u * cx;
    derivs[2, 2] := A.u * ay + B.u * by + C.u * cy;
    derivs[3, 1] := A.v * ax + B.v * bx + C.v * cx;
    derivs[3, 2] := A.v * ay + B.v * by + C.v * cy;
    derivs[4, 1] := AE.focLight * ax + BE.focLight * bx + CE.focLight * cx;
    derivs[4, 2] := AE.focLight * ay + BE.focLight * by + CE.focLight * cy;
    derivs[5, 1] := AE.pt[1] * ax + BE.pt[1] * bx +
      CE.pt[1] * cx;
    derivs[5, 2] := AE.pt[1] * ay + BE.pt[1] * by +
      CE.pt[1] * cy;
    derivs[6, 1] := AE.pt[2] * ax + BE.pt[2] * bx +
      CE.pt[2] * cx;
    derivs[6, 2] := AE.pt[2] * ay + BE.pt[2] * by +
      CE.pt[2] * cy;
    derivs[7, 1] := A.c.proj[3] * ax + B.c.proj[3] * bx + C.c.proj[3] * cx;
    derivs[7, 2] := A.c.proj[3] * ay + B.c.proj[3] * by + C.c.proj[3] * cy;
    derivs[8, 1] := AE.pt[3] * ax + BE.pt[3] * bx +
      CE.pt[3] * cx;
    derivs[8, 2] := AE.pt[3] * ay + BE.pt[3] * by +
      CE.pt[3] * cy;
    det := derivs[6, 2] * derivs[5, 1] - derivs[6, 1] * derivs[5, 2];
    noFBorder := abs(det) > 1.0E-7;
    if noFBorder then begin
      dfp := (derivs[7, 1] * derivs[6, 2] - derivs[7, 2] * derivs[6, 1]) /
        det;
      dfq := (derivs[7, 2] * derivs[5, 1] - derivs[7, 1] * derivs[5, 2]) /
        det;
    end;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := cInitVal[1, k] * ax + cInitVal[2, k] * bx +
        cInitVal[3, k] * cx;
      cValDerivs[k, 2] := cInitVal[1, k] * ay + cInitVal[2, k] * by +
        cInitVal[3, k] * cy;
    end;
    haveDerivs := True;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    for ii := 1 to 8 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
    for ii := 0 to gClipTop do
      cVal[ii] := cInitVal[1, ii] + cValDerivs[ii, 1] * difi + cValDerivs[ii,
        2]
        *
        difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 8 do
      val[ii] := val[ii] + derivs[ii, 1];
    for ii := 0 to gClipTop do
      cVal[ii] := cVal[ii] + cValDerivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].c.proj[2] < vertices[1].c.proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].c.proj[2] < vertices[permu[1]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].c.proj[2] < vertices[permu[2]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  AE := triangleF[permu[1]];
  BE := triangleF[permu[2]];
  CE := triangleF[permu[3]];

  ACx := A.c.proj[1] - C.c.proj[1];
  ACy := A.c.proj[2] - C.c.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.c.proj[1] - B.c.proj[1];
  ABy := A.c.proj[2] - B.c.proj[2];
  BCx := B.c.proj[1] - C.c.proj[1];
  BCy := B.c.proj[2] - C.c.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;

  auxab := A.c.ambLight + B.c.ambLight + C.c.ambLight;
  if auxab < 0 then begin
    A.c.ambLight := abs(A.c.ambLight) + 101;
    B.c.ambLight := abs(B.c.ambLight) + 101;
    C.c.ambLight := abs(C.c.ambLight) + 101;
  end
  else begin
    A.c.ambLight := abs(A.c.ambLight);
    B.c.ambLight := abs(B.c.ambLight);
    C.c.ambLight := abs(C.c.ambLight);
  end;

  coefAC := ACx / ACy;
  initialVal[1] := A.c.ambLight;
  initialVal[2] := A.u;
  initialVal[3] := A.v;
  initialVal[4] := AE.FocLight;
  initialVal[5] := AE.pt[1];
  initialVal[6] := AE.pt[2];
  initialVal[7] := A.c.proj[3];
  initialVal[8] := AE.pt[3];
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      cInitVal[i, j] := HeightOverClipS(gClips[j], vertices[permu[i]].c.pos);
  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.c.proj[2]));
    jSup := min(kHeightMinus1, floor(B.c.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.c.proj[2];
      auxac := A.c.proj[1] + difj * coefAC;
      auxab := A.c.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        line := bm.ScanLine[j];
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), kWidthMinus1);
        end;

        if here <= there then begin
          difi := here - A.c.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);
            if good then begin
              virt.x := Floor(val[5]);
              virt.y := Floor(val[6]);
              light := 0;
              if (virt.x >= 0) and (virt.x < kWidth) and (virt.y
                >= 0) and (virt.y < kHeight) and noFBorder then begin
                dfExtrapol := val[8] + dfp * (virt.x -
                  val[5]) + dfq * (virt.y - val[6]);
                if dfExtrapol < 1.002 * gfBuffer^[virt.x, virt.y]
                  then
                  light := val[4];
              end;
              if (val[7] < gZBuffer^[i, j]) then begin
                gZBuffer^[i, j] := val[7];
                gParamBuffer^[i, j, 1] := val[2];
                gParamBuffer^[i, j, 2] := val[3];
                light := light + val[1];
                pm := floor(light);
                if theMap <> nil then
                  theMap^[i, j] := pm;
                line^[i] := gPalette3D^[pm];
              end;
            end;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.c.proj[2]));
    jSup := min(kHeightMinus1, floor(C.c.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.c.proj[2];
    difb := jInf - 1 - B.c.proj[2];
    auxac := A.c.proj[1] + difj * coefAC;
    auxbc := B.c.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      line := bm.ScanLine[j];
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), kWidthMinus1);
      end;
      if here <= there then begin
        difi := here - A.c.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good then begin
            virt.x := Floor(val[5]);
            virt.y := Floor(val[6]);
            light := 0;
            if (virt.x >= 0) and (virt.x < kWidth) and (virt.y
              >= 0) and (virt.y < kHeight) and noFBorder then begin
              dfExtrapol := val[8] + dfp * (virt.x -
                val[5]) + dfq * (virt.y - val[6]);
              if dfExtrapol < 1.002 * gfBuffer^[virt.x, virt.y] then
                light := val[4];
            end;
            if (val[7] < gZBuffer^[i, j]) then begin
              gZBuffer^[i, j] := val[7];
              gParamBuffer^[i, j, 1] := val[2];
              gParamBuffer^[i, j, 2] := val[3];
              light := light + val[1];
              pm := floor(light);
              if theMap <> nil then
                theMap^[i, j] := pm;
              line^[i] := gPalette3D^[pm];
            end;
          end;
        end;
      end;
    end;
  end;
end;

procedure FillZBufferTriangle(vertices: triang);
var
  permu: threeBytes;
  A, B, C: ptDataAndParam;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  val, initialVal: array[1..3] of single;
  derivs: array[1..3, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there: Integer;
  auxab, auxac, auxbc: double;
  haveDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    derivs[1, 1] := A.c.proj[3] * ax + B.c.proj[3] * bx + C.c.proj[3] * cx;
    derivs[1, 2] := A.c.proj[3] * ay + B.c.proj[3] * by + C.c.proj[3] * cy;
    derivs[2, 1] := A.u * ax + B.u * bx + C.u * cx;
    derivs[2, 2] := A.u * ay + B.u * by + C.u * cy;
    derivs[3, 1] := A.v * ax + B.v * bx + C.v * cx;
    derivs[3, 2] := A.v * ay + B.v * by + C.v * cy;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    for ii := 1 to 3 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 3 do
      val[ii] := val[ii] + derivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].c.proj[2] < vertices[1].c.proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].c.proj[2] < vertices[permu[1]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].c.proj[2] < vertices[permu[2]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  ACx := A.c.proj[1] - C.c.proj[1];
  ACy := A.c.proj[2] - C.c.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.c.proj[1] - B.c.proj[1];
  ABy := A.c.proj[2] - B.c.proj[2];
  BCx := B.c.proj[1] - C.c.proj[1];
  BCy := B.c.proj[2] - C.c.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;


  coefAC := ACx / ACy;
  initialVal[1] := A.c.proj[3];
  initialVal[2] := A.u;
  initialVal[3] := A.v;

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.c.proj[2]));
    jSup := min(kHeightMinus1, floor(B.c.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.c.proj[2];
      auxac := A.c.proj[1] + difj * coefAC;
      auxab := A.c.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), kWidthMinus1);
        end;

        if here <= there then begin
          difi := here - A.c.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            if (val[1] < gZBuffer^[i, j]) then begin
              gZBuffer^[i, j] := val[1];
              gParamBuffer^[i, j, 1] := val[2];
              gParamBuffer^[i, j, 2] := val[3];
            end;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.c.proj[2]));
    jSup := min(kHeightMinus1, floor(C.c.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.c.proj[2];
    difb := jInf - 1 - B.c.proj[2];
    auxac := A.c.proj[1] + difj * coefAC;
    auxbc := B.c.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), kWidthMinus1);
      end;
      if here <= there then begin
        difi := here - A.c.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          if (val[1] < gZBuffer^[i, j]) then begin
            gZBuffer^[i, j] := val[1];
            gParamBuffer^[i, j, 1] := val[2];
            gParamBuffer^[i, j, 2] := val[3];
          end
        end;
      end;
    end;
  end;
end;

procedure ComputeZBuffer;
var
  i, j: integer;
  points: triang;
begin
  if (gPointsData = nil) or (high(gPointsData[0]) <> gGridV) or
    (high(gPointsData) <> gGridU) then begin
    gPointsData := nil;
    SetLength(gPointsData, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(gPointsData[i], gGridV + 1);
  end;

  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  InitializeZBuffer;
  InitializeParamBuffer;
  ComputeGridData(False);

  for i := 1 to gGridU do
    for j := 1 to gGridV do begin
      points[1].c := gPointsData[i, j];
      points[1].u := i;
      points[1].v := j;
      points[2].c := gPointsData[i - 1, j - 1];
      points[2].u := i - 1;
      points[2].v := j - 1;
      points[3].c := gPointsData[i - 1, j];
      points[3].u := i - 1;
      points[3].v := j;
      FillZBufferTriangle(points);
      points[3].c := gPointsData[i, j - 1];
      points[3].u := i;
      points[3].v := j - 1;
      FillZBufferTriangle(points);
    end;
end;

procedure FillZBufferTriangleWithClips(vertices: triang);
var
  permu: threeBytes;
  A, B, C: ptDataAndParam;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  val, initialVal: array[1..3] of single;
  derivs: array[1..3, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, k: Integer;
  auxab, auxac, auxbc: double;
  haveDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    derivs[1, 1] := A.c.proj[3] * ax + B.c.proj[3] * bx + C.c.proj[3] * cx;
    derivs[1, 2] := A.c.proj[3] * ay + B.c.proj[3] * by + C.c.proj[3] * cy;
    derivs[2, 1] := A.u * ax + B.u * bx + C.u * cx;
    derivs[2, 2] := A.u * ay + B.u * by + C.u * cy;
    derivs[3, 1] := A.v * ax + B.v * bx + C.v * cx;
    derivs[3, 2] := A.v * ay + B.v * by + C.v * cy;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := cInitVal[1, k] * ax + cInitVal[2, k] * bx +
        cInitVal[3, k] * cx;
      cValDerivs[k, 2] := cInitVal[1, k] * ay + cInitVal[2, k] * by +
        cInitVal[3, k] * cy;
    end;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    for ii := 1 to 3 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
    for ii := 0 to gClipTop do
      cVal[ii] := cInitVal[1, ii] + cValDerivs[ii, 1] * difi + cValDerivs[ii,
        2]
        *
        difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 3 do
      val[ii] := val[ii] + derivs[ii, 1];
    for ii := 0 to gClipTop do
      cVal[ii] := cVal[ii] + cValDerivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].c.proj[2] < vertices[1].c.proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].c.proj[2] < vertices[permu[1]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].c.proj[2] < vertices[permu[2]].c.proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  ACx := A.c.proj[1] - C.c.proj[1];
  ACy := A.c.proj[2] - C.c.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.c.proj[1] - B.c.proj[1];
  ABy := A.c.proj[2] - B.c.proj[2];
  BCx := B.c.proj[1] - C.c.proj[1];
  BCy := B.c.proj[2] - C.c.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;


  coefAC := ACx / ACy;
  initialVal[1] := A.c.proj[3];
  initialVal[2] := A.u;
  initialVal[3] := A.v;
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      cInitVal[i, j] := HeightOverClipS(gClips[j], vertices[permu[i]].c.pos);

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.c.proj[2]));
    jSup := min(kHeightMinus1, floor(B.c.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.c.proj[2];
      auxac := A.c.proj[1] + difj * coefAC;
      auxab := A.c.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), kWidthMinus1);
        end;

        if here <= there then begin
          difi := here - A.c.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);
            if good and (val[1] < gZBuffer^[i, j]) then begin
              gZBuffer^[i, j] := val[1];
              gParamBuffer^[i, j, 1] := val[2];
              gParamBuffer^[i, j, 2] := val[3];
            end;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.c.proj[2]));
    jSup := min(kHeightMinus1, floor(C.c.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.c.proj[2];
    difb := jInf - 1 - B.c.proj[2];
    auxac := A.c.proj[1] + difj * coefAC;
    auxbc := B.c.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), kWidthMinus1);
      end;
      if here <= there then begin
        difi := here - A.c.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good and (val[1] < gZBuffer^[i, j]) then begin
            gZBuffer^[i, j] := val[1];
            gParamBuffer^[i, j, 1] := val[2];
            gParamBuffer^[i, j, 2] := val[3];
          end
        end;
      end;
    end;
  end;
end;

procedure ComputeZBufferWithClips;
var
  i, j: integer;
  points: triang;
begin
  if (gPointsData = nil) or (high(gPointsData[0]) <> gGridV) or
    (high(gPointsData) <> gGridU) then begin
    gPointsData := nil;
    SetLength(gPointsData, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(gPointsData[i], gGridV + 1);
  end;
  gClipTop := Length(gClips) - 1;
  for i := 1 to 3 do begin
    cInitVal[i] := nil;
    SetLength(cInitVal[i], gClipTop + 1);
  end;
  cVal := nil;
  SetLength(cVal, gClipTop + 1);
  cValDerivs := nil;
  SetLength(cValDerivs, gClipTop + 1);
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  InitializeZBuffer;
  InitializeParamBuffer;
  ComputeGridData(False);
  for i := 1 to gGridU do
    for j := 1 to gGridV do begin
      points[1].c := gPointsData[i, j];
      points[1].u := i;
      points[1].v := j;
      points[2].c := gPointsData[i - 1, j - 1];
      points[2].u := i - 1;
      points[2].v := j - 1;
      points[3].c := gPointsData[i - 1, j];
      points[3].u := i - 1;
      points[3].v := j;
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and gPointsData[i - 1, j].inside then
        FillZBufferTriangle(points)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i - 1, j].outside) then
        FillZBufferTriangleWithClips(points);
      points[3].c := gPointsData[i, j - 1];
      points[3].u := i;
      points[3].v := j - 1;
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and gPointsData[i, j - 1].inside then
        FillZBufferTriangle(points)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i, j - 1].outside) then
        FillZBufferTriangleWithClips(points);
    end;
end;

procedure FillFBufferTriangle(triangleF:
  triangFoc);
var
  permu: threeBytes;
  A, B, C: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  df, initialDF, dfp, dfq: single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there: Integer;
  auxab, auxac, auxbc: double;
  haveDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    dfp := A.pt[3] * ax + B.pt[3] * bx + C.pt[3]
      * cx;
    dfq := A.pt[3] * ay + B.pt[3] * by + C.pt[3]
      * cy;
    haveDerivs := True;
  end;

  procedure ComputeData;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    df := initialDF + dfp * difi +
      dfq * difj;
  end;


begin
  haveDerivs := False;
  permu := gId;
  if triangleF[2].pt[2] < triangleF[1].pt[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triangleF[3].pt[2] < triangleF[permu[1]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triangleF[3].pt[2] < triangleF[permu[2]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triangleF[permu[1]];
  B := triangleF[permu[2]];
  C := triangleF[permu[3]];

  ACx := A.pt[1] - C.pt[1];
  ACy := A.pt[2] - C.pt[2];
  if ACy >= 0 then
    Exit;
  ABx := A.pt[1] - B.pt[1];
  ABy := A.pt[2] - B.pt[2];
  BCx := B.pt[1] - C.pt[1];
  BCy := B.pt[2] - C.pt[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;
  coefAC := ACx / ACy;
  initialDF := A.pt[3];
  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.pt[2]));
    jSup := min(kHeightMinus1, floor(B.pt[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.pt[2];
      auxac := A.pt[1] + difj * coefAC;
      auxab := A.pt[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), kWidthMinus1);
        end;

        if here <= there then begin
          difi := here - A.pt[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            df := df + dfp;
            if df < gfBuffer^[i, j] then
              gfBuffer^[i, j] := df;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.pt[2]));
    jSup := min(kHeightMinus1, floor(C.pt[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.pt[2];
    difb := jInf - 1 - B.pt[2];
    auxac := A.pt[1] + difj * coefAC;
    auxbc := B.pt[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), kWidthMinus1);
      end;
      if here <= there then begin
        difi := here - A.pt[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          df := df + dfp;
          if df < gfBuffer^[i, j] then
            gfBuffer^[i, j] := df;
        end;
      end;
    end;
  end;
end;


procedure FillFBufferTriangleWithClips(triangleF:
  triangFoc; poss: Matrix3Single);
var
  permu: threeBytes;
  A, B, C: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  df, initialDF, dfp, dfq: single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, k: Integer;
  auxab, auxac, auxbc: double;
  noFBorder, haveDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    dfp := A.pt[3] * ax + B.pt[3] * bx + C.pt[3]
      * cx;
    dfq := A.pt[3] * ay + B.pt[3] * by + C.pt[3]
      * cy;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := cInitVal[1, k] * ax + cInitVal[2, k] * bx +
        cInitVal[3, k] * cx;
      cValDerivs[k, 2] := cInitVal[1, k] * ay + cInitVal[2, k] * by +
        cInitVal[3, k] * cy;
    end;
    haveDerivs := True;
  end;

  procedure ComputeData;
  var
    k: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    df := initialDF + dfp * difi +
      dfq * difj;
    for k := 0 to gClipTop do
      cVal[k] := cInitVal[1, k] + cValDerivs[k, 1] * difi + cValDerivs[k, 2] *
        difj;
  end;

  procedure IncrementI;
  var
    k: Integer;
  begin
    df := df + dfp;
    for k := 0 to gClipTop do
      cVal[k] := cVal[k] + cValDerivs[k, 1];
  end;


begin
  haveDerivs := False;
  permu := gId;
  if triangleF[2].pt[2] < triangleF[1].pt[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triangleF[3].pt[2] < triangleF[permu[1]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triangleF[3].pt[2] < triangleF[permu[2]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triangleF[permu[1]];
  B := triangleF[permu[2]];
  C := triangleF[permu[3]];
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      cInitVal[i, j] := HeightOverClipS(gClips[j], poss[permu[i]]);
  ACx := A.pt[1] - C.pt[1];
  ACy := A.pt[2] - C.pt[2];
  if ACy >= 0 then
    Exit;
  ABx := A.pt[1] - B.pt[1];
  ABy := A.pt[2] - B.pt[2];
  BCx := B.pt[1] - C.pt[1];
  BCy := B.pt[2] - C.pt[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;
  coefAC := ACx / ACy;
  initialDF := A.pt[3];
  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.pt[2]));
    jSup := min(kHeightMinus1, floor(B.pt[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.pt[2];
      auxac := A.pt[1] + difj * coefAC;
      auxab := A.pt[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), kWidthMinus1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), kWidthMinus1);
        end;

        if here <= there then begin
          difi := here - A.pt[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);
            if good and (df < gfBuffer^[i, j]) then
              gfBuffer^[i, j] := df;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.pt[2]));
    jSup := min(kHeightMinus1, floor(C.pt[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.pt[2];
    difb := jInf - 1 - B.pt[2];
    auxac := A.pt[1] + difj * coefAC;
    auxbc := B.pt[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), kWidthMinus1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), kWidthMinus1);
      end;
      if here <= there then begin
        difi := here - A.pt[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good and (df < gfBuffer^[i, j]) then
            gfBuffer^[i, j] := df;
        end;
      end;
    end;
  end;
end;

procedure ComputeFBufferWithClips;
var
  i, j: integer;
  triangleF: triangFoc;
  poss: Matrix3Single;
begin
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  for i := 1 to gGridU do
    for j := 1 to gGridV do begin
      triangleF[1] := gFocusPtData[i, j];
      triangleF[2] := gFocusPtData[i - 1, j - 1];
      triangleF[3] := gFocusPtData[i - 1, j];
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and
        gPointsData[i - 1, j].inside then
        FillFBufferTriangle(triangleF)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i - 1, j].outside) then begin
        poss[1] := gPointsData[i, j].pos;
        poss[2] := gPointsData[i - 1, j - 1].pos;
        poss[3] := gPointsData[i - 1, j].pos;
        FillFBufferTriangleWithClips(triangleF, poss);
      end;
      triangleF[3] := gFocusPtData[i, j - 1];
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and
        gPointsData[i, j - 1].inside then
        FillFBufferTriangle(triangleF)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i, j - 1].outside) then begin
        poss[1] := gPointsData[i, j].pos;
        poss[2] := gPointsData[i - 1, j - 1].pos;
        poss[3] := gPointsData[i, j - 1].pos;
        FillFBufferTriangleWithClips(triangleF, poss);
      end;
    end;
end;

procedure ComputeFBuffer;
var
  i, j: integer;
  triangleF: triangFoc;
begin
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  for i := 1 to gGridU do
    for j := 1 to gGridV do begin
      triangleF[1] := gFocusPtData[i, j];
      triangleF[2] := gFocusPtData[i - 1, j - 1];
      triangleF[3] := gFocusPtData[i - 1, j];
      FillFBufferTriangle(triangleF);
      triangleF[3] := gFocusPtData[i, j - 1];
      FillFBufferTriangle(triangleF);
    end;
end;

function DrawSurfaceWithoutClips(bm: TBitmap): Boolean;
var
  i, j: Integer;
  vertices: triang;
  triangleF: triangFoc;
begin
  Result := False;
  if (gPointsData = nil) or (high(gPointsData[0]) <> gGridV) or
    (high(gPointsData) <> gGridU) then begin
    gPointsData := nil;
    SetLength(gPointsData, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(gPointsData[i], gGridV + 1);
  end;
  if (gFocusPtData = nil) or (high(gFocusPtData[0]) <> gGridV) or
    (high(gFocusPtData) <> gGridU) then begin
    gFocusPtData := nil;
    SetLength(gFocusPtData, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(gFocusPtData[i], gGridV + 1);
  end;
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  InitializeZBuffer;
  InitializeParamBuffer;
  InitializeFBuffer;
  ComputeGridData(True);
  ComputeFBuffer;
  for i := 1 to gGridU do begin
    if i mod 20 = 0 then begin
      Application.processMessages;
      if gStop then Exit;
    end;
    for j := 1 to gGridV do begin
      vertices[1].c := gPointsData[i, j];
      triangleF[1] := gFocusPtData[i, j];
      vertices[1].u := i;
      vertices[1].v := j;
      vertices[2].c := gPointsData[i - 1, j - 1];
      triangleF[2] := gFocusPtData[i - 1, j - 1];
      vertices[2].u := i - 1;
      vertices[2].v := j - 1;
      vertices[3].c := gPointsData[i - 1, j];
      triangleF[3] := gFocusPtData[i - 1, j];
      vertices[3].u := i - 1;
      vertices[3].v := j;
      FillTriangle(bm, vertices, triangleF);
      vertices[3].c := gPointsData[i, j - 1];
      triangleF[3] := gFocusPtData[i, j - 1];
      vertices[3].u := i;
      vertices[3].v := j - 1;
      FillTriangle(bm, vertices, triangleF);
    end;
  end;
  Result := True;
end;

function DrawSurfaceWithClips(bm: TBitmap): Boolean;
var
  i, j: Integer;
  vertices: triang;
  triangleF: triangFoc;
begin
  Result := False;
  if (gPointsData = nil) or (high(gPointsData[0]) <> gGridV) or
    (high(gPointsData) <> gGridU) then begin
    gPointsData := nil;
    SetLength(gPointsData, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(gPointsData[i], gGridV + 1);
  end;
  if (gFocusPtData = nil) or (high(gFocusPtData[0]) <> gGridV) or
    (high(gFocusPtData) <> gGridU) then begin
    gFocusPtData := nil;
    SetLength(gFocusPtData, gGridU + 1);
    for i := 0 to gGridU do
      SetLength(gFocusPtData[i], gGridV + 1);
  end;
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  gClipTop := Length(gClips) - 1;
  for i := 1 to 3 do begin
    cInitVal[i] := nil;
    SetLength(cInitVal[i], gClipTop + 1);
  end;
  cVal := nil;
  SetLength(cVal, gClipTop + 1);
  cValDerivs := nil;
  SetLength(cValDerivs, gClipTop + 1);

  InitializeZBuffer;
  InitializeParamBuffer;
  InitializeFBuffer;
  ComputeGridData(True);
  ComputeFBufferWithClips;
  for i := 1 to gGridU do begin
    if i mod 20 = 0 then begin
      Application.processMessages;
      if gStop then Exit;
    end;
    for j := 1 to gGridV do begin
      vertices[1].c := gPointsData[i, j];
      triangleF[1] := gFocusPtData[i, j];
      vertices[1].u := i;
      vertices[1].v := j;
      vertices[2].c := gPointsData[i - 1, j - 1];
      triangleF[2] := gFocusPtData[i - 1, j - 1];
      vertices[2].u := i - 1;
      vertices[2].v := j - 1;
      vertices[3].c := gPointsData[i - 1, j];
      triangleF[3] := gFocusPtData[i - 1, j];
      vertices[3].u := i - 1;
      vertices[3].v := j;
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and gPointsData[i - 1, j].inside then
        FillTriangle(bm, vertices, triangleF)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i - 1, j].outside) then
        FillTriangleWithClips(bm, vertices, triangleF);
      vertices[3].c := gPointsData[i, j - 1];
      triangleF[3] := gFocusPtData[i, j - 1];
      vertices[3].u := i;
      vertices[3].v := j - 1;
      if gPointsData[i, j].inside and gPointsData[i - 1, j - 1].inside
        and gPointsData[i, j - 1].inside then
        FillTriangle(bm, vertices, triangleF)
      else if not (gPointsData[i, j].outside or gPointsData[i - 1, j -
        1].outside or gPointsData[i, j - 1].outside) then
        FillTriangleWithClips(bm, vertices, triangleF);
    end;
  end;

  Result := True;
end;

procedure ComputeParamSurface(bm: TBitmap; var map: ColorMapPtr);
begin
  if (map <> nil) and (map <> gNormalPairMap) then
    FillChar(map^, kWidth * kHeight, #$00);
  theMap := map;
  if gClips <> nil then
    DrawSurfaceWithClips(bm)
  else
    DrawSurfaceWithoutClips(bm)
end;

procedure ComputeNormalPairParam;
const
  sinEyes: double = 1 / 15;
  cosEyes: double = 0.9977753031;
var
  nLeft, hLeft, vLeft: Vector3;
  j, k: integer;
  d: double;
  prevC0, prevC1, prevC2, prevC3, prevPcx: double;
  cen, prevH, prevV, prevN, prevWp: Vector3;
  line: pTColorArray;
begin
  FillChar(gNormalPairMap^, kWidth * kHeight, 0);
  prevC0 := gC0;
  gC0 := prevC0 / 2;
  prevPcx := gpcx;
  gpcx := round(prevPcx / 2);
  cen := gCenter3D;
  d := EuclDistance3(gViewpoint, cen);
  vLeft := gV;
  for j := 1 to 3 do begin
    nLeft[j] := cosEyes * gN[j] + sinEyes * gH[j];
    hLeft[j] := -sinEyes * gN[j] + cosEyes * gH[j];
  end;
  gMovingBmp.canvas.FillRect(gRect);
  ComputeParamSurface(gMovingBmp, gNormalPairMap);

  prevC1 := gC1;
  prevC2 := gC2;
  prevC3 := gC3;
  gC3 := EuclInnerProd3(nLeft, cen) - d;
  gC1 := EuclInnerProd3(hLeft, cen);
  gC2 := EuclInnerProd3(vLeft, cen);
  prevWp := gViewpoint;
  for j := 1 to 3 do
    gViewpoint[j] := gC3 * nLeft[j] + gC1 * hLeft[j] + gC2 * vLeft[j];
  gpcx := gpcx + kWidth div 2;
  prevH := gH;
  prevN := gN;
  prevV := gV;
  gH := hLeft;
  gV := vLeft;
  gN := nLeft;
  ComputeParamSurface(gMovingBmp, gNormalPairMap);

  gc0 := prevC0;
  gC1 := prevC1;
  gC2 := prevC2;
  gC3 := prevC3;
  gpcx := Round(prevPcx);
  gH := prevH;
  gV := prevV;
  gN := prevN;
  gViewpoint := prevWp;

  ComputeZBuffer;
  Antialiase(gNormalPairMap);
  for k := 0 to kHeightMinus1 do begin
    line := gMovingBmp.ScanLine[k];
    for j := 0 to kWidthMinus1 do
      line^[j] := gPalette3D^[gNormalPairMap^[j, k]];
  end;
end; {de ComputeNormalPairParam}

function DrawParametricSurface: Boolean;
begin
  result := False;
  if gType = Tube then begin
    surface.uMaxEdit.text := FloatToStrF(gMax2D[1], ffFixed, 5, 2);
    surface.uMinEdit.text := FloatToStrF(gMin2D[1], ffFixed, 5, 2);
    if gEdges2D[1] <= 1.0E-10 then begin
      surface.Inform(eBadTube);
      Exit;
    end;
  end;
  SetFocus;
  FrameFromFocus;
  ComputeParamSurface(gStdBmp, gMap);
  Antialiase(gMap);
  ReceiveColorMap(gStdBmp, gMap);
  ComputeVPCubeKey;
  ComputeCube;
  PrepareToSetClip;
  result := True;
end;

procedure Finalize;
var
  i: Integer;
begin
  for i := 1 to 3 do
    cInitVal[i] := nil;
  cVal := nil;
  cValDerivs := nil;
end;

initialization
  ;
finalization
  Finalize;

end.
