

{-----------------------------------------------------------------------------
 Unit Name: FSelectToPrint
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: by means of TForm2, the user can choose the contents and resolution
    for the printing of a surface
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}



unit FSelectToPrint;

interface

uses
  SysUtils, Classes,
{$IFDEF MSWINDOWS}
  Windows, Messages, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ActnList, StdActns, ExtCtrls
{$ENDIF}
{$IFDEF LINUX}
  Types, QGraphics, QControls, QForms, QDialogs,
  QStdCtrls, QActnList, QStdActns, QExtCtrls
{$ENDIF}
   ;
type
  TForm2 = class(TForm)
    CoordinatesCheck: TCheckBox;
    Label1: TLabel;
    CurvaturesCheck: TCheckBox;
    AsymptoticsCheck: TCheckBox;
    IntrinsicCheck: TCheckBox;
    Label2: TLabel;
    CancelBt: TButton;
    OKBt: TButton;
    GeodBallsCheck: TCheckBox;
    PrintedPageContents: TRadioGroup;
    LightRaysCheck: TCheckBox;
    ParalTransportCheck: TCheckBox;
    Resolution: TRadioGroup;
    Label3: TLabel;
    procedure OnClose(Sender: TObject; var Action: TCloseAction);
    procedure OnShowForm(Sender: TObject);
    procedure OnOK(Sender: TObject);
    procedure OnCancel(Sender: TObject);
    procedure OnPrintedPageContents(Sender: TObject);

  private
    { Private declarations }
  public
  { Public declarations }
  end;

var
  Form2: TForm2;

implementation
{$IFDEF MSWINDOWS}
  {$R *.DFM}
{$ENDIF}
{$IFDEF LINUX}
  {$R *.xfm}
{$ENDIF}
uses
  Globals;



procedure TForm2.OnClose(Sender: TObject; var Action: TCloseAction);
begin
  if (modalResult <> mrOK) then
    ModalResult := mrCancel;
end;

procedure TForm2.OnShowForm(Sender: TObject);
begin
  visible := True;
  if gType = PseudoRiemannian then begin
    PrintedPageContents.Items[0] := 'Print Background only';
    PrintedPageContents.Items[1] := 'Background and Lines';
    PrintedPageContents.Controls[2].Enabled := False;
    CoordinatesCheck.Enabled := False;
    CurvaturesCheck.Enabled := False;
    AsymptoticsCheck.Caption := 'Lines of Const. Curv.';
    IntrinsicCheck.caption := 'Geodesics';
  end
  else begin
    PrintedPageContents.Controls[1].Enabled := (gMode = Standard);
    PrintedPageContents.Controls[2].Enabled := (gMode = Standard) and (gType <> Implicit);
  end;
  OnPrintedPageContents(nil);
  resolution.Enabled := not (gMode in gWireModes);
end;

procedure TForm2.OnOK(Sender: TObject);
var
  count, i: Integer;
{$IFDEF MSWINDOWS}
  answer: Integer;
{$ENDIF}
{$IFDEF LINUX}
  answer: TMessageButton;
{$ENDIF}
begin
  if gType <> PseudoRiemannian then begin
    gLinesData[kCoordinates].toPrint := CoordinatesCheck.checked;
    gLinesData[kCurvature].toPrint := CurvaturesCheck.checked;
    gLinesData[kAsymptotics].toPrint := AsymptoticsCheck.checked;
    gLinesData[kIntrinsic].toPrint := IntrinsicCheck.checked;
    gLinesData[kGeodBall].toPrint := GeodBallsCheck.checked;
    gLinesData[kLightRays].toPrint := LightRaysCheck.checked;
    gLinesData[kTransport].toPrint := ParalTransportCheck.checked;
    gLinesData[kArrows].toPrint := ParalTransportCheck.checked;
  end
  else begin
    gLinesData[kConstantCurvatureLorentz].toPrint := AsymptoticsCheck.checked;
    gLinesData[kConstantCurvatureRiem].toPrint := AsymptoticsCheck.checked;
    gLinesData[kIntrinsicLorentz].toPrint := IntrinsicCheck.checked;
    gLinesData[kIntrinsicRiem].toPrint := IntrinsicCheck.checked;
    gLinesData[kGeodBall].toPrint := GeodBallsCheck.checked;
    gLinesData[kLightRays].toPrint := LightRaysCheck.checked;
    gLinesData[kTransport].toPrint := ParalTransportCheck.checked;
    gLinesData[kArrows].toPrint := ParalTransportCheck.checked;
  end;
  count := 0;
  gPrintLines := PrintedPageContents.ItemIndex > 0;
  gPrintSurface := PrintedPageContents.ItemIndex < 2;
  for i := kCoordinates to kTransport do
    count := count + ord(gLinesData[i].toPrint);
  gResolution := resolution.ItemIndex + 1;
  if gPrintSurface or (count > 0) then
    ModalResult := mrOK
  else begin
   {$IFDEF MSWINDOWS}
    answer := Application.MessageBox('You did select nothing to print!',
      'Warning', MB_OKCANCEL + MB_DEFBUTTON1);
    if answer <> IDCancel then
      ModalResult := mrCancel
    else
      modalResult := 0;
{$ENDIF}
{$IFDEF LINUX}
    answer := Application.MessageBox('You did select nothing to print! ',
      'Warning', [smbOK , smbCancel],smsWarning, smbOK, smbCancel);
    if answer <> smbCancel then
      ModalResult := mrCancel
    else
      modalResult := 0;
{$ENDIF}
  end;
end;

procedure TForm2.OnCancel(Sender: TObject);
begin
  ModalResult := mrCancel;
end;



procedure TForm2.OnPrintedPageContents(Sender: TObject);
begin
  if PrintedPageContents.ItemIndex = 0 then begin
    CoordinatesCheck.Enabled := False;
    CurvaturesCheck.Enabled := False;
    AsymptoticsCheck.Enabled := False;
    IntrinsicCheck.Enabled := False;
    GeodBallsCheck.Enabled := False;
    LightRaysCheck.Enabled := False;
    ParalTransportCheck.Enabled := False;
    Label1.Enabled := False;
    Label2.Enabled := False;
  end
  else begin
    CoordinatesCheck.Enabled := (gMode = Standard);
    Resolution.Enabled := (PrintedPageContents.ItemIndex <> 2);


    CurvaturesCheck.Enabled := (gMode = Standard);
    AsymptoticsCheck.Enabled := (gMode = Standard);
    IntrinsicCheck.Enabled := (gMode = Standard);
    GeodBallsCheck.Enabled := (gMode = Standard);
    LightRaysCheck.Enabled := (gMode = Standard);
    ParalTransportCheck.Enabled := (gMode = Standard);
    Label1.Enabled := (gMode = Standard);
    Label2.Enabled := (gMode = Standard);
    CoordinatesCheck.Enabled := (gMode = Standard) and (gType <> PseudoRiemannian);
    CurvaturesCheck.Enabled := (gMode = Standard) and (gType <> PseudoRiemannian);
    ParalTransportCheck.Checked := (gMode = Standard) and (gType <> Implicit);
    ParalTransportCheck.Enabled := (gMode = Standard) and (gType <> Implicit);
  end;
end;


end.
