

{-----------------------------------------------------------------------------
 Unit Name: FileInOut
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: routines for reading and writing to disk
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}

unit FileInOut;

interface

uses
  SysUtils;

function ReadFromDisk(pathAndName: TFilename): Boolean;
function WriteToDisk(pathAndName: TFilename): Boolean;

implementation

uses
  Math,
{$IFDEF MSWINDOWS}
  Dialogs,
{$ENDIF}
{$IFDEF LINUX}
  QDialogs, 
{$ENDIF}
  Globals, ResStrings;

{******************  For compatibility with old versions ******************}

{$ALIGN 2} {For agreement with previous versions that used this alignment
            {in records. This will affect only to these old types.}

const
  kMaxActors = 500;
  kMaxData = 47;
type
  vector2Extended = array[1..2] of extended;
  vectorExtended = array[1..3] of extended;
  dataExt = record
    step2D, lados2D, semilados2D, centro2D: vector2Extended;
    numLineas, numFlags, ppLine2D: Integer;
    amplitud2D, amplitud3D, radio3D, c0, c1, c2, c3, qh, qv,
      dTolerada, muyChico, minDistanciaZ, escaladorZ,
      uMin, uMax, vMin, vMax, wMin, wMax, radio2D, maxLongitudSegmento2D,
      maxLongitudSegmento3D, dosin22, dosin12, cuatroin1in2,
      cuatroIn1, cuatroin2: extended;
    claveMira: CubeKey;
    cabo, rabo, lados, semilados3D, centro, mira, hPant, vPant, nPant, incr:
    vectorExtended;
  end;
  MedSizeStr = array[0..kMaxActors] of Byte;
  IniDataExt = array[0..kMaxData] of extended;

  FunctionAsArrayExt = record
    theActors: MedSizeStr;
    data: IniDataExt;
  end;

{$ALIGN 8}

{*************************************************************************}

var
  theFile: file;

procedure ReportIOError(er: Integer);
var
  e: Integer;
  m: AnsiString;
begin
  if er = -1 then
    m := eUnknownSurfaceType
  else if er = -2 then
    m := eBadFileType
  else if er = -3 then
    m := eFileNotAssigned
  else begin
    e := GetLastError;
    m := SysErrorMessage(e);
  end;
  ShowMessage(m);
end;

procedure FillDataToSave(var dts: DataToSave);
begin
  dts.gViewPoint := gViewPoint;
  dts.gH := gH;
  dts.gV := gV;
  dts.gN := gN;
  dts.gC0 := gC0;
  dts.gC1 := gC1;
  dts.gC2 := gC2;
  dts.gC3 := gC3;
  dts.gpcx := gpcx;
  dts.gpcy := gpcy;
  dts.gVPCubeKey := gVPCubeKey;
end;

procedure ExtractData(dts: DataToSave);
begin
  gViewPoint := dts.gViewPoint;
  gH := dts.gH;
  gv := dts.gV;
  gN := dts.gN;
  gC0 := dts.gC0;
  gC1 := dts.gC1;
  gC2 := dts.gC2;
  gC3 := dts.gC3;
  gpcx := dts.gpcx;
  gpcy := dts.gpcy;
  gVPCubeKey := dts.gVPCubeKey;
end;

procedure FromOldMaptoMap(source, destination: ColorMapPtr);
var
  i, j: integer;
  theByte: byte;
  r, a: double;
begin
  if gType = PseudoRiemannian then
  begin
    r := 100 / 118;
    a := -882 / 118;
  end
  else
  begin
    r := 100 / 96;
    a := -3104 / 96;
  end;
  for i := 0 to kWidthMinus1 do
    for j := 0 to kHeightMinus1 do
    begin
      theByte := source^[i, j];
      if theByte < 10 then
        destination^[i, j] := 0
      else
        destination^[i, j] := round(r * theByte + a);
    end;
end;

function WriteToDisk(pathAndName: TFilename): Boolean;
var
  theFile: file;
  str: AnsiString;
  dts: DataToSave;
  i, j, len, vers, byteNumber: Integer;
  x: Double;
  aMap: ColorMapPtr;
begin
{$I-}
  result := False;
  if (pathAndName = '') then begin
    ReportIOError(-3);
    Exit;
  end;
  try
    AssignFile(theFile, pathAndName);
    Rewrite(theFile, 1);
    str := kCreator;
    BlockWrite(theFile, str[1], kCreatorLength);
    case gType of
      Parametric:
        str := kFileTypeParam;
      Implicit:
        str := kFileTypeImplicit;
      PseudoRiemannian:
        str := kFileTypeRiem;
      Tube:
        str := kFileTypeTube;
    else
      ;
    end;
    BlockWrite(theFile, str[1], kTypeLength);
    vers := gVersion;
    byteNumber := SizeOf(vers);
    BlockWrite(theFile, vers, byteNumber);
    byteNumber := SizeOf(Integer);
{$O-}
    for i := 1 to kNumOfSavedToDiskStrings do begin
      len := gSurfaceStrLengths[i];
      BlockWrite(theFile, len, byteNumber);
    end;
    for i := 1 to kNumOfSavedToDiskStrings do begin
      byteNumber := gSurfaceStrLengths[i];
      if byteNumber > 0 then begin
        str := gSurfaceStr[i];
        BlockWrite(theFile, str[1], byteNumber);
      end;
    end;
{$O+}
    FillDataToSave(dts);
    byteNumber := SizeOf(dts);
    BlockWrite(theFile, dts, byteNumber);
    byteNumber := SizeOf(ColorMap);
    aMap := gMap;
    BlockWrite(theFile, aMap^, byteNumber);
    if gClips <> nil then
      len := Length(gClips)
    else
      len := 0;
    byteNumber := SizeOf(Integer);
    BlockWrite(theFile, len, byteNumber);
    byteNumber := SizeOf(double);
    if len > 0 then
      for i := 0 to len - 1 do begin
        for j := 1 to 3 do begin
          x := gClips[i].u[j];
          BlockWrite(theFile, x, byteNumber);
        end;
        x := gClips[i].a;
        BlockWrite(theFile, x, byteNumber);
      end;
  finally
    result := (ioResult = 0);
    if not result then
      ReportIOError(ioResult)
    else
      gSaved := True;
    CloseFile(theFile);
  end;
{$I+}
end;

function ReadFromDiskOldVersion: Integer;
var
  vers: Int64;
  x: Double;
  header: array[0..7] of char;
  short: string[4];
  i, j, count, readNo, position: Integer;
  aChar, separator: char;
  oldMap, aMap: ColorMapPtr;
  oldData: dataExt;
begin
  result := 0;
  try
    Seek(theFile, 0);
    BlockRead(theFile, header, 8);
    short := Copy(header, 5, 4);
    if (short = 'x(uv') then
      gType := Parametric
    else if (short = 'f=0f') then
      gType := Implicit
    else if (short = 'tubo') then
      gType := Tube
    else if (short = 'efgR') then
      gType := PseudoRiemannian
    else begin
      result := -1; {Unknown surface type}
      Exit;
    end;
    BlockRead(theFile, vers, 8, readNo);
    if (vers <> 0) and (vers <> 3500) then
      Exit;
    gVersion := vers;
    position := FilePos(theFile);
    gSurfaceStr[1] := '';
    aChar := 'a';
    readNo := 1;
    while (aChar <> #0) and (readNo < 1000) do begin
      BlockRead(theFile, aChar, 1);
      gSurfaceStr[1] := gSurfaceStr[1] + aChar;
      inc(readNo);
    end;
    gSurfaceStrLengths[1] := Length(gSurfaceStr[1]);

    gSurfaceStr[2] := '';
    aChar := 'a';
    position := position + 1000;
    Seek(theFile, position);
    readNo := 1;
    while (aChar <> #0) and (readNo < 1000) do begin
      BlockRead(theFile, aChar, 1);
      gSurfaceStr[2] := gSurfaceStr[2] + aChar;
      inc(readNo);
    end;
    gSurfaceStrLengths[2] := Length(gSurfaceStr[2]);

    gSurfaceStr[3] := '';
    aChar := 'a';
    position := position + 1000;
    Seek(theFile, position);
    readNo := 1;
    while (aChar <> #0) and (readNo < 1000) do begin
      BlockRead(theFile, aChar, 1);
      gSurfaceStr[3] := gSurfaceStr[3] + aChar;
      inc(readNo);
    end;
    gSurfaceStrLengths[3] := Length(gSurfaceStr[3]);

    position := position + 1000 + 3 * SizeOf(FunctionAsArrayExt);
    Seek(theFile, position);
    count := SizeOf(dataExt);
    BlockRead(theFile, oldData, count);
    count := SizeOf(ColorMap);
    oldMap := gNormalPairMap;
    aMap := gMap;
    BlockRead(theFile, oldMap^, count);
    position := FilePos(theFile) + 1000;
    FromOldMapToMap(oldMap, aMap);
    gSurfaceStr[4] := '';
    if gType = Tube then begin
      aChar := 'a';
      position := FilePos(theFile);
      readNo := 1;
      while (aChar <> #0) and (readNo < 999) do begin
        BlockRead(theFile, aChar, 1);
        gSurfaceStr[4] := gSurfaceStr[4] + aChar;
        inc(readNo);
      end;
    end;
    gSurfaceStrLengths[4] := Length(gSurfaceStr[4]);
    gClips := nil;
    if vers >= 3500 then begin
      count := SizeOf(gClipTop);
      BlockRead(theFile, gClipTop, count);
      dec(gClipTop);
      if gClipTop >= 0 then begin
        Seek(theFile, position);
        SetLength(gClips, gClipTop + 1);
        count := SizeOf(double);
        for i := 0 to gClipTop do
        begin
          for j := 1 to 3 do
          begin
            BlockRead(theFile, x, count);
            gClips[i].u[j] := x;
          end;
          BlockRead(theFile, x, count);
          gClips[i].a := x;
        end;
      end;
    end;
    separator := DecimalSeparator;
    DecimalSeparator := '.';
    gSurfaceStr[5] := FloatToStrF(oldData.uMin, FFFixed, 5, 3);
    gSurfaceStr[6] := FloatToStrF(oldData.uMax, FFFixed, 5, 3);
    gSurfaceStr[7] := FloatToStrF(oldData.vMin, FFFixed, 5, 3);
    gSurfaceStr[8] := FloatToStrF(oldData.vMax, FFFixed, 5, 3);
    gSurfaceStr[9] := FloatToStrF(oldData.wMin, FFFixed, 5, 3);
    gSurfaceStr[10] := FloatToStrF(oldData.wMax, FFFixed, 5, 3);
    DecimalSeparator := separator;
    for i := 5 to 10 do
      gSurfaceStrLengths[i] := Length(gSurfaceStr[i]);
    for i := 1 to 3 do begin
      gViewPoint[i] := oldData.mira[i];
      gH[i] := oldData.hPant[i];
      gV[i] := oldData.vPant[i];
      gN[i] := -oldData.nPant[i];
    end;
    gC0 := -oldData.c0;
    gC1 := oldData.c2;
    gC2 := oldData.c3;
    gC3 := -oldData.c1;
    gpcx := Ceil(-oldData.qh);
    gpcy := Ceil(-oldData.qv);
    gVPCubeKey := oldData.claveMira;
  finally
  end;
end;

function ReadFromDisk(pathAndName: TFilename): Boolean;
var
  str: AnsiString;
  dts: DataToSave;
  x: Double;
  i, j, byteNumber, len, vers, er: Integer;
  aMap: ColorMapPtr;
begin
{$I-}
  Result := False;
  er := 0;
  if gBusy then
    Exit;
  try
    AssignFile(theFile, pathAndName);
    FileMode := fmOpenRead;
    Reset(theFile, 1);
    SetLength(str, kCreatorLength);
    BlockRead(theFile, str[1], kCreatorLength);
    if (str <> kCreator) then begin
      if Copy(str, 1, 4) = 'GTVA' then
        er := ReadFromDiskOldVersion
      else
        er := -2;
    end
    else begin
      SetLength(str, kTypeLength);
      BlockRead(theFile, str[1], kTypeLength);
      if (str = kFileTypeParam) or gBusy then
        gType := Parametric
      else if (str = kFileTypeImplicit) then
        gType := Implicit
      else if (str = kFileTypeTube) then
        gType := Tube
      else if (str = kFileTypeRiem) then
        gType := PseudoRiemannian
      else begin
        CloseFile(theFile);
        er := -1;
      end;
      if (er = 0) then begin
        byteNumber := SizeOf(vers);
        BlockRead(theFile, vers, byteNumber);
        gVersion := vers;
        byteNumber := SizeOf(Integer);
{$O-}
        for i := 1 to kNumOfSavedToDiskStrings do begin
          BlockRead(theFile, len, byteNumber);
          gSurfaceStrLengths[i] := len;
        end;
        for i := 1 to kNumOfSavedToDiskStrings do
          if gSurfaceStrLengths[i] > 0 then begin
            len := gSurfaceStrLengths[i];
            SetLength(str, len);
            BlockRead(theFile, str[1], len);
            gSurfaceStr[i] := str;
          end
          else
            gSurfaceStr[i] := '';
{$O+}
        byteNumber := SizeOf(dts);
        BlockRead(theFile, dts, byteNumber);
        ExtractData(dts);
        byteNumber := SizeOf(ColorMap);
        aMap := gMap;
        BlockRead(theFile, aMap^, byteNumber);
        byteNumber := SizeOf(Integer);
        BlockRead(theFile, len, byteNumber);
        gClips := nil;
        gClipTop := len - 1;
        if len > 0 then begin
          SetLength(gClips, len);
          byteNumber := SizeOf(double);
          for i := 0 to len - 1 do begin
            for j := 1 to 3 do begin
              BlockRead(theFile, x, byteNumber);
              gClips[i].u[j] := x;
            end;
            BlockRead(theFile, x, byteNumber);
            gClips[i].a := x;
          end;
        end;
      end;
    end;
  finally
    if ioResult <> 0 then
      ReportIOError(ioResult)
    else if er <> 0 then
      ReportIOError(er);
    CloseFile(theFile);
    FileMode := fmOpenReadWrite;
  end;
  if (ioResult = 0) and (er = 0) then begin
    gSaved := True;
    Result := True;
  end;
{$I+}
end;

end.
