

{-----------------------------------------------------------------------------
 Unit Name: FineDrawing
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Draws the surface for printing with resolution greater than that used
      in the screen.
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit FineDrawing;

interface

function DrawWithHighResolution: Boolean;
{Draws the surface to gForPrintingBmp for printing with resolution greater than 400 x 400}
{Result shall be true if all goes well and user does not cancel, or false otherwise}

implementation

uses
  Math, Types, 
{$IFDEF MSWINDOWS}
  Windows, Graphics, Forms, Controls,
{$ENDIF}
{$IFDEF LINUX}
  QGraphics, QForms, QControls,
{$ENDIF}
  OCFunct, Globals, Utilities, DrawImplicitSurf, BasicRiemUtils;

type
  uDataBuffer = array of array of single;
  uTColorArray = array[0..4000] of TColor;
  upTColorArray = ^TColorArray;

  threeBytes = array[1..3] of Byte;
  triang = array[1..3] of PointData;
  triangFoc = array[1..3] of FocusSpacePtData;
var
  gId: threeBytes;
  cInitVal: array[1..3] of array of single;
  cVal: array of Single;
  cValDerivs: array of Vector2Single;
  zBuffer, fBuffer: uDataBuffer;
  uPointsData: PointDataDeposit; {for the grid to draw a parametric surface}
  uFocusPtData: FocusSpacePtDataDeposit;
  uStep2D: Vector2;
  uC0, upcx, upcy, uFC0: Double;
  prWidth, prHeight: Integer;
  uGrid, vGrid: Integer;


function uProjectSToPaneRealAndDist(x: Vector3Single): Vector3Single;
begin
  result[3] := (gN[1] * x[1] + gN[2] * x[2] + gN[3] * x[3] - gC3);
  result[1] := (gH[1] * x[1] + gH[2] * x[2] + gH[3] * x[3] - gC1) * uC0 /
    result[3]
    +
    upcx;
  result[2] := (gV[1] * x[1] + gV[2] * x[2] + gV[3] * x[3] - gC2) * uC0 /
    result[3]
    +
    upcy;
end;

function uProjectToFocusPlaneRealAndDist(x: Vector3Single): Vector3Single;
begin
  result[3] := gFN[1] * x[1] + gFN[2] * x[2] + gFN[3] * x[3] - gFC3;
  result[1] := (gFH[1] * x[1] + gFH[2] * x[2] + gFH[3] * x[3] - gFC1) * uFC0 /
    result[3]
    + prWidth div 2;
  result[2] := (gFV[1] * x[1] + gFV[2] * x[2] + gFV[3] * x[3] - gFC2) * uFC0 /
    result[3]
    + prHeight div 2;
end;


function InitializeBigVars: Boolean;
var
  i, j: Integer;
begin
  result := True;
  try
    zBuffer := nil;
    SetLength(zBuffer, prWidth);
    for i := 0 to prWidth - 1 do
      SetLength(zBuffer[i], prHeight);
    fBuffer := nil;
    SetLength(fBuffer, prWidth);
    for i := 0 to prWidth - 1 do
      SetLength(fBuffer[i], prHeight);
    for i := 0 to prWidth - 1 do
      for j := 0 to prHeight - 1 do begin
        zBuffer[i, j] := maxInt;
        fBuffer[i, j] := maxInt;
      end;
    uPointsData := nil;
    SetLength(uPointsData, uGrid + 1);
    for i := 0 to uGrid do
      SetLength(uPointsData[i], vGrid + 1);
    uFocusPtData := nil;
    SetLength(uFocusPtData, uGrid + 1);
    for i := 0 to uGrid do
      SetLength(uFocusPtData[i], vGrid + 1);
  except
    if zBuffer <> nil then begin
      for i := 0 to prWidth - 1 do
        zBuffer[i] := nil;
      zBuffer := nil;
    end;
    if fBuffer <> nil then begin
      for i := 0 to prWidth - 1 do
        fBuffer[i] := nil;
      fBuffer := nil;
    end;
    if uPointsData <> nil then begin
      for i := 0 to uGrid do
        uPointsData[i] := nil;
      uPointsData := nil;
    end;
    if uFocusPtData <> nil then begin
      for i := 0 to uGrid do
        uFocusPtData[i] := nil;
      uFocusPtData := nil;
    end;
    result := False;
  end;
end;

function ComputeSpecialNormal(i, j: Integer; var n: Vector3Single; var norm:
  single):
  Boolean;
var
  k: Integer;
  ph, pv: Vector3Single;
begin
  if i < uGrid then begin
    if j < vGrid then
      for k := 1 to 3 do begin
        ph[k] := uPointsData[i + 1, j + 1].pos[k] - uPointsData[i,
          j].pos[k];
        pv[k] := uPointsData[i, j + 1].pos[k] - uPointsData[i + 1,
          j].pos[k];
      end
    else
      for k := 1 to 3 do begin
        ph[k] := uPointsData[i, j - 1].pos[k] - uPointsData[i + 1,
          j].pos[k];
        pv[k] := uPointsData[i + 1,
          j - 1].pos[k] - uPointsData[i, j].pos[k];
      end
  end
  else begin
    if j < vGrid then
      for k := 1 to 3 do begin
        ph[k] := uPointsData[i - 1, j].pos[k] - uPointsData[i,
          j + 1].pos[k];
        pv[k] := uPointsData[i, j].pos[k] - uPointsData[i - 1,
          j + 1].pos[k];
      end
    else
      for k := 1 to 3 do begin
        ph[k] := uPointsData[i, j].pos[k] - uPointsData[i - 1,
          j - 1].pos[k];
        pv[k] := uPointsData[i - 1, j].pos[k] - uPointsData[i,
          j - 1].pos[k];
      end
  end;
  n[1] := ph[2] * pv[3] - ph[3] * pv[2];
  n[2] := ph[3] * pv[1] - ph[1] * pv[3];
  n[3] := ph[1] * pv[2] - ph[2] * pv[1];
  norm := EuclNorm3S(n);
  Result := norm > 1.0E-11;
end;

procedure AssignCol(i, j: Integer; ph, pv:
  Vector3);
const
  dullness: double = 0.3;
  OneMinusDullness: double = 1 - 0.3 {1 - dullness};
  intFocus: double = 60;
  intAmb: double = 40;
var
  ii: Integer;
  ray, normal, fRay: Vector3Single;
  rayNorm, fRayNorm, aux, sigma, lambda, omega, fComp: single;
begin
  for ii := 1 to 3 do begin
    ray[ii] := gViewpoint[ii] - uPointsData[i, j].pos[ii];
    fRay[ii] := gFocus[ii] - uPointsData[i, j].pos[ii];
  end;
  rayNorm := EuclNorm3S(ray);
  fRayNorm := EuclNorm3S(fRay);
  for ii := 1 to 3 do begin
    ray[ii] := ray[ii] / rayNorm;
    fRay[ii] := fRay[ii] / fRayNorm;
  end;
  normal[1] := ph[2] * pv[3] - ph[3] * pv[2];
  normal[2] := ph[3] * pv[1] - ph[1] * pv[3];
  normal[3] := ph[1] * pv[2] - ph[2] * pv[1];
  aux := EuclNorm3S(normal);
  if (aux > 1.0E-11) or ComputeSpecialNormal(i, j, normal, aux) then begin
    for ii := 1 to 3 do
      normal[ii] := normal[ii] / aux;
    sigma := ray[1] * normal[1] + ray[2] * normal[2] + ray[3] * normal[3];
    lambda := fRay[1] * normal[1] + fRay[2] * normal[2] + fRay[3] *
      normal[3];
    omega := 2 * lambda * sigma;
    uPointsData[i, j].ambLight := intAmb * (1 + abs(sigma)) / 2;
    if sigma < 0 then
      uPointsData[i, j].ambLight := -uPointsData[i, j].ambLight;
    if omega > 0 then begin
      lambda := abs(lambda);
      omega := (1 + omega - (fRay[1] * ray[1] + fRay[2] * ray[2] +
        fRay[3] * ray[3])) / 2;
      fComp := lambda * (OneMinusDullness * sqr(omega) + dullness * (1 +
        abs(sigma)) / 2);
      uFocusPtData[i, j].focLight := Min(99 - abs(uPointsData[i,
        j].ambLight), intFocus * fComp);
    end
    else
      uFocusPtData[i, j].focLight := 0;
  end
  else begin
    uPointsData[i, j].ambLight := 100;
    uFocusPtData[i, j].focLight := 0;
  end;

end;



procedure FillPointData(di, dj: Integer;
  ph, pv: Vector3; allData: Boolean);
var
  i: Integer;
  p: Vector3Single;
  a: Single;
begin
  p := uPointsData[di, dj].pos;
  uPointsData[di, dj].outside := False;
  uPointsData[di, dj].inside := True;
  uPointsData[di, dj].proj := uProjectSToPaneRealAndDist(p);
  if allData then begin
    uFocusPtData[di, dj].pt := uProjectToFocusPlaneRealAndDist(p);
    AssignCol(di, dj, ph, pv);
  end;
  if gClips <> nil then
    for i := 0 to High(gClips) do begin
      a := HeightOverClipS(gClips[i], p);
      if a > 0 then begin
        uPointsData[di, dj].inside := False;
        a := a - (abs(gClips[i].u[1]) * (abs(ph[1]) + abs(pv[1])) +
          abs(gClips[i].u[2]) * (abs(ph[2]) + abs(pv[2])) +
          abs(gClips[i].u[3]) * (abs(ph[3]) + abs(pv[3])));
        uPointsData[di, dj].outside := (a > 0);
        if uPointsData[di, dj].outside then Exit;
      end
    end;
end;



procedure ComputeGridData(allData: Boolean);
 {We compute the illumination that would reach the}
 {eye from each point of the grid as if it were visible from the eye. }
 {We take advantage of the stored data as far as possible.}
const
  cc = 16;
var
  i, j, k: Integer;
  u, w, hu, hv: double;
  p, pv, ph: Vector3;
  aPointData: PointData;
  aFocusSpaceData: FocusSpacePtData;
begin
  hu := uStep2D[1] / 10;
  hv := uStep2D[2] / 10;
  for i := 0 to uGrid do
    for j := 0 to vGrid do begin
      theFPar(p, gMin2D[1] + i * uStep2D[1], gMin2D[2] + j * uStep2D[2]);
      uPointsData[i, j].pos[1] := p[1];
      uPointsData[i, j].pos[2] := p[2];
      uPointsData[i, j].pos[3] := p[3];
    end;

 {First, the corners}
  u := gMin2D[1];
  w := gMin2D[2];
  theFPar(ph, u + hu, w);
  theFPar(pv, u, w + hv);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - uPointsData[0, 0].pos[k]);
    pv[k] := cc * (pv[k] - uPointsData[0, 0].pos[k]);
  end;
  FillPointData(0, 0, ph, pv, allData);

  u := gMax2D[1];
  theFPar(ph, u, w + hv);
  theFPar(pv, u - hu, w);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - uPointsData[uGrid, 0].pos[k]);
    pv[k] := cc * (pv[k] - uPointsData[uGrid, 0].pos[k]);
  end;
  FillPointData(uGrid, 0, ph, pv, allData);

  w := gMax2D[2];
  theFPar(ph, u - hu, w);
  theFPar(pv, u, w - hv);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - uPointsData[uGrid, vGrid].pos[k]);
    pv[k] := cc * (pv[k] - uPointsData[uGrid, vGrid].pos[k]);
  end;
  FillPointData(uGrid, vGrid, ph, pv, allData);

  u := gMin2D[1];
  theFPar(ph, u, w - hv);
  theFPar(pv, u + hu, w);
  for k := 1 to 3 do begin
    ph[k] := cc * (ph[k] - uPointsData[0, vGrid].pos[k]);
    pv[k] := cc * (pv[k] - uPointsData[0, vGrid].pos[k]);
  end;
  FillPointData(0, vGrid, ph, pv, allData);

 {now the limiting rows}
  for i := 1 to uGrid - 1 do begin
    u := gMin2D[1] + i * uStep2D[1];
    w := gMin2D[2];
    theFPar(pv, u, w + hv);
    theFPar(ph, u + hu, w);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - uPointsData[i, 0].pos[k]);
      pv[k] := cc * (pv[k] - uPointsData[i, 0].pos[k]);
    end;
    FillPointData(i, 0, ph, pv, allData);

    w := gMax2D[2];
    theFPar(pv, u, w - hv);
    theFPar(ph, u - hu, w);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - uPointsData[i, vGrid].pos[k]);
      pv[k] := cc * (pv[k] - uPointsData[i, vGrid].pos[k]);
    end;
    FillPointData(i, vGrid, ph, pv, allData);
  end;

  for i := 1 to vGrid - 1 do begin
    w := gMin2D[2] + i * uStep2D[2];
    u := gMax2D[1];
    theFPar(pv, u - hu, w);
    theFPar(ph, u, w + hv);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - uPointsData[uGrid, i].pos[k]);
      pv[k] := cc * (pv[k] - uPointsData[uGrid, i].pos[k]);
    end;
    FillPointData(uGrid, i, ph, pv, allData);

    u := gMin2D[1];
    theFPar(ph, u + hu, w);
    theFPar(pv, u, w + hv);
    for k := 1 to 3 do begin
      ph[k] := cc * (ph[k] - uPointsData[0, i].pos[k]);
      pv[k] := cc * (pv[k] - uPointsData[0, i].pos[k]);
    end;
    FillPointData(0, i, ph, pv, allData);

  end;

  {And, finally, the interior of the domain}
  for i := 1 to uGrid - 1 do
    for j := 1 to vGrid - 1 do begin
      for k := 1 to 3 do begin
        ph[k] := uPointsData[i + 1, j].pos[k] - uPointsData[i,
          j].pos[k];
        pv[k] := uPointsData[i, j + 1].pos[k] - uPointsData[i,
          j].pos[k];
      end;
      FillPointData(i, j, ph, pv, allData);
    end;
end;


procedure FillTriangle(vertices: triang; triangleF:
  triangFoc);
var
  permu: threeBytes;
  virt: TPoint;
  A, B, C: PointData;
  AE, BE, CE: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy, light: single;
  dfp, dfq, dfExtrapol: single;
  val, initialVal: array[1..6] of single;
  derivs: array[1..6, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, pm: Integer;
  auxab, auxac, auxbc: double;
  line: upTColorArray;
  noFBorder, haveDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    derivs[1, 1] := A.ambLight * ax + B.ambLight * bx + C.ambLight * cx;
    derivs[1, 2] := A.ambLight * ay + B.ambLight * by + C.ambLight * cy;
    derivs[2, 1] := AE.focLight * ax + BE.focLight * bx + CE.focLight * cx;
    derivs[2, 2] := AE.focLight * ay + BE.focLight * by + CE.focLight * cy;
    derivs[3, 1] := AE.pt[1] * ax + BE.pt[1] * bx +
      CE.pt[1] * cx;
    derivs[3, 2] := AE.pt[1] * ay + BE.pt[1] * by +
      CE.pt[1] * cy;
    derivs[4, 1] := AE.pt[2] * ax + BE.pt[2] * bx +
      CE.pt[2] * cx;
    derivs[4, 2] := AE.pt[2] * ay + BE.pt[2] * by +
      CE.pt[2] * cy;
    derivs[5, 1] := A.proj[3] * ax + B.proj[3] * bx + C.proj[3] * cx;
    derivs[5, 2] := A.proj[3] * ay + B.proj[3] * by + C.proj[3] * cy;
    derivs[6, 1] := AE.pt[3] * ax + BE.pt[3] * bx +
      CE.pt[3] * cx;
    derivs[6, 2] := AE.pt[3] * ay + BE.pt[3] * by +
      CE.pt[3] * cy;
    det := derivs[4, 2] * derivs[3, 1] - derivs[4, 1] * derivs[3, 2];
    noFBorder := abs(det) > 1.0E-7;
    if noFBorder then begin
      dfp := (derivs[5, 1] * derivs[4, 2] - derivs[5, 2] * derivs[4, 1]) /
        det;
      dfq := (derivs[5, 2] * derivs[3, 1] - derivs[5, 1] * derivs[3, 2]) /
        det;
    end;
    haveDerivs := True;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    for ii := 1 to 6 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 6 do
      val[ii] := val[ii] + derivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].proj[2] < vertices[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].proj[2] < vertices[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].proj[2] < vertices[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  AE := triangleF[permu[1]];
  BE := triangleF[permu[2]];
  CE := triangleF[permu[3]];

  ACx := A.proj[1] - C.proj[1];
  ACy := A.proj[2] - C.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.proj[1] - B.proj[1];
  ABy := A.proj[2] - B.proj[2];
  BCx := B.proj[1] - C.proj[1];
  BCy := B.proj[2] - C.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;

  auxab := A.ambLight + B.ambLight + C.ambLight;
  if auxab < 0 then begin
    A.ambLight := abs(A.ambLight) + 101;
    B.ambLight := abs(B.ambLight) + 101;
    C.ambLight := abs(C.ambLight) + 101;
  end
  else begin
    A.ambLight := abs(A.ambLight);
    B.ambLight := abs(B.ambLight);
    C.ambLight := abs(C.ambLight);
  end;

  coefAC := ACx / ACy;
  initialVal[1] := A.ambLight;
  initialVal[2] := AE.FocLight;
  initialVal[3] := AE.pt[1];
  initialVal[4] := AE.pt[2];
  initialVal[5] := A.proj[3];
  initialVal[6] := AE.pt[3];

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(prHeight - 1, floor(B.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.proj[2];
      auxac := A.proj[1] + difj * coefAC;
      auxab := A.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        line := gForPrintingBmp.ScanLine[j];
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), prWidth - 1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), prWidth - 1);
        end;

        if here <= there then begin
          difi := here - A.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            virt.x := Floor(val[3]);
            virt.y := Floor(val[4]);
            light := 0;
            if (virt.x >= 0) and (virt.x < prWidth) and (virt.y
              >= 0) and (virt.y < prHeight) and noFBorder then begin
              dfExtrapol := val[6] + dfp * (virt.x -
                val[3]) + dfq * (virt.y - val[4]);
              if dfExtrapol < 1.002 * fBuffer[virt.x, virt.y] then
                light := val[2];
            end;
            if (val[5] < zBuffer[i, j]) then begin
              zBuffer[i, j] := val[5];
              light := light + val[1];
              pm := floor(light);
              line^[i] := gPalette3D^[pm];
            end
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(prHeight - 1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.proj[2];
    difb := jInf - 1 - B.proj[2];
    auxac := A.proj[1] + difj * coefAC;
    auxbc := B.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      line := gForPrintingBmp.ScanLine[j];
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), prWidth - 1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), prWidth - 1);
      end;
      if here <= there then begin
        difi := here - A.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          virt.x := Floor(val[3]);
          virt.y := Floor(val[4]);
          light := 0;
          if (virt.x >= 0) and (virt.x < prWidth) and (virt.y
            >= 0) and (virt.y < prHeight) and noFBorder then begin
            dfExtrapol := val[6] + dfp * (virt.x -
              val[3]) + dfq * (virt.y - val[4]);
            if dfExtrapol < 1.002 * fBuffer[virt.x, virt.y] then
              light := val[2];
          end;
          if (val[5] < zBuffer[i, j]) then begin
            zBuffer[i, j] := val[5];
            light := light + val[1];
            pm := floor(light);
            line^[i] := gPalette3D^[pm];
          end
        end;
      end;
    end;
  end;
end;

procedure FillTriangleWithClips(vertices: triang; triangleF:
  triangFoc);
var
{x,y Drawing3D coordinates; u,v parameters of the point}
{p,q coordinates in the virtual screen used for the Focus}
  permu: threeBytes;
  virt: TPoint;
  A, B, C: PointData;
  AE, BE, CE: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy, light: single;
  dfp, dfq, dfExtrapol: single;
  val, initialVal: array[1..6] of single;
  derivs: array[1..6, 1..2] of single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, pm, k: Integer;
  auxab, auxac, auxbc: double;
  line: upTColorArray;
  noFBorder, haveDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    derivs[1, 1] := A.ambLight * ax + B.ambLight * bx + C.ambLight * cx;
    derivs[1, 2] := A.ambLight * ay + B.ambLight * by + C.ambLight * cy;
    derivs[2, 1] := AE.focLight * ax + BE.focLight * bx + CE.focLight * cx;
    derivs[2, 2] := AE.focLight * ay + BE.focLight * by + CE.focLight * cy;
    derivs[3, 1] := AE.pt[1] * ax + BE.pt[1] * bx +
      CE.pt[1] * cx;
    derivs[3, 2] := AE.pt[1] * ay + BE.pt[1] * by +
      CE.pt[1] * cy;
    derivs[4, 1] := AE.pt[2] * ax + BE.pt[2] * bx +
      CE.pt[2] * cx;
    derivs[4, 2] := AE.pt[2] * ay + BE.pt[2] * by +
      CE.pt[2] * cy;
    derivs[5, 1] := A.proj[3] * ax + B.proj[3] * bx + C.proj[3] * cx;
    derivs[5, 2] := A.proj[3] * ay + B.proj[3] * by + C.proj[3] * cy;
    derivs[6, 1] := AE.pt[3] * ax + BE.pt[3] * bx +
      CE.pt[3] * cx;
    derivs[6, 2] := AE.pt[3] * ay + BE.pt[3] * by +
      CE.pt[3] * cy;
    det := derivs[4, 2] * derivs[3, 1] - derivs[4, 1] * derivs[3, 2];
    noFBorder := abs(det) > 1.0E-7;
    if noFBorder then begin
      dfp := (derivs[5, 1] * derivs[4, 2] - derivs[5, 2] * derivs[4, 1]) /
        det;
      dfq := (derivs[5, 2] * derivs[3, 1] - derivs[5, 1] * derivs[3, 2]) /
        det;
    end;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := cInitVal[1, k] * ax + cInitVal[2, k] * bx +
        cInitVal[3, k] * cx;
      cValDerivs[k, 2] := cInitVal[1, k] * ay + cInitVal[2, k] * by +
        cInitVal[3, k] * cy;
    end;
    haveDerivs := True;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    for ii := 1 to 6 do
      val[ii] := initialVal[ii] + derivs[ii, 1] * difi +
        derivs[ii, 2] * difj;
    for ii := 0 to gClipTop do
      cVal[ii] := cInitVal[1, ii] + cValDerivs[ii, 1] * difi + cValDerivs[ii,
        2]
        *
        difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    for ii := 1 to 6 do
      val[ii] := val[ii] + derivs[ii, 1];
    for ii := 0 to gClipTop do
      cVal[ii] := cVal[ii] + cValDerivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].proj[2] < vertices[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].proj[2] < vertices[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].proj[2] < vertices[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  AE := triangleF[permu[1]];
  BE := triangleF[permu[2]];
  CE := triangleF[permu[3]];

  ACx := A.proj[1] - C.proj[1];
  ACy := A.proj[2] - C.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.proj[1] - B.proj[1];
  ABy := A.proj[2] - B.proj[2];
  BCx := B.proj[1] - C.proj[1];
  BCy := B.proj[2] - C.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;

  auxab := A.ambLight + B.ambLight + C.ambLight;
  if auxab < 0 then begin
    A.ambLight := abs(A.ambLight) + 101;
    B.ambLight := abs(B.ambLight) + 101;
    C.ambLight := abs(C.ambLight) + 101;
  end
  else begin
    A.ambLight := abs(A.ambLight);
    B.ambLight := abs(B.ambLight);
    C.ambLight := abs(C.ambLight);
  end;

  coefAC := ACx / ACy;
  initialVal[1] := A.ambLight;
  initialVal[2] := AE.FocLight;
  initialVal[3] := AE.pt[1];
  initialVal[4] := AE.pt[2];
  initialVal[5] := A.proj[3];
  initialVal[6] := AE.pt[3];
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      cInitVal[i, j] := HeightOverClipS(gClips[j], vertices[permu[i]].pos);
  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(prHeight - 1, floor(B.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.proj[2];
      auxac := A.proj[1] + difj * coefAC;
      auxab := A.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        line := gForPrintingBmp.ScanLine[j];
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), prWidth - 1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), prWidth - 1);
        end;

        if here <= there then begin
          difi := here - A.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);
            if good then begin
              virt.x := Floor(val[3]);
              virt.y := Floor(val[4]);
              light := 0;
              if (virt.x >= 0) and (virt.x < prWidth) and (virt.y
                >= 0) and (virt.y < prHeight) and noFBorder then begin
                dfExtrapol := val[6] + dfp * (virt.x -
                  val[3]) + dfq * (virt.y - val[4]);
                if dfExtrapol < 1.002 * fBuffer[virt.x, virt.y]
                  then
                  light := val[2];
              end;
              if (val[5] < zBuffer[i, j]) then begin
                zBuffer[i, j] := val[5];
                light := light + val[1];
                pm := floor(light);
                line^[i] := gPalette3D^[pm];
              end;
            end;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(prHeight - 1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.proj[2];
    difb := jInf - 1 - B.proj[2];
    auxac := A.proj[1] + difj * coefAC;
    auxbc := B.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      line := gForPrintingBmp.ScanLine[j];
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), prWidth - 1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), prWidth - 1);
      end;
      if here <= there then begin
        difi := here - A.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good then begin
            virt.x := Floor(val[3]);
            virt.y := Floor(val[4]);
            light := 0;
            if (virt.x >= 0) and (virt.x < prWidth) and (virt.y
              >= 0) and (virt.y < prHeight) and noFBorder then begin
              dfExtrapol := val[6] + dfp * (virt.x -
                val[3]) + dfq * (virt.y - val[4]);
              if dfExtrapol < 1.002 * fBuffer[virt.x, virt.y] then
                light := val[2];
            end;
            if (val[5] < zBuffer[i, j]) then begin
              zBuffer[i, j] := val[5];
              light := light + val[1];
              pm := floor(light);
              line^[i] := gPalette3D^[pm];
            end;
          end;
        end;
      end;
    end;
  end;
end;

procedure FillZBufferTriangle(vertices: triang);
var
  permu: threeBytes;
  A, B, C: PointData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  val, initialVal: single;
  derivs: Vector2Single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there: Integer;
  auxab, auxac, auxbc: double;
  haveDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    derivs[1] := A.proj[3] * ax + B.proj[3] * bx + C.proj[3] * cx;
    derivs[2] := A.proj[3] * ay + B.proj[3] * by + C.proj[3] * cy;
  end;

  procedure ComputeData;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    val := initialVal + derivs[1] * difi +
      derivs[2] * difj;
  end;


  procedure IncrementI;
  begin
    val := val + derivs[1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].proj[2] < vertices[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].proj[2] < vertices[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].proj[2] < vertices[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  ACx := A.proj[1] - C.proj[1];
  ACy := A.proj[2] - C.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.proj[1] - B.proj[1];
  ABy := A.proj[2] - B.proj[2];
  BCx := B.proj[1] - C.proj[1];
  BCy := B.proj[2] - C.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;


  coefAC := ACx / ACy;
  initialVal := A.proj[3];

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(prHeight - 1, floor(B.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.proj[2];
      auxac := A.proj[1] + difj * coefAC;
      auxab := A.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), prWidth - 1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), prWidth - 1);
        end;

        if here <= there then begin
          difi := here - A.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            if (val < zBuffer[i, j]) then begin
              zBuffer[i, j] := val;
            end;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(prHeight - 1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.proj[2];
    difb := jInf - 1 - B.proj[2];
    auxac := A.proj[1] + difj * coefAC;
    auxbc := B.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), prWidth - 1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), prWidth - 1);
      end;
      if here <= there then begin
        difi := here - A.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          if (val < zBuffer[i, j]) then begin
            zBuffer[i, j] := val;
          end
        end;
      end;
    end;
  end;
end;

procedure ComputeZBuffer;
var
  i, j: integer;
  points: triang;
begin
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  ComputeGridData(False);
  for i := 1 to uGrid do
    for j := 1 to vGrid do begin
      points[1] := uPointsData[i, j];
      points[2] := uPointsData[i - 1, j - 1];
      points[3] := uPointsData[i - 1, j];
      FillZBufferTriangle(points);
      points[3] := uPointsData[i, j - 1];
      FillZBufferTriangle(points);
    end;
end;

procedure FillZBufferTriangleWithClips(vertices: triang);
var
  permu: threeBytes;
  A, B, C: PointData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  val, initialVal: single;
  derivs: Vector2single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, k: Integer;
  auxab, auxac, auxbc: double;
  haveDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    derivs[1] := A.proj[3] * ax + B.proj[3] * bx + C.proj[3] * cx;
    derivs[2] := A.proj[3] * ay + B.proj[3] * by + C.proj[3] * cy;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := cInitVal[1, k] * ax + cInitVal[2, k] * bx +
        cInitVal[3, k] * cx;
      cValDerivs[k, 2] := cInitVal[1, k] * ay + cInitVal[2, k] * by +
        cInitVal[3, k] * cy;
    end;
  end;

  procedure ComputeData;
  var
    ii: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    val := initialVal + derivs[1] * difi + derivs[2] * difj;
    for ii := 0 to gClipTop do
      cVal[ii] := cInitVal[1, ii] + cValDerivs[ii, 1] * difi + cValDerivs[ii,
        2]
        *
        difj;
  end;


  procedure IncrementI;
  var
    ii: Integer;
  begin
    val := val + derivs[1];
    for ii := 0 to gClipTop do
      cVal[ii] := cVal[ii] + cValDerivs[ii, 1];
  end;

begin
  haveDerivs := False;
  permu := gId;
  if vertices[2].proj[2] < vertices[1].proj[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if vertices[3].proj[2] < vertices[permu[1]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if vertices[3].proj[2] < vertices[permu[2]].proj[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := vertices[permu[1]];
  B := vertices[permu[2]];
  C := vertices[permu[3]];

  ACx := A.proj[1] - C.proj[1];
  ACy := A.proj[2] - C.proj[2];
  if ACy >= 0 then
    Exit;
  ABx := A.proj[1] - B.proj[1];
  ABy := A.proj[2] - B.proj[2];
  BCx := B.proj[1] - C.proj[1];
  BCy := B.proj[2] - C.proj[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;


  coefAC := ACx / ACy;
  initialVal := A.proj[3];
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      cInitVal[i, j] := HeightOverClipS(gClips[j], vertices[permu[i]].pos);

  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.proj[2]));
    jSup := min(prHeight - 1, floor(B.proj[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.proj[2];
      auxac := A.proj[1] + difj * coefAC;
      auxab := A.proj[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), prWidth - 1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), prWidth - 1);
        end;

        if here <= there then begin
          difi := here - A.proj[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);
            if good and (val < zBuffer[i, j]) then begin
              zBuffer[i, j] := val;
            end;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.proj[2]));
    jSup := min(prHeight - 1, floor(C.proj[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.proj[2];
    difb := jInf - 1 - B.proj[2];
    auxac := A.proj[1] + difj * coefAC;
    auxbc := B.proj[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), prWidth - 1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), prWidth - 1);
      end;
      if here <= there then begin
        difi := here - A.proj[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good and (val < zBuffer[i, j]) then begin
            zBuffer[i, j] := val;
          end
        end;
      end;
    end;
  end;
end;

procedure ComputeZBufferWithClips;
var
  i, j: integer;
  points: triang;
begin
  gClipTop := Length(gClips) - 1;
  for i := 1 to 3 do begin
    cInitVal[i] := nil;
    SetLength(cInitVal[i], gClipTop + 1);
  end;
  cVal := nil;
  SetLength(cVal, gClipTop + 1);
  cValDerivs := nil;
  SetLength(cValDerivs, gClipTop + 1);
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  ComputeGridData(False);
  for i := 1 to uGrid do
    for j := 1 to vGrid do begin
      points[1] := uPointsData[i, j];
      points[2] := uPointsData[i - 1, j - 1];
      points[3] := uPointsData[i - 1, j];
      if uPointsData[i, j].inside and uPointsData[i - 1, j - 1].inside
        and uPointsData[i - 1, j].inside then
        FillZBufferTriangle(points)
      else if not (uPointsData[i, j].outside or uPointsData[i - 1, j -
        1].outside or uPointsData[i - 1, j].outside) then
        FillZBufferTriangleWithClips(points);
      points[3] := uPointsData[i, j - 1];
      if uPointsData[i, j].inside and uPointsData[i - 1, j - 1].inside
        and uPointsData[i, j - 1].inside then
        FillZBufferTriangle(points)
      else if not (uPointsData[i, j].outside or uPointsData[i - 1, j -
        1].outside or uPointsData[i, j - 1].outside) then
        FillZBufferTriangleWithClips(points);
    end;
end;

procedure FillFBufferTriangle(triangleF:
  triangFoc);
var
  permu: threeBytes;
  A, B, C: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  df, initialDF, dfp, dfq: single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there: Integer;
  auxab, auxac, auxbc: double;
  haveDerivs, CAtLeftOfB: Boolean;

  procedure ComputeDerivatives;
  begin
    dfp := A.pt[3] * ax + B.pt[3] * bx + C.pt[3]
      * cx;
    dfq := A.pt[3] * ay + B.pt[3] * by + C.pt[3]
      * cy;
    haveDerivs := True;
  end;

  procedure ComputeData;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    df := initialDF + dfp * difi +
      dfq * difj;
  end;


begin
  haveDerivs := False;
  permu := gId;
  if triangleF[2].pt[2] < triangleF[1].pt[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triangleF[3].pt[2] < triangleF[permu[1]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triangleF[3].pt[2] < triangleF[permu[2]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triangleF[permu[1]];
  B := triangleF[permu[2]];
  C := triangleF[permu[3]];

  ACx := A.pt[1] - C.pt[1];
  ACy := A.pt[2] - C.pt[2];
  if ACy >= 0 then
    Exit;
  ABx := A.pt[1] - B.pt[1];
  ABy := A.pt[2] - B.pt[2];
  BCx := B.pt[1] - C.pt[1];
  BCy := B.pt[2] - C.pt[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;
  coefAC := ACx / ACy;
  initialDF := A.pt[3];
  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.pt[2]));
    jSup := min(prHeight - 1, floor(B.pt[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.pt[2];
      auxac := A.pt[1] + difj * coefAC;
      auxab := A.pt[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), prWidth - 1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), prWidth - 1);
        end;

        if here <= there then begin
          difi := here - A.pt[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            df := df + dfp;
            if df < fBuffer[i, j] then
              fBuffer[i, j] := df;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.pt[2]));
    jSup := min(prHeight - 1, floor(C.pt[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.pt[2];
    difb := jInf - 1 - B.pt[2];
    auxac := A.pt[1] + difj * coefAC;
    auxbc := B.pt[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), prWidth - 1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), prWidth - 1);
      end;
      if here <= there then begin
        difi := here - A.pt[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          df := df + dfp;
          if df < fBuffer[i, j] then
            fBuffer[i, j] := df;
        end;
      end;
    end;
  end;
end;


procedure FillFBufferTriangleWithClips(triangleF:
  triangFoc; poss: Matrix3Single);
var
  permu: threeBytes;
  A, B, C: FocusSpacePtData;
  ACx, ACy, ABx, ABy, BCx, BCy, ax, bx, cx, ay, by, cy: single;
  df, initialDF, dfp, dfq: single;
  difi, difj, difb, det, coefAC, coefAB, coefBC: single;
  j, jInf, jSup, i, here, there, k: Integer;
  auxab, auxac, auxbc: double;
  noFBorder, haveDerivs, CAtLeftOfB, good: Boolean;

  procedure ComputeDerivatives;
  var
    k: Integer;
  begin
    dfp := A.pt[3] * ax + B.pt[3] * bx + C.pt[3]
      * cx;
    dfq := A.pt[3] * ay + B.pt[3] * by + C.pt[3]
      * cy;
    for k := 0 to gClipTop do begin
      cValDerivs[k, 1] := cInitVal[1, k] * ax + cInitVal[2, k] * bx +
        cInitVal[3, k] * cx;
      cValDerivs[k, 2] := cInitVal[1, k] * ay + cInitVal[2, k] * by +
        cInitVal[3, k] * cy;
    end;
    haveDerivs := True;
  end;

  procedure ComputeData;
  var
    k: Integer;
  begin
    if not haveDerivs then
      ComputeDerivatives;
    df := initialDF + dfp * difi +
      dfq * difj;
    for k := 0 to gClipTop do
      cVal[k] := cInitVal[1, k] + cValDerivs[k, 1] * difi + cValDerivs[k, 2] *
        difj;
  end;

  procedure IncrementI;
  var
    k: Integer;
  begin
    df := df + dfp;
    for k := 0 to gClipTop do
      cVal[k] := cVal[k] + cValDerivs[k, 1];
  end;


begin
  haveDerivs := False;
  permu := gId;
  if triangleF[2].pt[2] < triangleF[1].pt[2] then begin
    permu[2] := 1;
    permu[1] := 2;
  end;
  if triangleF[3].pt[2] < triangleF[permu[1]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := permu[1];
    permu[1] := 3;
  end
  else if triangleF[3].pt[2] < triangleF[permu[2]].pt[2] then begin
    permu[3] := permu[2];
    permu[2] := 3;
  end;
  A := triangleF[permu[1]];
  B := triangleF[permu[2]];
  C := triangleF[permu[3]];
  for i := 1 to 3 do
    for j := 0 to gClipTop do
      cInitVal[i, j] := HeightOverClipS(gClips[j], poss[permu[i]]);
  ACx := A.pt[1] - C.pt[1];
  ACy := A.pt[2] - C.pt[2];
  if ACy >= 0 then
    Exit;
  ABx := A.pt[1] - B.pt[1];
  ABy := A.pt[2] - B.pt[2];
  BCx := B.pt[1] - C.pt[1];
  BCy := B.pt[2] - C.pt[2];
  det := ACx * BCy - ACy * BCx;
  if abs(det) < 1.0E-9 then
    Exit;
  det := 1 / det;
  ax := BCy * det;
  ay := -BCx * det;
  bx := -ACy * det;
  by := ACx * det;
  cx := ABy * det;
  cy := -ABx * det;
  coefAC := ACx / ACy;
  initialDF := A.pt[3];
  if ABy < 0 then begin
    coefAB := ABx / ABy;
    CAtLeftOfB := coefAC < coefAB;
    jInf := max(0, ceil(A.pt[2]));
    jSup := min(prHeight - 1, floor(B.pt[2]));
    if jSup >= jInf then begin
      difj := jInf - 1 - A.pt[2];
      auxac := A.pt[1] + difj * coefAC;
      auxab := A.pt[1] + difj * coefAB;
      for j := jInf to jSup do begin
        difj := difj + 1;
        auxac := auxac + coefAC;
        auxab := auxab + coefAB;
        if CAtLeftOfB then begin
          here := Max(0, ceil(auxac));
          there := Min(floor(auxab), prWidth - 1);
        end
        else begin
          here := Max(0, ceil(auxab));
          there := Min(floor(auxac), prWidth - 1);
        end;

        if here <= there then begin
          difi := here - A.pt[1] - 1;
          ComputeData;

          for i := here to there do begin
            difi := difi + 1;
            IncrementI;
            good := True;
            for k := 0 to gClipTop do
              good := good and (cVal[k] < 0);
            if good and (df < fBuffer[i, j]) then
              fBuffer[i, j] := df;
          end;
        end;
      end;
    end;
  end;

  if BCy < 0 then begin
    coefBC := BCx / BCy;
    jInf := max(0, ceil(B.pt[2]));
    jSup := min(prHeight - 1, floor(C.pt[2]));
    if jSup < jInf then
      Exit;
    difj := jInf - 1 - A.pt[2];
    difb := jInf - 1 - B.pt[2];
    auxac := A.pt[1] + difj * coefAC;
    auxbc := B.pt[1] + difb * coefBC;
    for j := jInf to jSup do begin
      difj := difj + 1;
      difb := difb + 1;
      auxac := auxac + coefAC;
      auxbc := auxbc + coefBC;
      if auxac < auxbc then begin
        here := Max(0, ceil(auxac));
        there := Min(floor(auxbc), prWidth - 1);
      end
      else begin
        here := Max(0, ceil(auxbc));
        there := Min(floor(auxac), prWidth - 1);
      end;
      if here <= there then begin
        difi := here - A.pt[1] - 1;
        ComputeData;
        for i := here to there do begin
          difi := difi + 1;
          IncrementI;
          good := True;
          for k := 0 to gClipTop do
            good := good and (cVal[k] < 0);
          if good and (df < fBuffer[i, j]) then
            fBuffer[i, j] := df;
        end;
      end;
    end;
  end;
end;

procedure ComputeFBufferWithClips;
var
  i, j: integer;
  triangleF: triangFoc;
  poss: Matrix3Single;
begin
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  for i := 1 to uGrid do
    for j := 1 to vGrid do begin
      triangleF[1] := uFocusPtData[i, j];
      triangleF[2] := uFocusPtData[i - 1, j - 1];
      triangleF[3] := uFocusPtData[i - 1, j];
      if uPointsData[i, j].inside and uPointsData[i - 1, j - 1].inside
        and
        uPointsData[i - 1, j].inside then
        FillFBufferTriangle(triangleF)
      else if not (uPointsData[i, j].outside or uPointsData[i - 1, j -
        1].outside or uPointsData[i - 1, j].outside) then begin
        poss[1] := uPointsData[i, j].pos;
        poss[2] := uPointsData[i - 1, j - 1].pos;
        poss[3] := uPointsData[i - 1, j].pos;
        FillFBufferTriangleWithClips(triangleF, poss);
      end;
      triangleF[3] := uFocusPtData[i, j - 1];
      if uPointsData[i, j].inside and uPointsData[i - 1, j - 1].inside
        and
        uPointsData[i, j - 1].inside then
        FillFBufferTriangle(triangleF)
      else if not (uPointsData[i, j].outside or uPointsData[i - 1, j -
        1].outside or uPointsData[i, j - 1].outside) then begin
        poss[1] := uPointsData[i, j].pos;
        poss[2] := uPointsData[i - 1, j - 1].pos;
        poss[3] := uPointsData[i, j - 1].pos;
        FillFBufferTriangleWithClips(triangleF, poss);
      end;
    end;
end;

procedure ComputeFBuffer;
var
  i, j: integer;
  triangleF: triangFoc;
begin
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  for i := 1 to uGrid do
    for j := 1 to vGrid do begin
      triangleF[1] := uFocusPtData[i, j];
      triangleF[2] := uFocusPtData[i - 1, j - 1];
      triangleF[3] := uFocusPtData[i - 1, j];
      FillFBufferTriangle(triangleF);
      triangleF[3] := uFocusPtData[i, j - 1];
      FillFBufferTriangle(triangleF);
    end;
end;

function DrawSurfaceWithoutClips: Boolean;
var
  i, j: Integer;
  vertices: triang;
  triangleF: triangFoc;
begin
  Result := False;
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  ComputeGridData(True);
  ComputeFBuffer;
  for i := 1 to uGrid do begin
    if i mod 20 = 0 then begin
      Application.processMessages;
      if gStop then Exit;
    end;
    for j := 1 to vGrid do begin
      vertices[1] := uPointsData[i, j];
      triangleF[1] := uFocusPtData[i, j];
      vertices[2] := uPointsData[i - 1, j - 1];
      triangleF[2] := uFocusPtData[i - 1, j - 1];
      vertices[3] := uPointsData[i - 1, j];
      triangleF[3] := uFocusPtData[i - 1, j];
      FillTriangle(vertices, triangleF);
      vertices[3] := uPointsData[i, j - 1];
      triangleF[3] := uFocusPtData[i, j - 1];
      FillTriangle(vertices, triangleF);
    end;
  end;
  Result := True;
end;

function DrawSurfaceWithClips: Boolean;
var
  i, j: Integer;
  vertices: triang;
  triangleF: triangFoc;
begin
  Result := False;
  gId[1] := 1;
  gId[2] := 2;
  gId[3] := 3;
  gClipTop := Length(gClips) - 1;
  for i := 1 to 3 do begin
    cInitVal[i] := nil;
    SetLength(cInitVal[i], gClipTop + 1);
  end;
  cVal := nil;
  SetLength(cVal, gClipTop + 1);
  cValDerivs := nil;
  SetLength(cValDerivs, gClipTop + 1);
  ComputeGridData(True);
  ComputeFBufferWithClips;
  for i := 1 to uGrid do begin
    if i mod 20 = 0 then begin
      Application.processMessages;
      if gStop then Exit;
    end;
    for j := 1 to vGrid do begin
      vertices[1] := uPointsData[i, j];
      triangleF[1] := uFocusPtData[i, j];
      vertices[2] := uPointsData[i - 1, j - 1];
      triangleF[2] := uFocusPtData[i - 1, j - 1];
      vertices[3] := uPointsData[i - 1, j];
      triangleF[3] := uFocusPtData[i - 1, j];
      if uPointsData[i, j].inside and uPointsData[i - 1, j - 1].inside
        and uPointsData[i - 1, j].inside then
        FillTriangle(vertices, triangleF)
      else if not (uPointsData[i, j].outside or uPointsData[i - 1, j -
        1].outside or uPointsData[i - 1, j].outside) then
        FillTriangleWithClips(vertices, triangleF);
      vertices[3] := uPointsData[i, j - 1];
      triangleF[3] := uFocusPtData[i, j - 1];
      if uPointsData[i, j].inside and uPointsData[i - 1, j - 1].inside
        and uPointsData[i, j - 1].inside then
        FillTriangle(vertices, triangleF)
      else if not (uPointsData[i, j].outside or uPointsData[i - 1, j -
        1].outside or uPointsData[i, j - 1].outside) then
        FillTriangleWithClips(vertices, triangleF);
    end;
  end;
  Result := True;
end;

function ComputeParamSurface: Boolean;
begin
  if gClips <> nil then
    result := DrawSurfaceWithClips
  else
    result := DrawSurfaceWithoutClips
end;

function ComputeNormalPairParam: Boolean;
const
  sinEyes: double = 1 / 15;
  cosEyes: double = 0.9977753031;
var
  nLeft, hLeft, vLeft: Vector3;
  j, k: integer;
  d: double;
  prevC0, prevC1, prevC2, prevC3, prevPcx: double;
  cen, prevH, prevV, prevN, prevWp: Vector3;
  line: upTColorArray;
begin
  result := False;
  prevC0 := uC0;
  uC0 := prevC0 / 2;
  prevPcx := upcx;
  upcx := round(prevPcx / 2);
  cen := gCenter3D;
  d := EuclDistance3(gViewpoint, cen);
  vLeft := gV;
  for j := 1 to 3 do begin
    nLeft[j] := cosEyes * gN[j] + sinEyes * gH[j];
    hLeft[j] := -sinEyes * gN[j] + cosEyes * gH[j];
  end;
  gForPrintingBmp.canvas.FillRect(Rect(0,0,prWidth,prHeight));
  if ComputeParamSurface then begin
    prevC1 := gC1;
    prevC2 := gC2;
    prevC3 := gC3;
    gC3 := EuclInnerProd3(nLeft, cen) - d;
    gC1 := EuclInnerProd3(hLeft, cen);
    gC2 := EuclInnerProd3(vLeft, cen);
    prevWp := gViewpoint;
    for j := 1 to 3 do
      gViewpoint[j] := gC3 * nLeft[j] + gC1 * hLeft[j] + gC2 * vLeft[j];
    upcx := upcx + prWidth div 2;
    prevH := gH;
    prevN := gN;
    prevV := gV;
    gH := hLeft;
    gV := vLeft;
    gN := nLeft;
    if ComputeParamSurface then
      Result := True;
    uc0 := prevC0;
    upcx := Round(prevPcx);
    gH := prevH;
    gV := prevV;
    gN := prevN;
    gViewpoint := prevWp;
  end;
end;

function LocalPickedPtImplicit(pix: TPoint; var pos: Vector3): Boolean;
var
  a, b, root, ah, av: double;
  ray: Vector3;
  k: Integer;
begin
  Result := False;
  ah := pix.x - upcx;
  av := pix.y - upcy;
  for k := 1 to 3 do
    ray[k] := ah * gH[k] + av * gV[k] + uC0 * gN[k];
  SegmentInCube(gViewpoint, ray, gVpCubeKey, a, b);
  if a < b then begin
    if FirstZero(gViewpoint, ray, a, b, root) then begin
      for k := 1 to 3 do
        pos[k] := gViewpoint[k] + root * ray[k];
      Result := True;
    end;
  end;
end;




function LocalReflectedLight(ptSurface: TPoint; var anver: integer; var light:
  double): Boolean;
const
  dullness: double = 0.3;
  OneMinusDullness: double = 1 - 0.3;
  FocusIntens: double = 0.60;
  OneMinusFocusIntens: double = 0.4;

var
  lambda, sigma, omega, compF, compAmb, f2x, m2x: double;
  normal, thePos, toVp, toFocus: Vector3;
  i: Integer;
  visFromFocus, goodPt: Boolean;
begin
  goodPt := LocalPickedPtImplicit(ptSurface, thePos);
  result := goodPt;
  if goodPt then begin
    for i := 1 to 3 do begin
      toVp[i] := gViewpoint[i] - thePos[i];
      toFocus[i] := gFocus[i] - thePos[i];
    end;
    m2x := EuclNorm3(toVp);
    normal := ComputeNormal(thePos);
    sigma := EuclInnerProd3(toVp, Normal) / m2x;
    compAmb := (1 + abs(sigma)) / 2;
    visFromFocus := IsVisibleImplicit(gFocus, thePos, gFCubeKey, TRUE);
    if visFromFocus then begin
      f2x := EuclNorm3(toFocus);
      lambda := EuclInnerProd3(toFocus, normal) / f2x;
      omega := 2 * lambda * sigma;
      visFromFocus := (omega > 0);
    end;
    if visFromFocus then begin
      lambda := abs(lambda);
      omega := (1 + omega - EuclInnerProd3(toFocus, toVp) / (m2x * f2x)) /
        2;
          //omega := sqr(sqr(omega));
      compF := {lambda * dullness + OneMinusDullness * omega;} lambda *
        (OneMinusDullness
        * sqr(omega) + dullness * (1 +
        abs(sigma)) / 2);
    end
    else
      compF := 0;
    light := Min(0.99, FocusIntens * compF + OneMinusFocusIntens * compAmb);
    if (sigma > 0) then
      anver := 0
    else
      anver := 1;
  end; {de IF goodPt}
end;

function ComputeImplicitSurface: Boolean;
const
  maxFocus: double = 5.0;
var
  j, k: integer;
  thePos: TPoint;
  light: double;
  index: byte;
  anver: integer;
  line: pTColorArray;
begin
  result := False;
  for k := 0 to prHeight - 1 do begin
    if (k mod 3 = 0) then
      Application.ProcessMessages;
    if gStop then
      Exit;
    thePos.y := k;
    line := gForPrintingBmp.ScanLine[k];
    for j := 0 to prWidth - 1 do begin
      thePos.x := j;
      if LocalReflectedLight(thePos, anver, light) then begin
        index := round(99.7 * light);
        if (anver = 0) then
          index := 101 + index
        else
          index := abs(index) + 1;
        line^[j] := gPalette3D^[index];
      end;
    end;
  end;
  result := True;
end;




function ComputeNormalPairImplicit: Boolean;
const
  EySin: double = 1 / 15;
  EyCos: double = 0.9977753031;
var
  nLeft, hLeft, vLeft: Vector3;
  j, k, anver, half, index: integer;
  thePos: TPoint;
  light, d: double;
  prevC0, prevC1, prevC2, prevC3: double;
  prevPcx: double;
  cen, prevH, prevV, prevN, prevVp: Vector3;
  line: pTColorArray;

  procedure ResetParameters;
  begin
    uC0 := prevC0;
    gC1 := prevC1;
    gC2 := prevC2;
    gC3 := prevC3;
    upcx := prevPcx;
    gViewPoint := prevVp;
    gH := prevH;
    gV := prevV;
    gN := prevN;
  end;
begin
  result := False;
  prevC0 := uC0;
  uC0 := prevC0 / 2;
  prevPcx := upcx;
  upcx := prevPcx / 2;
  cen := gCenter3D;
  d := EuclDistance3(gViewpoint, cen);
  vLeft := gV;
  for j := 1 to 3 do begin
    nLeft[j] := EyCos * gN[j] + EySin * gH[j];
    hLeft[j] := -EySin * gN[j] + EyCos * gH[j];
  end;
  half := prWidth div 2;
  for k := 0 to prHeight - 1 do begin
    if (k mod 3 = 0) then Application.ProcessMessages;
    if gStop then
      Exit;

    thePos.y := k;
    line := gForPrintingBmp.ScanLine[k];
    for j := 0 to half - 1 do begin
      thePos.x := j;
      if LocalReflectedLight(thePos, anver, light) then begin
        index := round(99.7 * light);
        if (anver = 0) then
          index := 101 + index
        else
          index := abs(index) + 1;
        line^[j] := gPalette3D^[index];
      end;
    end;
  end;


  prevC1 := gC1;
  prevC2 := gC2;
  prevC3 := gC3;
  gC3 := EuclInnerProd3(nLeft, cen) - d;
  gC1 := EuclInnerProd3(hLeft, cen);
  gC2 := EuclInnerProd3(vLeft, cen);
  prevVp := gViewPoint;
  for j := 1 to 3 do
    gViewPoint[j] := gC3 * nLeft[j] + gC1 * hLeft[j] + gC2 * vLeft[j];
  upcx := upcx + prWidth div 2;
  prevH := gH;
  prevN := gN;
  prevV := gV;
  gH := hLeft;
  gV := vLeft;
  gN := nLeft;

  for k := 0 to prHeight - 1 do begin
    if (k mod 3 = 0) then Application.ProcessMessages;
    if gStop then begin
      ResetParameters;
      Exit;
    end;
    thePos.y := k;
    line := gForPrintingBmp.ScanLine[k];
    for j := half + 1 to prWidth - 1 do begin
      thePos.x := j;
      if LocalReflectedLight(thePos, anver, light) then begin
        index := round(99.7 * light);
        if (anver = 0) then
          index := 101 + index
        else
          index := abs(index) + 1;
        line^[j] := gPalette3D^[index];
      end;
    end;
  end;
  ResetParameters;
  result := True;
end;

function uDrawRiemannianCurvature: Boolean;
var
  i, j: Integer;
  x: vector2;
  K, udp1, udp2, udp3, udp4: double;
  index: Byte;
  metricType: Integer;
  goodPt: Boolean;
  line: pTColorArray;
begin
  Result := False;
  K := (kWidth - 1) / (prWidth - 1);
  udp1 := dp1 * K;
  udp2 := dp2;
  udp3 := dp3 * K;
  udp4 := dp4;
  for j := 0 to prHeight - 1 do begin
    if (j mod 30 = 0) then Application.ProcessMessages;
    if gStop then
      Exit;
    line := gForPrintingBmp.ScanLine[j];
    x[2] := udp3 * j + udp4;
    if (abs(x[2] - gCenter2D[2]) > gHalfEdges2D[2]) then
      continue;
    for i := 0 to prWidth - 1 do begin
      x[1] := udp1 * i + udp2;
      goodPt := (abs(x[1] - gCenter2D[1]) <=
        gHalfEdges2D[1]);
      if goodPt then
        metricType := ComputePseudoRiemCurvature(x, K);
      if goodPt and (metricType <> 0) then begin
        if metricType = kRiemMetric then
          index := trunc(1 + 50 * (1 - K / (1 + abs(K))))
        else
          index := trunc(101 + 50 * (1 - K / (1 + abs(K))));
        line^[i] := gPalette2D^[index];
      end;
    end;
  end;
  Result := True;
end;

function DrawWithHighResolution: Boolean;
begin
  result := False;
  prWidth := gResolution * kWidth;
  prHeight := gResolution * kHeight;
  PreparegForPrintingBmp;
  uc0 := gc0 * gResolution;
  upcx := gpcx * gResolution;
  upcy := gpcy * gResolution;
  uFC0 := gFC0 * gResolution;
  if gType = Implicit then begin
    if gMode = ColoredStereoPair then
      result := ComputeNormalPairImplicit
    else
      result := ComputeImplicitSurface
  end
  else if gType = PseudoRiemannian then
    result := uDrawRiemannianCurvature
  else begin
    if gType = Parametric then begin
      uGrid := 127;
      vGrid := 127;
    end
    else if gType = Tube then begin
      uGrid := kTubeGridU;
      vGrid := 51;
    end;
    uStep2D[1] := gEdges2D[1] / uGrid;
    uStep2D[2] := gEdges2D[2] / vGrid;
    if not InitializeBigVars then Exit;
    if gMode = ColoredStereoPair then
      result := ComputeNormalPairParam
    else
      result := ComputeParamSurface;
  end;
end;

procedure Finalize;
var
  i: Integer;
begin
  for i := 1 to 3 do
    cInitVal[i] := nil;
  cVal := nil;
  cValDerivs := nil;
  if zBuffer <> nil then begin
    for i := 0 to High(zBuffer) do
      zBuffer[i] := nil;
    zBuffer := nil;
  end;
  if fBuffer <> nil then begin
    for i := 0 to High(fBuffer) do
      fBuffer[i] := nil;
    fBuffer := nil;
  end;
  if uPointsData <> nil then begin
    for i := 0 to High(uPointsData) do
      uPointsData[i] := nil;
    uPointsData := nil;
  end;
  if uFocusPtData <> nil then begin
    for i := 0 to High(uFocusPtData) do
      uFocusPtData[i] := nil;
    uFocusPtData := nil;
  end;
end;


initialization
  ;
finalization
  Finalize;


end.
