

{-----------------------------------------------------------------------------
 Unit Name: Globals
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose:  Constants, Types and Variables that are used along the project
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit Globals;
interface
uses
  Classes, SysUtils
{$IFDEF MSWINDOWS}
  , Windows, Forms, Graphics
{$ENDIF}
{$IFDEF LINUX}
  , Types, QGraphics,  QForms, QTypes
{$ENDIF}
    ;

const
{For identifying files}
  kFilter = 'gtv';
  kCreator = 'AMonAmi';
  kCreatorLength = Length(kCreator);
  kFileTypeParam = 'ParametrSurface';
  kFileTypeImplicit = 'ImplicitSurface';
  kFileTypeTube = 'TubeAroundCurve';
  kFileTypeRiem = 'PseudoRiemPlane';
  kTypeLength = Length(kFileTypeParam);
    {the length of any of the File types. ALL
                                        MUST BE THE SAME}
  kVersion: Integer = 221001;

  kNumOfSavedToDiskStrings = 10;

  pi: double = 3.1415926535897932384;
  kThreeOver2 = 3 / 2;
  kTwoOver3 = 2 / 3;
  kFourOver3 = 4 / 3;
  kOneThird = 1 / 3;
  kOneSixth = 1 / 6;
  kTol = 1.0E-5;
  kWidth = 400; {of the drawing panels}
  kHeight = 400; {id}
  kWidthMinus1 = kWidth - 1;
  kHeightMinus1 = kHeight - 1;
  kHalfWidth = kWidth div 2;
  kHalfHeight = kHeight div 2;

  kTubeGridU = 125;
  kTubeGridV = 31;

  kPixelsArrow = 25;
  kPPLine2D = 40;
  kMaxPPLine2D = 50;
  kMaxNumFlags = 4 * kMaxPPLine2D;

  {classes of lines}

  kCoordinates = 0;
  kConstantCurvatureLorentz = kCoordinates;

  kCurvature = 1;
  kConstantCurvatureRiem = kCurvature;

  kAsymptotics = 2;
  kIntrinsicLorentz = kAsymptotics;
  kIntrinsic = 3;
  kIntrinsicRiem = kIntrinsic;
  kGeodBall = 4;
  kLightRays = 5;
  kTransport = 6;
  kArrows = 7;

  kWiresPerLine = 15;
  kSegmentsAlongLine = 30;

  {TColors}
  cBlack: TColor = $00000000;
  cWhite: TColor = $00FFFFFF;
  cGreen: TColor = $0000FF00;
  cBlue: TColor = $00FF0000;
  cRed: TColor = $000000FF;
  cTeal: TColor = $00FFFF00;
  cBloodRed: TColor = $000000D0;
  cLightGreen: TColor = $0038FF38;
  cEyeRed: TColor = $000000FF;
  cPrinterRed: TColor = $007070FF;
  cEyeBlue: TColor = $00B0B000;
  cMedGreen: TColor = $0050FF40;
  cOrange: TColor = $0020B0FF;
  cBaseGoldenPalette: TColor = 13408;
  cBaseGreenPalette: TColor = 3166730;

  {flags}

  kRiemMetric = 1;

  {miscellaneous}
  kReturn = char(13);
  kLineFeed = char(10);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';
{$IFDEF MSWINDOWS}
  kBullet = chr(149);
{$ENDIF}
{$IFDEF LINUX}
  kBullet = chr(164);
{$ENDIF}

type

  SurfaceType = (Parametric, Implicit, PseudoRiemannian, Tube);
  Mode = (Standard, ColoredStereoPair, Movable, WireframeMono,
    WireframeStereo,
    WireframeStereoPair);
  GeomOption = (CoordLines, LightRays, Geodesics, ExponentialMap,
    GeodBetweenPts,
    GeodesicBall, LinesOfConstGeodCurv, ParallelTransport, CurvatureLines,
    AsymptoticLines, Clipping, NormalSection, Nothing);

  ColorMap = packed array[0..kWidthMinus1, 0..kHeightMinus1] of Byte;
  ColorMapPtr = ^ColorMap;

  Vector2 = array[1..2] of Double;
  Vector3 = array[1..3] of Double;
  Vector3Single = array[1..3] of Single;
  Vector2Single = array[1..2] of Single;
  Segment2 = array[1..2] of Vector2;
  Segment3 = array[1..2] of Vector3;
  Segm2Single = array[1..2] of Vector2Single;
  Segm3Single = array[1..3] of Vector2Single;
  Matrix3 = array[1..3] of Vector3;
  Matrix3Single = array[1..3] of Vector3Single;

  PointData = record
    pos, proj: Vector3Single;
    ambLight: Single;
    inside, outside: Boolean;
  end;

  FocusSpacePtData = record
    pt: Vector3Single;
    focLight: Single;
  end;

  PointDataDeposit = array of array of PointData;
  FocusSpacePtDataDeposit = array of array of FocusSpacePtData;

  DataBuffer = array[0..kWidthMinus1, 0..kHeightMinus1] of single;
  DataBufferPtr = ^dataBuffer;
  ParamDataBuffer = array[0..kWidthMinus1, 0..kHeightMinus1] of Vector2Single;
  ParamDataBufferPtr = ^ParamDataBuffer;

  Clip = record
    u: Vector3;
    a: Double;
  end;

  CubeKey = array[1..3] of 0..2;

  Node3D = record
    key: Integer;
    pos: Vector3Single;
  end;

  Wire3D = array of Node3D;

  SegCol3D = record
    color: Integer;
    seg: Segm2Single;
  end;

  Cube = array of TPoint;

  TColorArray = array[0..kWidth] of TColor;
  pTColorArray = ^TColorArray;
  Pal200 = array[0..200] of TColor;
  Pal200Ptr = ^Pal200;

  gamma16 = array[1..16] of TColor;
  gamma16Ptr = ^gamma16;

  LinesData = record
    toPrint, noMoreRoom: Boolean;
    numSegm2D, numSegm3D, maxSegm: Integer;
    text: AnsiString;
    pen2DColor: TColor;
    pen2DWidth: Integer;
    gammaP: gamma16Ptr;
  end;

  Lines2D = array of Segm2Single;
  Lines3D = array of SegCol3D;

  FPar = procedure(var result: Vector3; u, uve: double);
  FPar2 = procedure(var resul: Vector3; aqui: vector2);

  DataToSave = record
    gViewPoint, gH, gV, gN: Vector3;
    gC0, gC1, gC2, gC3: Double;
    gpcx, gpcy: Integer;
    gVPCubeKey: CubeKey;
  end;

  EditFStrings = array[1..kNumOfSavedToDiskStrings] of AnsiString;


var
  gRect: TRect = (left: 0; top: 0; right: kWidth; bottom: kHeight);
  gXOptions: set of GeomOption;
  gSecondOrderOptions: set of GeomOption;
  gWireModes: set of Mode;
  gStdBmp, gMovingBmp, gForPrintingBmp: TBitmap;
  gMap, gNormalPairMap: ColorMapPtr;
  theFPar: FPar;
  theFPar2: FPar2;

  gPointsData: PointDataDeposit; {for the grid to draw a parametric surface}
  gFocusPtData: FocusSpacePtDataDeposit; {the same for the Focus viewpoint}
  gZBuffer, gFBuffer: DataBufferPtr; {for the viewPoint and the Focus}
  gParamBuffer: ParamDataBufferPtr;
    {for the projection of the grid points with}
                                     {single precision.}
  gVersion: Integer;
  gType: SurfaceType;
  gPathAndName: TFilename;
  gTitle: AnsiString;

  gSurfaceStr: EditFStrings;
  gSurfaceStrLengths: array[1..kNumOfSavedToDiskStrings] of Integer;
  gStrInitParametric, gStrInitTube, gStrInitImplicit,
    gStrInitPseudoRiemannian:
  EditFStrings;

  gGridU, gGridV: Integer;
    {Grid points per line for drawing the parametric surface.}
  gStep2D: Vector2; {for the grid points}
  gCenter3D, gEdges3D, gHalfEdges3D, {of the cube that contains the surface;}
    gMin3D, gMax3D, {cube corners of least and greatest coordinates.}
    gIncr3D, {increments to compute derivatives.}
    gViewPoint, {from which one sees the surface in space}
    gFocus {the point where is the focus that illuminates the surface}
    : Vector3;
  gSpan3D, {diagonal of the cube}
    gRadius3D,
    gMaxSegmentLength3D,
    gTolDistance,
    gVerySmall {used in 'FirstZero'}
    : Double;

  dp1, dp2, dp3, dp4, {Coefficients to pass from Drawing2D coordinates to }
                        {parameter space coordinates}
  pd1, pd2, pd3, pd4 {Coefficients to pass from parameter space coordinates}
                        {to Drawing2D.}
  : Double;

  gH, gV, gN: Vector3; {For the projection to Drawing3D}
  gC0, gC1, gC2, gC3: Double; {Coefficientes for the projection to Drawing3D}
  gpcx, gpcY: Integer;
    {Drawing3D coordinates of the projection of the surface center}
  gDist: Double; {The distance from gViewpoint to gCenter3D}
  gFH, gFV, gFN: Vector3;
    {For the projection to the plane orthogonal to the focus}
  gFC0, gFC1, gFC2, gFC3: Double; {id, id}
  gVPCubeKey, gFCubeKey: CubeKey;

  gCenter2D, gEdges2D, gHalfEdges2D, {of the rectangle of parameters u and v}
    gMin2D, gMax2D, {rectangle corners of least and greatest coordinates.}
    gIncr2D, g2Incr2D, g24Incr2D, g24SqrIncr2D, g4Incr2D, g2SqrIncr2D
    {increments to compute derivatives.}
  : Vector2;
  g4Incr1Incr2: Double; {An pre-computed auxiliar number}

  gSpan2D, {Diagonal of the rectangle}
    gRadius2D, {distante equivalent to one half pixel in the space of parameters}
    gMaxSegmentLength2D {max parameter distance spanned by an onscreen-drawn segment}
    : Double;

  gEpsilon: Integer;
    {+1 if the space is Euclidean; -1 if the space is Minkowskian}

  gNewClip: Clip;
  gClips: array of Clip;
  gClipTop: Integer;
  gClipCut: Wire3D;
  gCCParams: array of vector2;
  gLastSegmCC: Integer;

  gWire: Wire3D;
  gNumNodes3D: Integer;
  gNoMemoryForWire: Boolean;

  gPPLine2D, gNumFlags: Integer; {for the computation of implicit curves}

  gCube: array of Segment2 {Cube};
  gCubeLast: Integer;

  gPalette3D, gPalette2D: Pal200Ptr;
  gammaGrey, gammaBlue, gammaPurple, gammaRed, gammaGreen: gamma16Ptr;

  gLines2D: array[kCoordinates..kArrows] of Lines2D;
  gLines3D: array[kCoordinates..kArrows] of Lines3D;
  gLinesData: array[kCoordinates..kArrows] of LinesData;
  gActiveLine: Integer;
  gPrintLines, gPrintSurface: Boolean;
  gResolution: Integer; {Resolution for the printed page}

  gGrain: Integer; {Estimated grain for moving an implicit surface}


  {flags}
  gAlreadyDrawn, {the desired surface has been drawn succesfully }
    gGotProjParams, {The projection parameters gC0, etc. are already set}
    gStop, gBusy, gSaved: Boolean;
  gThereIsX2D, gThereIsX3D, gThereIsCross2D, gThereIsCross3D: Boolean;
  gX2D, gCross2D, gX3D, gCross3D: array[0..4] of TPoint;
  gUnitConic: array[0..84] of TPoint;
  gThereIsUnitConic: Integer;
    {0, there is not; 1 there is a Ellipse; 2, there are two hyperbolas}

  gKStr, gHStr, gK1Str, gK2Str: AnsiString;

  //Preferences flags
  gShowUnitConic, gShowCube: Boolean;
  gSideToDrawOn: Integer; {If it is 2, both sides; if 0, the normal side;}
                             {if 1, the other side.}
  gMode: Mode;
  gGeomOption: GeomOption;
  gMeshNumber: Integer;

  gFirstPtDrawing2D, gFirstPtDrawing3D, gLastPtDrawing2D, gLastPtDrawing3D,
    gCurPtDrawing2D, gCurPtDrawing3D: TPoint;
  gFirst2DPos, gLast2DPos, gCur2DPos: Vector2;
  gFirst3DPos, gLast3DPos, gCur3DPos, gVel: Vector3;
  gFirstPtGood, gCurPtGood: Boolean;
  gTracking: Boolean;



procedure ChangeGamma(ini: Integer; c: TColor);
procedure SetWiresData;
procedure InitNewSurfaceVars;

implementation
uses
   Math,
{$IFDEF MSWINDOWS}
   Dialogs,
{$ENDIF}
{$IFDEF LINUX}
   QDialogs,
{$ENDIF}
   Utilities, ResStrings;


function CreateVars: Boolean;
begin
  Result := False;
  New(gMap);
  if gMap = nil then
    Exit;
  New(gNormalPairMap);
  if gNormalPairMap = nil then
    Exit;
  New(gZBuffer);
  if gZbuffer = nil then
    Exit;
  New(gFBuffer);
  if gFbuffer = nil then
    Exit;
  New(gParamBuffer);
  if gParambuffer = nil then
    Exit;
  gStdBmp := TBitmap.Create;
  gStdBmp.Height := kHeight;
  gStdBmp.Width := kWidth;
  gStdBmp.PixelFormat := pf32bit;
  gMovingBmp := TBitmap.Create;
  gMovingBmp.Height := kHeight;
  gMovingBmp.Width := kWidth;
  gMovingBmp.PixelFormat := pf32bit;
  gMovingBmp.Canvas.Brush.Color := clBlack;
  gMovingBmp.Canvas.Brush.Style := bsSolid;
  gForPrintingBmp := TBitmap.Create;
  gForPrintingBmp.Height := kHeight;
  gForPrintingBmp.Width := kWidth;
  gForPrintingBmp.PixelFormat := pf32bit;
  gForPrintingBmp.Canvas.Brush.Color := clWhite;
  gForPrintingBmp.Canvas.Brush.Style := bsSolid;

  New(gammaGrey);
  if gammaGrey = nil then
    Exit;
  New(gammaBlue);
  if gammaBlue = nil then
    Exit;
  New(gammaPurple);
  if gammaPurple = nil then
    Exit;
  New(gammaRed);
  if gammaRed = nil then
    Exit;
  New(gammaGreen);
  if gammaGreen = nil then
    Exit;
  New(gPalette3D);
  if gPalette3D = nil then
    Exit;
  New(gPalette2D);
  if gPalette2D = nil then
    Exit;
  Result := True;
end;

procedure InitFixedVars;
begin
  gXOptions := [LightRays, CurvatureLines, AsymptoticLines];
  gSecondOrderOptions := [Geodesics, ExponentialMap, GeodBetweenPts,
    GeodesicBall,
    LinesOfConstGeodCurv];
  gWireModes := [WireframeMono, WireframeStereo, WireframeStereoPair];
end;


procedure InitNewSurfaceVars;
begin
  gVersion := kVersion;
  gBusy := False;
  gEpsilon := 1;
  gCube := nil;
  gShowUnitConic := False;
  gShowCube := False;
  gSideToDrawOn := 2;
  gMode := Standard;
  gGeomOption := Nothing;
  gActiveLine := kCoordinates;
  gMeshNumber := 16;
  gStop := False;
  gAlreadyDrawn := False;
  gThereIsX2D := False;
  gThereIsX3D := False;
  gThereIsCross2D := False;
  gThereIsCross3D := False;
  gThereIsUnitConic := 0;
  gFirstPtGood := False;
  gCurPtGood := False;
  gTracking := False;
  InitWire;
  InitSavedLines;
  gMovingBmp.Canvas.FillRect(gRect);
  gMovingBmp.Canvas.Brush.Color := clWhite;
  gStdBmp.Canvas.FillRect(gRect);
end;

procedure InitFormStrings;
begin
  gStrInitParametric[1] := '2 cos(u) sin(v)';
  gStrInitParametric[2] := '3 sin(u) sin(v)';
  gStrInitParametric[3] := 'cos(v)';
  gStrInitParametric[4] := '';
  gStrInitParametric[5] := '0';
  gStrInitParametric[6] := '2 pi';
  gStrInitParametric[7] := '0';
  gStrInitParametric[8] := 'pi';
  gStrInitParametric[9] := '';
  gStrInitParametric[10] := '60';

  gStrInitTube[1] := 'cos(t/2)';
  gStrInitTube[2] := 'sin(2 t)';
  gStrInitTube[3] := 'sin(3 t)';
  gStrInitTube[4] := '(2 + cos(t))/9';
  gStrInitTube[5] := '0';
  gStrInitTube[6] := '2 pi';
  gStrInitTube[7] := '';
  gStrInitTube[8] := '';
  gStrInitTube[9] := '';
  gStrInitTube[10] := '';


  gStrInitImplicit[1] := '(sqrt(2 x^2+y^2)-2)^2+ z^2-1';
  gStrInitImplicit[2] := '';
  gStrInitImplicit[3] := '';
  gStrInitImplicit[4] := '';
  gStrInitImplicit[5] := '-3';
  gStrInitImplicit[6] := '3';
  gStrInitImplicit[7] := '-3';
  gStrInitImplicit[8] := '3';
  gStrInitImplicit[9] := '-1';
  gStrInitImplicit[10] := '1';

  gStrInitPseudoRiemannian[1] := '1+ cos(x)^2';
  gStrInitPseudoRiemannian[2] := 'sin(x-y)';
  gStrInitPseudoRiemannian[3] := '1 + sin(y+x)';
  gStrInitPseudoRiemannian[4] := '';
  gStrInitPseudoRiemannian[5] := '- pi';
  gStrInitPseudoRiemannian[6] := 'pi';
  gStrInitPseudoRiemannian[7] := '0';
  gStrInitPseudoRiemannian[8] := '2 pi';
  gStrInitPseudoRiemannian[9] := '';
  gStrInitPseudoRiemannian[10] := '';
end;

procedure BuildGamma16(r, g, b: Integer; var gamut: gamma16Ptr);
const
  MM = 65535;
  num = 16;
var
  i: Integer;
  redComp, greenComp, blueComp: Integer;
  m, i0, pr, pg, pb, kr, kg, kb, aux, factor: double;
begin
  m := r;
  if g > m then
    m := g;
  if b > m then
    m := b;
  aux := 1;
  if m * num < MM then
    aux := 1.1 * MM / (m * num);
  i0 := (m * aux * num - MM) / (MM - aux * m);
  kr := r * aux / (1 + i0);
  kg := g * aux / (1 + i0);
  kb := b * aux / (1 + i0);
  factor := exp(5 * ln(num - 1));
  pr := (MM - kr * (num + i0)) / factor;
  pg := (MM - kg * (num + i0)) / factor;
  pb := (MM - kb * (num + i0)) / factor;
  for i := 1 to num do begin
    if i = 1 then
      factor := 0
    else
      factor := exp(0.97 * 5 * ln(i - 1));
    greenComp := round(kg * (i + i0) + pg * factor);
    blueComp := round(kb * (i + i0) + pb * factor);
    redComp := round(kr * (i + i0) + pr * factor);
    gamut^[i] := redComp div 256 + ((greenComp div 256) shl 8) + ((blueComp
      div
      256)
      shl 16);
  end;
end;

procedure BuildGamma200(initial: Integer;
  r, g, b, pow: Integer; var theGamut: Pal200Ptr);
const
  MM = 65533;
  num = 100;
var
  i, nr, ng, nb: Integer;
  redComp, greenComp, blueComp: Integer;
  m, i0, pr, pg, pb, kr, kg, kb, aux, factor: double;
begin
  nr := Min(MM, r);
  ng := Min(MM, g);
  nb := Min(MM, b);
  nr := Max(1, nr);
  ng := Max(1, ng);
  nb := Max(1, nb);
  m := nr;
  if ng > m then
    m := ng;
  if nb > m then
    m := nb;
  aux := (MM - m) / (m * (num - 1));
  i0 := initial - 1 / aux;
  kr := aux * nr;
  kg := aux * ng;
  kb := aux * nb;
  pr := MM * (m - nr) / m;
  pg := MM * (m - ng) / m;
  pb := MM * (m - nb) / m;
  greenComp := ng;
  blueComp := nb;
  redComp := nr;
  theGamut^[initial] := blueComp div 256 + ((greenComp div 256) shl 8) +
    ((redComp div 256)
    shl 16);
  if num > 1 then
    for i := initial + 1 to initial + num - 1 do begin
      aux := i - i0;
      factor := exp(pow * ln((i - initial) / (num - 1)));
      greenComp := round(kg * aux + pg * factor);
      blueComp := round(kb * aux + pb * factor);
      redComp := round(kr * aux + pr * factor);
      theGamut^[i] := blueComp div 256 + ((greenComp div 256) shl 8) +
        ((redComp div 256)
        shl 16);
    end;
end;

procedure BuildPseudoRiemGamut200;
var
  redComp, greenComp, blueComp: Integer;
  i: Integer;
begin
  greenComp := $FFFF;
  blueComp := $FFFF;
  redComp := $FFFF;
  gPalette2D^[0] := blueComp div 256 + ((greenComp div 256) shl 8) + ((redComp
    div 256)
    shl 16); {cBlanco}
  for i := 1 to 50 do begin
    greenComp := round(1310 * i);
    redComp := $FFFF;
    blueComp := greenComp;
    gPalette2D^[i] := blueComp div 256 + ((greenComp div 256) shl 8) +
      ((redComp div 256)
      shl 16); {gama rojo a blanco}
  end;
  for i := 1 to 50 do begin
    greenComp := round(65600 - 118 * i);
    redComp := round(65600 - 300 * i);
    blueComp := round(66600 - i * 1289);
    gPalette2D^[i + 50] := blueComp div 256 + ((greenComp div 256) shl 8) +
      ((redComp div 256)
      shl 16); {gama blanco a amarillo}
  end;
  for i := 1 to 50 do begin
    greenComp := round(1300 * i);
    redComp := greenComp;
    blueComp := $FFFF;
    gPalette2D^[i + 100] := blueComp div 256 + ((greenComp div 256) shl 8) +
      ((redComp div 256)
      shl 16); {gama azul a blanco}
  end;
  for i := 1 to 50 do begin
    greenComp := round(66600 - i * 1289);
    redComp := $FFFF;
    blueComp := $FFFF;
    gPalette2D^[i + 150] := blueComp div 256 + ((greenComp div 256) shl 8) +
      ((redComp div 256)
      shl 16); {gama blanco a lila}
  end;
end;

procedure SetWiresData;
var
  i: Integer;
begin
  for i := kCoordinates to kArrows do begin
    gLinesData[i].noMoreRoom := False;
    gLinesData[i].pen2DWidth := kArrows + 1 - i;
  end;
  gLinesData[kArrows].pen2DColor := gammaRed^[4];
  gLinesData[kArrows].text := '';
  gLinesData[kArrows].gammaP := gammaRed;
  gLinesData[kArrows].maxSegm := 1000;
  if gType = PseudoRiemannian then begin
    gLinesData[kConstantCurvatureLorentz].pen2DColor := cBloodRed;
    gLinesData[kConstantCurvatureLorentz].maxSegm := 5000;
    gLinesData[kConstantCurvatureRiem].pen2DColor := clOlive;
    gLinesData[kConstantCurvatureRiem].maxSegm := 25000;
    gLinesData[kIntrinsicLorentz].pen2DColor := cTeal;
    gLinesData[kIntrinsicLorentz].maxSegm := 5000;
    gLinesData[kIntrinsicRiem].pen2DColor := cBlue;
    gLinesData[kIntrinsicRiem].maxSEgm := 25000;
    gLinesData[kGeodBall].pen2DColor := clGray;
    gLinesData[kGeodBall].maxSegm := 10000;
    gLinesData[kLightRays].pen2DColor := cLightGreen;
    gLinesData[kLightRays].maxSegm := 10000;
    gLinesData[kTransport].pen2DColor := cGreen;
    gLinesData[kTransport].maxSegm := 10000;
    gLinesData[kConstantCurvatureLorentz].text := eCGC;
    gLinesData[kConstantCurvatureRiem].text := eCGC;
    gLinesData[kIntrinsicLorentz].text := eGeodesics;
    gLinesData[kIntrinsicRiem].text := eGeodesics;
    gLinesData[kGeodBall].text := eBG;
    gLinesData[kLightRays].text := eLightRays;
    gLinesData[kTransport].text := eTP;
  end
  else begin
    gLinesData[kCoordinates].pen2DColor := clSilver;
    gLinesData[kCoordinates].maxSegm := 20000;
    gLinesData[kCurvature].pen2DColor := clRed;
    gLinesData[kCurvature].maxSegm := 25000;
    gLinesData[kAsymptotics].pen2DColor := clBlue;
    gLinesData[kAsymptotics].maxSegm := 10000;
    gLinesData[kIntrinsic].pen2DColor := clFuchsia;
    gLinesData[kIntrinsic].maxSegm := 30000;
    gLinesData[kGeodBall].pen2DColor := clGray;
    gLinesData[kGeodBall].maxSegm := 10000;
    gLinesData[kLightRays].pen2DColor := clSilver;
    gLinesData[kLightRays].maxSegm := 5000;
    gLinesData[kTransport].pen2DColor := cGreen;
    gLinesData[kTransport].maxSegm := 10000;

    gLinesData[kCoordinates].text := eTCoord;
    gLinesData[kCurvature].text := eTCurv;
    gLinesData[kAsymptotics].text := eTAsint;
    gLinesData[kIntrinsic].text := eTIntrins;
    gLinesData[kGeodBall].text := eBG;
    gLinesData[kLightRays].text := eLightRays;
    gLinesData[kTransport].text := eTP;

    gLinesData[kCoordinates].gammaP := gammaGrey;
    gLinesData[kCurvature].gammaP := gammaRed;
    gLinesData[kAsymptotics].gammaP := gammaBlue;
    gLinesData[kIntrinsic].gammaP := gammaPurple;
    gLinesData[kGeodBall].gammaP := gammaGrey;
    gLinesData[kLightRays].gammaP := gammaGrey;
    gLinesData[kTransport].gammaP := gammaGreen;
  end;
end;

procedure BuildPalettes;
begin
  gPalette3D^[0] := $00FFFFFF;
  BuildGamma200(1, 24600, 13320, 10, 6, gPalette3D);
  BuildGamma200(101, 2700, 21000, 12400, 6, gPalette3D);
  BuildGamma16(30000, 30000, 30000, gammaGrey);
  BuildGamma16(0, 8000, 22000, gammaBlue);
  BuildGamma16(22000, 0, 22000, gammaPurple);
  BuildGamma16(22000, 0, 0, gammaRed);
  BuildGamma16(0, 22000, 0, gammaGreen);
  BuildPseudoRiemGamut200;
  SetWiresData;
end;

procedure ChangeGamma(ini: Integer; c: TColor);
var
  r, g, b: Integer;
begin
  b := ((c shr 16) mod 256) shl 8;
  g := ((c shr 8) mod 256) shl 8;
  r := (c mod 256) shl 8;
  BuildGamma200(ini, r, g, b, 5, gPalette3D);
end;

procedure initialize;
begin
  if not CreateVars then
    Application.Terminate;
  gType := Parametric;
  InitFixedVars;
  InitNewSurfaceVars;
  InitFormStrings;
  BuildPalettes;
  {$IFDEF MSWINDOWS}
    ForceCurrentDirectory := True;
  {$ENDIF}
end;

procedure Finalize;
begin
  gStdBmp.Free;
  gMovingBmp.Free;
  gForPrintingBmp.Free;
  if gMap <> nil then
    Dispose(gMap);
  if gNormalPairMap <> nil then
    Dispose(gNormalPairMap);
  if gZBuffer <> nil then
    Dispose(gZBuffer);
  if gFBuffer <> nil then
    Dispose(gFBuffer);
  if gParamBuffer <> nil then
    Dispose(gParamBuffer);
  gPointsData := nil;
  gFocusPtData := nil;
  if gPalette3D <> nil then
    Dispose(gPalette3D);
  if gPalette2D <> nil then
    Dispose(gPalette2D);
  gClips := nil;
end;

initialization
  initialize;

finalization
  finalize;
end.
