

{-----------------------------------------------------------------------------
 Unit Name: ImplicitUtils
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Utilities specific to surfaces given in implicit form
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}




unit ImplicitUtils;
interface
uses
{$IFDEF MSWINDOWS}
  Windows, Graphics,
{$ENDIF}
{$IFDEF LINUX}
  Types, QGraphics,
{$ENDIF}
  Globals;

procedure PartialsOfFImplicitSurface(position: Vector3; var Hessian: Matrix3;
  var
  grad: Vector3);
procedure ComputeCurvaturesImplicitSurface(pos: Vector3);
function ComputeVelImplicitSurface(var velocity: Vector3; normalized:
  Boolean): Boolean;
function ComputeClipImplicitSurface(var theClip: Clip):
  Boolean;
procedure AddClipImplicitSurface(pos, vel: Vector3);
function InRangeAndInClipsImplicitSurface(var depth: Double; x: Vector3):
  Boolean;
function InRangeAndInClipsImplicitSurfaceS(var depth: double; x:
  Vector3Single):
  Boolean;
procedure RecordAndDrawSegmentImplicitSurface(x1, x2: Vector3; toSave:
  Boolean);

function CheckRangeImplicit(lastX, nextX: Vector3; findProlong: Boolean;
  var newX, newVel: Vector3): Integer;
function CheckCameBackImplicitSurface(firstX, prevX, curX:
  Vector3; counter: Integer): Boolean;
function CheckHandImplicit(x, t, v: Vector3): Double;
procedure DrawXImplicitSurface;
function DirectionsImplicitSurf(var deriv: Vector3; prevDeriv, here: Vector3;
  axis: Integer): Boolean;
function GeodAccelImplicitSurf(var accel: Vector3; vel, x: vector3; toRight:
  Integer):
  Boolean;
procedure DrawCurvesImplicitSurface(fromHere: Vector3);
procedure DrawMeshImplicitSurface;

procedure RecordAndDrawSegmentConst1(previousX, currentX: Vector2; key:
  Integer);
procedure RecordAndDrawSegmentConst2(previousX, currentX: Vector2; key:
  Integer);
procedure RecordAndDrawSegmentConst3(previousX, currentX: Vector2; key:
  Integer);
procedure RecordInWireImplicitSurf1(previousX, currentX: Vector2; key:
  Integer);
procedure RecordInWireImplicitSurf2(previousX, currentX: Vector2; key:
  Integer);
procedure RecordInWireImplicitSurf3(previousX, currentX: Vector2; key:
  Integer);
function CheckRangeWireImplicitSurf1(var depth: Double; p: Vector2): Boolean;
function CheckRangeWireImplicitSurf2(var depth: Double; p: Vector2): Boolean;
function CheckRangeWireImplicitSurf3(var depth: Double; p: Vector2): Boolean;
procedure AdjustPointImplicitSurf1(var x, vel: Vector3);
procedure AdjustPointImplicitSurf2(var x, vel: Vector3);

function ComputeFrameImplicitSurface(var vel, orthToVel: Vector3; var phi,
  topT: Double; x: Vector3): double;
{Given the vector 'vel' tangent to the surface at the point
'x', 'orthToVel' shall return a tangent vector at x that is
orthogonal to 'vel', and with same length. The result is the determinant of the
first fundamental form at 'x', or zero if the computations fails.}

procedure ComputeWireImplicitSurface;

implementation
uses
  SysUtils, Math,
{$IFDEF MSWINDOWS}
   Controls, Printers,
{$ENDIF}
{$IFDEF LINUX}
   QControls, QPrinters,
{$ENDIF}
  FSurface, OcFunct, Utilities, FRunKut, DrawImplicitSurf, FImplicitCurve;
  
var
  uZ: Vector3;
  uConstantCoordinate: Single;

procedure PartialsOfFImplicitSurface(position: Vector3; var Hessian: Matrix3;
  var
  grad: Vector3);
var
  place: Vector3;
  fu, fmu, fv, fmv, fw, fmw, fuv, fmumv, fvw, fmvmw, fwu, fmwmu, f002, numuu,
    numvv, numww, numuv, numvw, numwu: double;
begin
  place := position;
  f002 := gTheF1.V(place);
  f002 := 2 * f002;
  place[1] := position[1] + gIncr3D[1];
  fu := gTheF1.V(place);
  place[1] := position[1] - gIncr3D[1];
  fmu := gTheF1.V(place);
  place[1] := position[1];
  place[2] := position[2] + gIncr3D[2];
  fv := gTheF1.V(place);
  place[2] := position[2] - gIncr3D[2];
  fmv := gTheF1.V(place);
  place[2] := position[2];
  place[3] := position[3] + gIncr3D[3];
  fw := gTheF1.V(place);
  place[3] := position[3] - gIncr3D[3];
  fmw := gTheF1.V(place);
  place[1] := position[1] - gIncr3D[1];
  fmwmu := gTheF1.V(place);
  place[1] := position[1];
  place[2] := position[2] - gIncr3D[2];
  fmvmw := gTheF1.V(place);
  place[3] := position[3];
  place[1] := position[1] - gIncr3D[1];
  fmumv := gTheF1.V(place);
  place[1] := position[1] + gIncr3D[1];
  place[2] := position[2] + gIncr3D[2];
  fuv := gTheF1.V(place);
  place[2] := position[2];
  place[3] := position[3] + gIncr3D[3];
  fwu := gTheF1.V(place);
  place[1] := position[1];
  place[2] := position[2] + gIncr3D[2];
  fvw := gTheF1.V(place);
  grad[1] := 0.5 * (fu - fmu) / gIncr3D[1];
  grad[2] := 0.5 * (fv - fmv) / gIncr3D[2];
  grad[3] := 0.5 * (fw - fmw) / gIncr3D[3];
  numuu := fu + fmu - f002;
  numvv := fv + fmv - f002;
  numww := fw + fmw - f002;
  numuv := 0.5 * (fuv + fmumv - f002 - numuu - numvv);
  numvw := 0.5 * (fvw + fmvmw - f002 - numvv - numww);
  numwu := 0.5 * (fwu + fmwmu - f002 - numww - numuu);
  Hessian[1][1] := numuu / sqr(gIncr3D[1]);
  Hessian[2][2] := numvv / sqr(gIncr3D[2]);
  Hessian[3][3] := numww / sqr(gIncr3D[3]);
  Hessian[1][2] := numuv / (gIncr3D[1] * gIncr3D[2]);
  Hessian[2][1] := Hessian[1][2];
  Hessian[2][3] := numvw / (gIncr3D[2] * gIncr3D[3]);
  Hessian[3][2] := Hessian[2][3];
  Hessian[3][1] := numwu / (gIncr3D[3] * gIncr3D[1]);
  Hessian[1][3] := Hessian[3][1];
end;




procedure ComputeCurvaturesImplicitSurface(pos: Vector3);
 {Gives in 'gKStr, gHStr, gK1Str, gK2Str' the Gauss curvature, mean curv.,}
 {and both principal curvatures at the point 'pos'.}
var
  gg, ggm, sign, traceH, auxH, auxK: double;
  adjH: array[1..3] of Vector3;
  grad: Vector3;
  Hessian: matrix3;
begin
  gKStr := '';
  gHStr := '';
  gK1Str := '';
  gK2Str := '';
  PartialsOfFImplicitSurface(pos, Hessian, grad);
  gg := EuclInnerProd3(grad, grad);
  if gEpsilon = 1 then
    ggm := gg
  else begin
    ggm := gg - 2 * sqr(grad[3]);
  end;
  if abs(ggm) < 1.0E-7 then
    Exit;
  if ggm < 0 then
    sign := 1
  else
    sign := -1;
  adjH[1][1] := Hessian[2][2] * Hessian[3][3] - sqr(Hessian[2][3]);
  adjH[1][2] := Hessian[3][2] * Hessian[1][3] - Hessian[2][1] *
    Hessian[3][3];
  adjH[2][1] := adjH[1][2];
  adjH[1][3] := Hessian[1][2] * Hessian[2][3] - Hessian[3][1] *
    Hessian[2][2];
  adjH[3][1] := adjH[1][3];
  adjH[2][2] := Hessian[1][1] * Hessian[3][3] - sqr(Hessian[3][1]);
  adjH[2][3] := Hessian[1][2] * Hessian[1][3] - Hessian[1][1] *
    Hessian[3][2];
  adjH[3][2] := adjH[2][3];
  adjH[3][3] := Hessian[1][1] * Hessian[2][2] - sqr(Hessian[2][1]);

  auxK := (adjH[1][1] * grad[1] + adjH[2][1] * grad[2] + adjH[3][1] * grad[3])
    *
    grad[1];
  auxK := auxK + (adjH[1][2] * grad[1] + adjH[2][2] * grad[2] + adjH[3][2] *
    grad[3]) * grad[2];
  auxK := auxK + (adjH[1][3] * grad[1] + adjH[2][3] * grad[2] + adjH[3][3] *
    grad[3]) * grad[3];
  auxK := auxK / sqr(gg);
  if gEpsilon = -1 then
    auxK := sign * sqr(gg / ggm) * auxK;
  if (gEpsilon = -1) and (ggm > 0) then
    gKStr := kBullet + FloatToStrF(auxK, ffFixed, 7, 4)
  else
    gKStr := FloatToStrF(auxK, ffFixed, 8, 4);

  grad[3] := gEpsilon * grad[3];
  traceH := Hessian[1][1] + Hessian[2][2] + gEpsilon * Hessian[3][3];
  auxH := (Hessian[1][1] * grad[1] + Hessian[2][1] * grad[2] + Hessian[3][1]
    * grad[3]) * grad[1];
  auxH := auxH + (Hessian[1][2] * grad[1] + Hessian[2][2] * grad[2] +
    Hessian[3][2] * grad[3]) * grad[2];
  auxH := auxH + (Hessian[1][3] * grad[1] + Hessian[2][3] * grad[2] +
    Hessian[3][3] * grad[3]) * grad[3];
  auxH := -sign * (ggm * traceH - auxH) / (2 * ggm * sqrt(abs(ggm)));
  gHStr := FloatToStrF(auxH, ffFixed, 14, 3);

  gg := sqr(auxH) - auxK;
  if gg > -1.0E-7 then begin
    gg := sqrt(abs(gg));
    ggm := auxH + gg;
    gK1Str := FloatToStrF(ggm, ffFixed, 14, 3);
    ggm := auxH - gg;
    gK2Str := FloatToStrF(ggm, ffFixed, 14, 3);
  end;
end;



procedure DrawXImplicitSurface;
var
  i: integer;
  e1, e2, he1, he2, dir1, dir2, pos1, pos2, pos3, pos4, grad: Vector3;
  aux, eLow, fLow, fLow2, gLow, m1, m2, g1, g2, l1, l2, p, q: double;
  gg, ggm, sign, traceH, auxH, auxK: double;
  adjH: array[1..3] of Vector3;
  Hessian: Matrix3;
  can: TCanvas;
begin
  EraseX;
  can := surface.Drawing3D.canvas;
  if (gGeomOption = LightRays) and (gEpsilon = 1) then
    Exit;
  gKStr := '';
  gHStr := '';
  gK1Str := '';
  gK2Str := '';
  PartialsOfFImplicitSurface(gCur3DPos, Hessian, grad);
  gg := EuclInnerProd3(grad, grad);
  if gEpsilon = 1 then
    ggm := gg
  else
    ggm := gg - 2 * sqr(grad[3]);
  if abs(ggm) < 1.0E-7 then
    Exit;
  if ggm < 0 then
    sign := 1
  else
    sign := -1;
  adjH[1][1] := Hessian[2][2] * Hessian[3][3] - sqr(Hessian[2][3]);
  adjH[1][2] := Hessian[3][2] * Hessian[1][3] - Hessian[2][1] *
    Hessian[3][3];
  adjH[2][1] := adjH[1][2];
  adjH[1][3] := Hessian[1][2] * Hessian[2][3] - Hessian[3][1] *
    Hessian[2][2];
  adjH[3][1] := adjH[1][3];
  adjH[2][2] := Hessian[1][1] * Hessian[3][3] - sqr(Hessian[3][1]);
  adjH[2][3] := Hessian[1][2] * Hessian[1][3] - Hessian[1][1] *
    Hessian[3][2];
  adjH[3][2] := adjH[2][3];
  adjH[3][3] := Hessian[1][1] * Hessian[2][2] - sqr(Hessian[2][1]);

  auxK := (adjH[1][1] * grad[1] + adjH[2][1] * grad[2] + adjH[3][1] * grad[3])
    *
    grad[1];
  auxK := auxK + (adjH[1][2] * grad[1] + adjH[2][2] * grad[2] + adjH[3][2] *
    grad[3]) * grad[2];
  auxK := auxK + (adjH[1][3] * grad[1] + adjH[2][3] * grad[2] + adjH[3][3] *
    grad[3]) * grad[3];
  auxK := auxK / sqr(gg);
  if gEpsilon = -1 then
    auxK := sign * sqr(gg / ggm) * auxK;
  if (gEpsilon = -1) and (ggm > 0) then
    gKStr := '' + FloatToStrF(auxK, ffFixed, 7, 4)
  else
    gKStr := FloatToStrF(auxK, ffFixed, 8, 4);

  grad[3] := gEpsilon * grad[3];
  traceH := Hessian[1][1] + Hessian[2][2] + gEpsilon * Hessian[3][3];
  auxH := (Hessian[1][1] * grad[1] + Hessian[2][1] * grad[2] + Hessian[3][1]
    * grad[3]) * grad[1];
  auxH := auxH + (Hessian[1][2] * grad[1] + Hessian[2][2] * grad[2] +
    Hessian[3][2] * grad[3]) * grad[2];
  auxH := auxH + (Hessian[1][3] * grad[1] + Hessian[2][3] * grad[2] +
    Hessian[3][3] * grad[3]) * grad[3];
  auxH := -sign * (ggm * traceH - auxH) / (2 * ggm * sqrt(abs(ggm)));
  gHStr := FloatToStrF(auxH, ffFixed, 14, 3);

  gg := sqr(auxH) - auxK;
  if gg > -1.0E-7 then begin
    gg := sqrt(abs(gg));
    ggm := auxH + gg;
    gK1Str := FloatToStrF(ggm, ffFixed, 14, 3);
    ggm := auxH - gg;
    gK2Str := FloatToStrF(ggm, ffFixed, 14, 3);
  end;
  surface.ShowCurvatures(false);
  grad[3] := gEpsilon * grad[3];
  if (gGeomOption = LightRays) then begin
    p := sqr(grad[1]) + sqr(grad[2]);
    q := p - sqr(grad[3]);
    if q < 1.0E-9 then
      Exit;
    q := sqrt(q);
    dir1[1] := grad[3] * grad[1] - q * grad[2];
    dir1[2] := grad[3] * grad[2] + q * grad[1];
    dir1[3] := -p;
    dir2[1] := grad[3] * grad[1] + q * grad[2];
    dir2[2] := grad[3] * grad[2] - q * grad[1];
    dir2[3] := -p;
  end
  else begin
    uZ[1] := 0.57655438900645;
    uZ[2] := 3 / sqrt(13);
    uZ[3] := 13 / sqrt(201);
    e1 := EuclCrossProduct(grad, uZ);
    aux := EuclNorm3(e1);
    if aux < 1.0E-7 then
      Exit;
    for i := 1 to 3 do
      e1[i] := e1[i] / aux;
    e2 := EuclCrossProduct(grad, e1);
    aux := EuclNorm3(e2);
    for i := 1 to 3 do
      e2[i] := e2[i] / aux;
    for i := 1 to 3 do begin
      he1[i] := EuclInnerProd3(Hessian[i], e1);
      he2[i] := EuclInnerProd3(Hessian[i], e2);
    end;
    elow := EuclInnerProd3(he1, e1);
    fLow := EuclInnerProd3(he1, e2);
    gLow := EuclInnerProd3(he2, e2);

    if (gGeomOption = CurvatureLines) then begin
      if gEpsilon = 1 then begin
        aux := fLow;
        fLow := 0.5 * (gLow - eLow);
        eLow := aux;
        gLow := -aux;
      end
      else begin
        m1 := he1[1] * grad[1] + he1[2] * grad[2] - he1[3] * grad[3];
        m2 := he2[1] * grad[1] + he2[2] * grad[2] - he2[3] * grad[3];
        g1 := grad[1] * e1[1] + grad[2] * e1[2] - grad[3] * e1[3];
        g2 := grad[1] * e2[1] + grad[2] * e2[2] - grad[3] * e2[3];
        aux := sqr(grad[1]) + sqr(grad[2]) - sqr(grad[3]);
        if abs(aux) < 1.0E-9 then
          Exit;
        fLow := m1 * g1 - aux * (he1[1] * e1[1] + he1[2] * e1[2] - he1[3]
          * e1[3]);
        eLow := m1 * g2 - aux * (he1[1] * e2[1] + he1[2] * e2[2] - he1[3]
          * e2[3]);
        gLow := aux * (he2[1] * e1[1] + he2[2] * e1[2] - he2[3] * e1[3]) -
          m2 * g1;
        fLow2 := m2 * g2 - aux * (he2[1] * e2[1] + he2[2] * e2[2] - he2[3]
          * e2[3]);
        fLow := 0.5 * (fLow2 - fLow);
      end;
    end;

    aux := sqr(fLow) - eLow * gLow;
    if aux < 1.0E-10 then
      Exit;

    aux := sqrt(aux);
    l1 := sqr(-fLow + aux) + sqr(eLow);
    l2 := sqr(fLow + aux) + sqr(gLow);
    if l1 < l2 then begin
      l2 := 1 / sqrt(l2);
      p := l2 * gLow;
      q := -(fLow + aux) * l2;
    end
    else begin
      l1 := 1 / sqrt(l1);
      p := (aux - fLow) * l1;
      q := eLow * l1;
    end;
    for i := 1 to 3 do
      dir1[i] := p * e1[i] + q * e2[i];

    aux := -aux;
    l1 := sqr(-fLow + aux) + sqr(eLow);
    l2 := sqr(fLow + aux) + sqr(gLow);
    if l1 < l2 then begin
      l2 := 1 / sqrt(l2);
      p := l2 * gLow;
      q := -(fLow + aux) * l2;
    end
    else begin
      l1 := 1 / sqrt(l1);
      p := (aux - fLow) * l1;
      q := eLow * l1;
    end;
    for i := 1 to 3 do
      dir2[i] := p * e1[i] + q * e2[i];
  end; {DE clase<> LightRays}
  p := 3 * gMaxSegmentLength3D / sqrt(EuclInnerProd3(dir1, dir1));
  q := 3 * gMaxSegmentLength3D / sqrt(EuclInnerProd3(dir2, dir2));
  for i := 1 to 3 do begin
    dir1[i] := p * dir1[i];
    dir2[i] := q * dir2[i];
    pos1[i] := gCur3DPos[i] - dir1[i];
    pos2[i] := gCur3DPos[i] + dir1[i];
    pos3[i] := gCur3DPos[i] - dir2[i];
    pos4[i] := gCur3DPos[i] + dir2[i];
  end;
  gX3D[0] := gCurPtDrawing3D;
  gX3D[1] := ProjectToPaneInt(pos1);
  gX3D[2] := ProjectToPaneInt(pos2);
  gX3D[3] := ProjectToPaneInt(pos3);
  gX3D[4] := ProjectToPaneInt(pos4);

  can.Pen.Color := cWhite;
  can.Pen.Mode := pmXor;
  can.moveto(gX3D[1].x, gX3D[1].y);
  can.lineto(gX3D[2].x, gX3D[2].y);
  can.moveto(gX3D[3].x, gX3D[3].y);
  can.lineto(gX3D[4].x, gX3D[4].y);
  can.Pen.Mode := pmCopy;
  gThereIsX3D := True;
end;

function ComputeVelImplicitSurface(var velocity: Vector3; normalized:
  Boolean): Boolean;
var
  i: Integer;
  an, dg, t: double;
  director, normal: Vector3;
begin
  Result := False;
  for i := 1 to 3 do
    velocity[i] := 0;
  if (gCurPtDrawing3D.x = gFirstPtDrawing3D.x) and (gCurPtDrawing3D.y =
    gFirstPtDrawing3D.y) then
    Exit;
  if not gFirstPtGood then
    Exit;
  director := FindRay(gCurPtDrawing3D);
  normal := EuclideanGradOfF(gFirst3DPos);
  dg := EuclInnerProd3(director, normal);

  if abs(dg) > 1.0E-7 then begin
    t := (gFirst3DPos[1] - gViewpoint[1]) * normal[1] + (gFirst3DPos[2] -
      gViewpoint[2]) *
      normal[2] + (gFirst3DPos[3] - gViewpoint[3]) * normal[3];
    t := t / dg;
    for i := 1 to 3 do
      velocity[i] := gViewpoint[i] - gFirst3DPos[i] + t * director[i];
    if normalized then begin
      an := sqrt(abs(InnerProd3(velocity, velocity)));
      if an > 1.0E-9 then
        for i := 1 to 3 do
          velocity[i] := velocity[i] / an;
    end;
    result := True;
  end;
end;

function ComputeClipImplicitSurface(var theClip: Clip):
  Boolean;
var
  i: Integer;
  xu, vel, normal: Vector3;
  a: Double;
begin
  Result := ComputeVelImplicitSurface(Vel, True);
  if Result then begin
    normal := EuclideanGradOfF(gFirst3DPos);
    xu := EuclCrossProduct(normal, vel);
    a := EuclNorm3(xu);
    if a > 1.0E-9 then begin
      a := 1 / a;
      for i := 1 to 3 do
        theClip.u[i] := xu[i] * a;
      theClip.a := -EuclInnerProd3(xu, gFirst3DPos) * a;
    end
    else
      Result := False;
  end;
end;

procedure AddClipImplicitSurface(pos, vel: Vector3);
var
  i: Integer;
  xu, normal: Vector3;
  a: Double;
begin
  normal := EuclideanGradOfF(pos);
  xu := EuclCrossProduct(normal, vel);
  a := EuclNorm3(xu);
  if a > 1.0E-9 then begin
    if gClips <> nil then
      i := Length(gClips)
    else
      i := 0;
    SetLength(gClips, i + 1);
    gClipTop := i;
    a := 1 / a;
    if normal[1] * (gViewPoint[1] - gCenter3D[1]) + normal[2] * (gViewPoint[2]
      -
      gCenter3D[2]) + normal[3] * (gViewPoint[3] - gCenter3D[3]) < 0 then
      a := -a;
    for i := 1 to 3 do
      gClips[gClipTop].u[i] := xu[i] * a;
    gClips[gClipTop].a := -EuclInnerProd3(xu, pos) * a;
  end;
end;

function InRangeAndInClipsImplicitSurface(var depth: double; x: Vector3):
  Boolean;
var
  i: Integer;
begin
  Result := True;
  depth := min(gHalfEdges3D[1] - abs(x[1] -
    gCenter3D[1]), gHalfEdges3D[2] - abs(x[2] -
    gCenter3D[2]));
  depth := min(depth, gHalfEdges3D[3] - abs(x[3] -
    gCenter3D[3]));
  if gClips <> nil then
    for i := 0 to gClipTop do
      Result := Result and (HeightOverClip(gClips[i], x) < 0);
end;

function InRangeAndInClipsImplicitSurfaceS(var depth: double; x:
  Vector3Single):
  Boolean;
var
  i: Integer;
begin
  Result := True;
  depth := min(gHalfEdges3D[1] - abs(x[1] -
    gCenter3D[1]), gHalfEdges3D[2] - abs(x[2] -
    gCenter3D[2]));
  depth := min(depth, gHalfEdges3D[3] - abs(x[3] -
    gCenter3D[3]));
  if gClips <> nil then
    for i := 0 to gClipTop do
      Result := Result and (HeightOverClipS(gClips[i], x) < 0);
end;


procedure RecordAndDrawSegmentImplicitSurface(x1, x2: Vector3; toSave:
  Boolean);
var
  i, colorSegm, anver: Integer;
  brightness: double;
  d: Double;
  col: byte;
  pixR1, pixR2: Vector2;
  pixel1, pixel2, cent: Tpoint;
  vis1, vis2: boolean;
  mid, mid1, mid2: Vector3;
  can: TCanvas;
  g: gamma16Ptr;
begin
  can := surface.drawing3D.canvas;
  g := gLinesData[gActiveLine].gammaP;
  vis1 := InRangeAndInClipsImplicitSurface(d, x1);
  vis2 := InRangeAndInClipsImplicitSurface(d, x2);
  vis1 := vis1 and IsVisibleImplicit(gViewPoint, x1, gVPCubeKey, true);
  vis2 := vis2 and IsVisibleImplicit(gViewPoint, x2, gVPCubeKey, true);
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      pixR1 := ProjectToPaneRealAndDist(x1, d);
      pixel1.x := round(pixR1[1]);
      pixel1.y := round(pixR1[2]);
      pixR2 := ProjectToPaneRealAndDist(x2, d);
      pixel2.x := round(pixR2[1]);
      pixel2.y := round(pixR2[2]);
      if PtInRect(gRect, pixel1) and PtInRect(gRect, pixel2) then begin
        cent.x := (pixel1.x + pixel2.x) div 2;
        cent.y := (pixel1.y + pixel2.y) div 2;
        if gMap <> nil then begin
          col := gMap^[cent.x, cent.y];
          if col > 100 then begin
            brightness := (col - 100) / 100;
            anver := 1;
          end
          else if col > 0 then begin
            brightness := col / 100;
            anver := 0;
          end
          else begin
            brightness := 0;
            anver := -1;
          end;
          colorSegm := round(1 + 15 * brightness);
          can.Pen.Color := g^[colorSegm];
          if ((gSideToDrawOn = 2) or (anver = gSideToDrawOn)) then begin
            can.Moveto(pixel1.x, pixel1.y);
            can.Lineto(pixel2.x, pixel2.y);
            AddSegment(pixR1, pixR2, 3, colorSegm);
          end;
        end;
      end;
    end
    else begin
      pixel1 := ProjectToPaneInt(x1);
      pixel2 := ProjectToPaneInt(x2);
      if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then begin
        for i := 1 to 3 do
          mid[i] := 0.5 * (x1[i] + x2[i]);
        mid1 := x1;
        mid2 := mid;
        RecordAndDrawSegmentImplicitSurface(mid1, mid2, toSave);
        mid1 := mid;
        mid2 := x2;
        RecordAndDrawSegmentImplicitSurface(mid1, mid2, toSave);
      end;
    end;
  end;
end;

function CheckRangeImplicit(lastX, nextX: Vector3; findProlong: Boolean;
  var newX, newVel: Vector3): Integer;
var
  i: Integer;
  depth, aux: Double;
  inside: Boolean;
begin
  depth := 1.0E100;
  for i := 1 to 3 do begin
    aux := gHalfEdges3D[i] - abs(nextX[i] - gCenter3D[i]);
    if depth > aux then
      depth := aux;
  end;
  inside := (depth >= 0);
  if inside and (gClips <> nil) then
    for i := 0 to gClipTop do
      inside := inside and (HeightOverClip(gClips[i], nextX) < 0);
  result := ord(inside);
end;

function CheckCameBackImplicitSurface(firstX, prevX, curX:
  Vector3; counter: Integer): Boolean;
var
  bb, bc, cc, aux: Double;
  b, c: Vector3;
  i: Integer;
begin
  result := False;
  if counter < 25 then
    Exit;
  for i := 1 to 3 do begin
    b[i] := firstX[i] - prevX[i];
    c[i] := curX[i] - prevX[i];
  end;
  bb := EuclInnerProd3(b, b);
  if sqrt(bb) < gMaxSegmentLength3D then begin
    bc := EuclInnerProd3(b, c);
    if bc > 0 then begin
      cc := EuclInnerProd3(c, c);
      if cc > bc then begin
        aux := sqrt(abs(bb - sqr(bc) / cc));
        result := aux < 2 * gRadius3D;
      end;
    end;
  end;
end;

function CheckHandImplicit(x, t, v: Vector3): Double;
var
  n, d: Vector3;
  j: Integer;
begin
  for j := 1 to 3 do
    d[j] := t[j] - x[j];
  n := EuclideanGradOfF(t);
  result := v[1] * (d[2] * n[3] - d[3] * n[2]) + v[2] * (d[3] * n[1] - d[1] *
    n[3]) +
    v[3] * (d[1] * n[2] - d[2] * n[1]);
end;

function DirectionsImplicitSurf(var deriv: Vector3; prevDeriv, here: Vector3;
  axis: Integer): Boolean;
var
  i: integer;
  e1, e2, he1, he2, uz, grad: Vector3;
  aux, eLow, fLow, fLow2, gLow, m1, m2, g1, g2, l1, l2, p, q: double;
  hessian: Matrix3;
begin
  result := False;
  PartialsOfFImplicitSurface(here, hessian, grad);
  repeat
    for i := 1 to 3 do
      uZ[i] := Random;
  until (uZ[1] <> 0) or (uZ[2] <> 0) or (uZ[3] <> 0);

  if (gGeomOption = LightRays) then begin
    p := sqr(grad[1]) + sqr(grad[2]);
    q := p - sqr(grad[3]);
    if q < 1.0E-9 then
      Exit;
    q := sqrt(q);
    deriv[1] := grad[3] * grad[1] - (1 - 2 * axis) * q * grad[2];
    deriv[2] := grad[3] * grad[2] + (1 - 2 * axis) * q * grad[1];
    deriv[3] := -p;
  end
  else begin
    e1 := EuclCrossProduct(grad, uZ);
    aux := EuclNorm3(e1);
    if aux < 1.0E-9 then
      Exit;
    for i := 1 to 3 do
      e1[i] := e1[i] / aux;
    e2 := EuclCrossProduct(grad, e1);
    aux := EuclNorm3(e2);
    for i := 1 to 3 do
      e2[i] := e2[i] / aux;
    for i := 1 to 3 do begin
      he1[i] := EuclInnerProd3(hessian[i], e1);
      he2[i] := EuclInnerProd3(hessian[i], e2);
    end;

    elow := EuclInnerProd3(he1, e1);
    fLow := EuclInnerProd3(he1, e2);
    gLow := EuclInnerProd3(he2, e2);

    if (gGeomOption = CurvatureLines) then begin
      if gEpsilon = 1 then begin
        aux := fLow;
        fLow := 0.5 * (gLow - eLow);
        eLow := aux;
        gLow := -aux;
      end
      else begin
        m1 := he1[1] * grad[1] + he1[2] * grad[2] - he1[3] * grad[3];
        m2 := he2[1] * grad[1] + he2[2] * grad[2] - he2[3] * grad[3];
        g1 := grad[1] * e1[1] + grad[2] * e1[2] - grad[3] * e1[3];
        g2 := grad[1] * e2[1] + grad[2] * e2[2] - grad[3] * e2[3];
        aux := sqr(grad[1]) + sqr(grad[2]) - sqr(grad[3]);
        if abs(aux) < 1.0E-9 then
          Exit;
        fLow := m1 * g1 - aux * (he1[1] * e1[1] + he1[2] * e1[2] - he1[3]
          * e1[3]);
        eLow := m1 * g2 - aux * (he1[1] * e2[1] + he1[2] * e2[2] - he1[3]
          * e2[3]);
        gLow := aux * (he2[1] * e1[1] + he2[2] * e1[2] - he2[3] * e1[3]) -
          m2 * g1;
        fLow2 := m2 * g2 - aux * (he2[1] * e2[1] + he2[2] * e2[2] - he2[3]
          * e2[3]);
        fLow := 0.5 * (fLow2 - fLow);
      end;
    end;

    aux := sqr(fLow) - eLow * gLow;
    if aux <= 0 then
      Exit;

    aux := (1 - 2 * axis) * sqrt(aux);
    l1 := sqr(-fLow + aux) + sqr(eLow);
    l2 := sqr(fLow + aux) + sqr(gLow);
    if l1 < l2 then begin
      l2 := 1 / sqrt(l2);
      p := l2 * gLow;
      q := -(fLow + aux) * l2;
    end
    else begin
      l1 := 1 / sqrt(l1);
      p := (aux - fLow) * l1;
      q := eLow * l1;
    end;
    for i := 1 to 3 do
      deriv[i] := p * e1[i] + q * e2[i];
  end;
  aux := EuclInnerProd3(deriv, prevDeriv);
  if aux < 0 then
    for i := 1 to 3 do
      deriv[i] := -deriv[i];
  result := True;
end;

function GeodAccelImplicitSurf(var accel: Vector3; vel, x: vector3; toRight:
  Integer):
  Boolean;
var
  i: Integer;
  aux, gg: double;
  hv, grad: Vector3;
  hessian: Matrix3;
begin
  result := TRUE;
  PartialsOfFImplicitSurface(x, hessian, grad);
  grad[3] := gEpsilon * grad[3];
  gg := InnerProd3(grad, grad);
  if abs(gg) < 1.0E-9 then begin
    Result := FALSE;
    Exit;
  end;
  for i := 1 to 3 do
    hv[i] := EuclInnerProd3(hessian[i], vel);
  aux := -EuclInnerProd3(hv, vel) / gg;
  for i := 1 to 3 do
    accel[i] := aux * grad[i];
  if toRight <> 0 then begin
    hv := CrossProduct(grad, vel);
    gg := toRight / sqrt(abs(gg));
    for i := 1 to 3 do
      accel[i] := accel[i] + gg * hv[i];
  end;
end;

procedure DrawFirstOrderLines(fromHere: Vector3);
var
  axis, j: Integer;
  initvel: array[0..1] of Vector3;
  vel, here: Vector3;
  key: Integer;
begin
  Randomize;
  surface.Drawing3D.canvas.Pen.Mode := pmCopy;
  if (gGeomOption = LightRays) and (gEpsilon = 1) then
    Exit;
  repeat
    for j := 1 to 3 do
      initVel[0, j] := Random;
  until (initVel[0, 1] <> 0) or (initVel[0, 2] <> 0) or (initVel[0, 3] <> 0);
  here := fromHere;
  if not DirectionsImplicitSurf(vel, initVel[0], here, 0) then
    Exit;
  DirectionsImplicitSurf(initVel[1], initVel[0], here, 1);
  initVel[0] := vel;
  theRunKut.SetEnvironment(3, 1, kTol, DirectionsImplicitSurf,
    CheckRangeImplicit,
    RecordAndDrawSegmentImplicitSurface, CheckCameBackImplicitSurface,
    AdjustPointImplicitSurf1);
  for axis := 0 to 1 do begin
    vel := initVel[axis];
    here := fromHere;
    key := theRunKut.ComputeIntegralCurve(here, vel, axis);
    if (key and kDoubled) = 0 then begin
      here := fromHere;
      for j := 1 to 3 do
        vel[j] := -initVel[axis, j];
      theRunKut.ComputeIntegralCurve(here, vel, axis);
    end;
  end;
end;

procedure RecordAndDrawSegmentConst1(previousX, currentX: Vector2; key:
  Integer);
var
  x1, x2: Vector3;
begin
  x1[1] := uConstantCoordinate;
  x1[2] := previousX[1];
  x1[3] := previousX[2];
  x2[1] := uConstantCoordinate;
  x2[2] := currentX[1];
  x2[3] := currentX[2];
  RecordAndDrawSegmentImplicitSurface(x1, x2, True);
end;

procedure RecordAndDrawSegmentConst2(previousX, currentX: Vector2; key:
  Integer);
var
  x1, x2: Vector3;
begin
  x1[1] := previousX[1];
  x1[2] := uConstantCoordinate;
  x1[3] := previousX[2];
  x2[1] := currentX[1];
  x2[2] := uConstantCoordinate;
  x2[3] := currentX[2];
  RecordAndDrawSegmentImplicitSurface(x1, x2, True);
end;

procedure RecordAndDrawSegmentConst3(previousX, currentX: Vector2; key:
  Integer);
var
  x1, x2: Vector3;
begin
  x1[1] := previousX[1];
  x1[2] := previousX[2];
  x1[3] := uConstantCoordinate;
  x2[1] := currentX[1];
  x2[2] := currentX[2];
  x2[3] := uConstantCoordinate;
  RecordAndDrawSegmentImplicitSurface(x1, x2, True);
end;

procedure RecordInWireImplicitSurf1(previousX, currentX: Vector2; key:
  Integer);
var
  pos: Vector3;
begin
  pos[1] := uConstantCoordinate;
  pos[2] := currentX[1];
  pos[3] := currentX[2];
  AddSegmentToWire(pos, key);
end;

procedure RecordInWireImplicitSurf2(previousX, currentX: Vector2; key:
  Integer);
var
  pos: Vector3;
begin
  pos[1] := currentX[1];
  pos[2] := uConstantCoordinate;
  pos[3] := currentX[2];
  AddSegmentToWire(pos, key);
end;



procedure RecordInWireImplicitSurf3(previousX, currentX: Vector2; key:
  Integer);
var
  pos: Vector3;
begin
  pos[1] := currentX[1];
  pos[2] := currentX[2];
  pos[3] := uConstantCoordinate;
  AddSegmentToWire(pos, key);
end;

function CheckRangeWireImplicitSurf1(var depth: Double; p: Vector2): Boolean;
var
  pos: Vector3;
begin
  pos[1] := uConstantCoordinate;
  pos[2] := p[1];
  pos[3] := p[2];
  result := InRangeAndInClipsImplicitSurface(depth, pos);
end;

function CheckRangeWireImplicitSurf2(var depth: Double; p: Vector2): Boolean;
var
  pos: Vector3;
begin
  pos[1] := p[1];
  pos[2] := uConstantCoordinate;
  pos[3] := p[2];
  result := InRangeAndInClipsImplicitSurface(depth, pos);
end;

function CheckRangeWireImplicitSurf3(var depth: Double; p: Vector2): Boolean;
var
  pos: Vector3;
begin
  pos[1] := p[1];
  pos[2] := p[2];
  pos[3] := uConstantCoordinate;
  result := InRangeAndInClipsImplicitSurface(depth, pos);
end;

procedure theFConst1(var f: double; p: Vector2);
var
  pos: Vector3;
begin
  pos[1] := uConstantCoordinate;
  pos[2] := p[1];
  pos[3] := p[2];
  f := gTheF1.V(pos);
end;

procedure theFConst2(var f: double; p: Vector2);
var
  pos: Vector3;
begin
  pos[1] := p[1];
  pos[2] := uConstantCoordinate;
  pos[3] := p[2];
  f := gTheF1.V(pos);
end;

procedure theFConst3(var f: double; p: Vector2);
var
  pos: Vector3;
begin
  pos[1] := p[1];
  pos[2] := p[2];
  pos[3] := uConstantCoordinate;
  f := gTheF1.V(pos);
end;

procedure DrawCutWithCoordinatePlane1(fromHere: Vector3; storeInWire:
  Boolean);
//Draws the cut of the surface with the plane  x[1] = fromHere[1].
var
  minC, maxC, x: Vector2;
begin
  uConstantCoordinate := fromHere[1];
  minC[1] := gMin3D[2];
  minC[2] := gMin3D[3];
  maxC[1] := gMax3D[2];
  maxC[2] := gMax3D[3];
  x[1] := fromHere[2];
  x[2] := fromHere[3];
  if storeInWire then
    theImplicitCurve.ComputeImplicitCurve(theFConst1,
      CheckRangeWireImplicitSurf1,
      RecordInWireImplicitSurf1, gVerySmall, minC, maxC, gPPLine2D, gNumFlags,
      True)
  else
    theImplicitCurve.ComputeImplicitCurve(theFConst1,
      CheckRangeWireImplicitSurf1,
      RecordAndDrawSegmentConst1, gVerySmall, minC, maxC, gPPLine2D,
      gNumFlags,
      False);
end;

procedure DrawCutWithCoordinatePlane2(fromHere: Vector3; storeInWire:
  Boolean);
//Draws the cut of the surface with the plane  x[2] = fromHere[2].
var
  minC, maxC, x: Vector2;
begin
  uConstantCoordinate := fromHere[2];
  minC[1] := gMin3D[1];
  minC[2] := gMin3D[3];
  maxC[1] := gMax3D[1];
  maxC[2] := gMax3D[3];
  x[1] := fromHere[1];
  x[2] := fromHere[3];
  if storeInWire then
    theImplicitCurve.ComputeImplicitCurve(theFConst2,
      CheckRangeWireImplicitSurf2,
      RecordInWireImplicitSurf2, gVerySmall, minC, maxC, gPPLine2D, gNumFlags,
      True)
  else
    theImplicitCurve.ComputeImplicitCurve(theFConst2,
      CheckRangeWireImplicitSurf2,
      RecordAndDrawSegmentConst2, gVerySmall, minC, maxC, gPPLine2D,
      gNumFlags,
      False);
end;

procedure DrawCutWithCoordinatePlane3(fromHere: Vector3; storeInWire:
  Boolean);
//Draws the cut of the surface with the plane  x[3] = fromHere[3].
var
  minC, maxC, x: Vector2;
begin
  uConstantCoordinate := fromHere[3];
  minC[1] := gMin3D[1];
  minC[2] := gMin3D[2];
  maxC[1] := gMax3D[1];
  maxC[2] := gMax3D[2];
  x[1] := fromHere[1];
  x[2] := fromHere[2];
  if storeInWire then
    theImplicitCurve.ComputeImplicitCurve(theFConst3,
      CheckRangeWireImplicitSurf3,
      RecordInWireIMplicitSurf3, gVerySmall, minC, maxC, gPPLine2D, gNumFlags,
      True)
  else
    theImplicitCurve.ComputeImplicitCurve(theFConst3,
      CheckRangeWireImplicitSurf3,
      RecordAndDrawSegmentConst3, gVerySmall, minC, maxC, gPPLine2D,
      gNumFlags,
      False);
end;



procedure DrawCoordLinesImplicitSurface(fromHere: Vector3);
begin
  DrawCutWithCoordinatePlane1(fromHere, False);
  DrawCutWithCoordinatePlane2(fromHere, False);
  DrawCutWithCoordinatePlane3(fromHere, False);
end;

procedure DrawCurvesImplicitSurface(fromHere: Vector3);
begin
  if (not gAlreadyDrawn) or (gMode <> Standard) then
    Exit;
  surface.cursor := crHourGlass;
  if gGeomOption = CoordLines then
    DrawCoordLinesImplicitSurface(fromHere)
  else begin
    EraseX;
    DrawFirstOrderLines(fromHere);
  end;
  surface.cursor := crDefault;
end;

procedure DrawMeshImplicitSurface;
var
  i: Integer;
  x: Vector3;
begin
  x := gMin3D;
  if (gMeshNumber > 1) and (gMeshNumber <= 100) then
    for i := 1 to (gMeshNumber - 1) do begin
      x[1] := gMin3D[1] + i * gEdges3D[1] / gMeshNumber;
      DrawCutWithCoordinatePlane1(x, False);
      x[2] := gMin3D[2] + i * gEdges3D[2] / gMeshNumber;
      DrawCutWithCoordinatePlane2(x, False);
      x[3] := gMin3D[3] + i * gEdges3D[3] / gMeshNumber;
      DrawCutWithCoordinatePlane3(x, False);
    end;
end;

procedure AdjustPointImplicitSurf1(var x, vel: Vector3);
var
  grad: Vector3;
  f0, aux1: double;
  i: integer;
begin
  grad := EuclideanGradOfF(x);
  f0 := gtheF1.V(x);
  aux1 := sqr(grad[1]) + sqr(grad[2]) + sqr(grad[3]);
  if aux1 > 1.0E-10 then begin
    aux1 := -f0 / aux1;
    for i := 1 to 3 do
      x[i] := x[i] + aux1 * grad[i];
  end;
end;

procedure AdjustPointImplicitSurf2(var x, vel: Vector3);
var
  grad: Vector3;
  f0, aux1, aux2: double;
  i: integer;
begin
  grad := EuclideanGradOfF(x);
  f0 := gTheF1.V(x);
  aux1 := EuclInnerProd3(grad, grad);
  if aux1 > 1.0E-10 then begin
    aux2 := -EuclInnerProd3(vel, grad) / aux1;
    aux1 := -f0 / aux1;
    for i := 1 to 3 do begin
      x[i] := x[i] + aux1 * grad[i];
      vel[i] := vel[i] + aux2 * grad[i];
    end;
  end;
end;

function ComputeFrameImplicitSurface(var vel, orthToVel: Vector3; var phi,
  topT: Double; x: Vector3): double;
var
  length, e, p, q: Double;
  normal: Vector3;
  i: Integer;
begin
  normal := EuclideanGradOfF(x);
  topT := sqrt(abs(InnerProd3(vel, vel)));
  length := InnerProd3(normal, normal);
  if (gEpsilon < 0) and (length > 1.0E-12) then begin
    result := -1;
    p := sqr(normal[1]) + sqr(normal[2]);
    q := p - sqr(normal[3]);
    if q < 1.0E-9 then
      Result := 0
    else begin
      e := 1 - 0.5 * q / p;
      Assert((e > 0) and (e < 1));
      phi := arccos(sqrt(e));
      e := p * (2 * p - q);
      Assert(e > 0);
      q := sqrt(e);
      vel[1] := normal[3] * normal[1] / q;
      vel[2] := normal[3] * normal[2] / q;
      vel[3] := -p / q;
      p := sqrt(p);
      orthToVel[1] := -normal[2] / p;
      orthToVel[2] := normal[1] / p;
      orthToVel[3] := 0;
    end
  end
  else if ((gEpsilon < 0) and (length < -1.0E-7)) or
    ((gEpsilon > 0) and (length > 1.0E-12)) then begin
    result := 1;
    orthToVel := CrossProduct(normal, vel);
    vel := CrossProduct(orthToVel, normal);
    p := sqrt(abs(InnerProd3(orthToVel, orthToVel)));
    q := sqrt(abs(InnerProd3(vel, vel)));
    if (p < 1.0E-10) or (q < 1.0E-10) then
      result := 0
    else
      for i := 1 to 3 do begin
        orthToVel[i] := orthToVel[i] * topT / p;
        vel[i] := vel[i] * topT / q;
      end;
  end
  else
    Result := 0;
end;


procedure ComputeWireImplicitSurface;
var
  i: Integer;
  x: Vector3;
begin
  InitWire;
  x := gMin3D;
  for i := 1 to kWiresPerLine - 1 do begin
    x[1] := gMin3D[1] + i * gEdges3D[1] / kWiresPerLine;
    DrawCutWithCoordinatePlane1(x, True);
    x[2] := gMin3D[2] + i * gEdges3D[2] / kWiresPerLine;
    DrawCutWithCoordinatePlane2(x, True);
    x[3] := gMin3D[3] + i * gEdges3D[3] / kWiresPerLine;
    DrawCutWithCoordinatePlane3(x, True);
  end;
  while (gNumNodes3D > 0) and (gWire[gNumNodes3D - 1].key = 1) do
    dec(gNumNodes3D);
end;

{procedure PrintImplicitSurface;
var
  marH, marV, anchuraTexto, y, shy, anchoPagina, i, num :
    integer;
  r, rOr : TRect;
  head, texto : AnsiString;
  pluma : TPen;
begin
  if gBusy or (not gAlreadyDrawn) then Exit;
  pluma := Printer.Canvas.Pen;
  Printer.orientation := poPortrait;
  num := Printer.copies;
  Printer.copies := 1;
  marH := Printer.PageWidth div 10;
  marV := Printer.PageHeight div 10;
  r.left := marH;
  r.top := marV;
  r.right := Printer.PageWidth - marH;
  r.bottom := r.top + r.right - r.left;
  anchoPagina := Printer.PageWidth - marH;
  rOr := r;
  InflateRect(rOr, 8, 8);
  if num < 1 then Exit;
  try
    Printer.BeginDoc;
    for i := 1 to num do
      begin
        Printer.Canvas.Brush.Style := bsClear;
        Printer.Canvas.Font.Name := 'Arial';
        Printer.Canvas.Font.Height := MulDiv(Printer.PageHeight, 20, 1000);
        Printer.Canvas.Font.Style := [fsBold];
        anchuraTexto := Printer.canvas.TextWidth(surface.caption);
        Printer.canvas.TextOut((Printer.PageWidth - anchuraTexto) div 2,
          r.top - 3 * Printer.Canvas.TextHeight(surface.caption),
          surface.caption);

        Printer.Canvas.Font.Style := [];
        Printer.Canvas.Font.Height := MulDiv(Printer.PageHeight, 15, 1000);
        head := '0 = ';

        shy := Printer.Canvas.Font.Height;
        y := r.bottom + 2 * shy;
        DrawLongFormula(Printer.canvas, head, gSurfaceStr[1], anchoPagina, marH,
          shy, y);

        texto := gSurfaceStr[5] + ' < x < ' + gSurfaceStr[6];
        anchuraTexto := Printer.canvas.TextWidth(texto);
        y := y + shy;
        Printer.canvas.TextOut((Printer.PageWidth - anchuraTexto) div 2, y, texto
          );
        texto := gSurfaceStr[7] + ' < y < ' + gSurfaceStr[8];
        anchuraTexto := Printer.canvas.TextWidth(texto);
        y := y + 3 * shy div 2;
        Printer.canvas.TextOut((Printer.PageWidth - anchuraTexto) div 2, y,
          texto
          );
        texto := gSurfaceStr[9] + ' < z < ' + gSurfaceStr[10];
        anchuraTexto := Printer.canvas.TextWidth(texto);
        y := y + 3 * shy div 2;
        Printer.canvas.TextOut((Printer.PageWidth - anchuraTexto) div 2, y,
          texto
          );
        y := y + 2 * shy;

        if (gMode = Standard) or (gMode = WireframeMono) or
          gPrintLines then
          begin
            Printer.canvas.Pen.Color := clBlack;
            Printer.canvas.moveto(rOr.left, rOr.top);
            Printer.canvas.Lineto(rOr.left, rOr.bottom);
            Printer.canvas.Lineto(rOr.right, rOr.bottom);
            Printer.canvas.Lineto(rOr.right, rOr.top);
            Printer.canvas.Lineto(rOr.left, rOr.top);
          end;
        ExcludeClipRect(Printer.canvas.handle, 0, 0, Printer.PageWidth, r.top
          );
        ExcludeClipRect(Printer.canvas.handle, 0, 0, r.left,
          Printer.PageHeight
          );
        ExcludeClipRect(Printer.canvas.handle, 0, r.bottom,
          Printer.PageWidth,
          Printer.PageHeight);
        ExcludeClipRect(Printer.canvas.handle, r.right, 0,
          Printer.PageWidth,
          Printer.PageHeight);
        if (gMode = Standard) or (gMode = ColoredStereoPair) then
            BitmapToPCanvas(Printer.Canvas, r, gStdBmp)
        else if (gMode in gWireModes) then
          Draw3DLinesForPrinter(Printer.Canvas, r);
        if num > 1 then Printer.NewPage;
      end;
  finally
    Printer.EndDoc;
    Printer.Canvas.Pen := pluma;
  end;
end;  }

end.
