

{-----------------------------------------------------------------------------
 Unit Name: ManageClips
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose:  Management of the "manual" clipping and the normal section clipping.
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit ManageClips;

interface
uses
{$IFDEF MSWINDOWS}
  Windows,
{$ENDIF}
{$IFDEF LINUX}
  Types,
{$ENDIF}
  Globals;

procedure FindClipCut;
procedure InitializeClipCut;
procedure DrawCutParamSurface(fromDraw3D:
  Boolean);
procedure DrawCutImplicitSurface;
procedure DrawClipEllipse;
procedure PrepareToSetClip;
procedure MoveClip(pt1, pt2: TPoint);
procedure FindWhatToMove(pt: TPoint);

implementation
uses
 Math,
{$IFDEF MSWINDOWS}
   Graphics,
{$ENDIF}
{$IFDEF LINUX}
  QGraphics,
{$ENDIF}
  FSurface, OcFunct, Utilities, ImplicitUtils, ParametricUtils, FImplicitCurve;

var
  uBE, uRE, uR2, uNear, c0D, c0D2, rContour, dcContour, tDefault, gDist2:
  Double;
  up, uq, cContour: Vector3;
  toMove: Integer;
  ellipseCenter2D, ellipseCenter3D: TPoint;

procedure InitializeClipCut;
begin
  gClipCut := nil;
  SetLength(gClipCut, 50);
  gLastSegmCC := -1;
  if gType <> Implicit then begin
    gCCParams := nil;
    SetLength(gCCParams, 50);
  end;
end;

procedure GetRoomCC;
var
  size: Integer;
begin
  size := Length(gClipCut) + 50;
  SetLength(gClipCut, size);
  if gType <> Implicit then
    SetLength(gCCParams, size);
end;

procedure RecordParamsParam(p1, p2: Vector2; k: Integer);
var
  i, j: Integer;
  x: Vector3;
begin
  inc(gLastSegmCC);
  i := gLastSegmCC;
  if i >= High(gClipCut) then
    GetRoomCC;
  gClipCut[i].key := k;
  theFPar2(x, p2);
  for j := 1 to 3 do
    gClipCut[i].pos[j] := x[j];
  gCCParams[i] := p2;
end;

procedure RecordParamsImpl1(p1, p2: Vector2; k: Integer);
var
  i: Integer;
begin
  inc(gLastSegmCC);
  i := gLastSegmCC;
  if i >= High(gClipCut) then
    GetRoomCC;
  gClipCut[i].key := k;
  gClipCut[i].pos[1] := -(gNewClip.u[2] * p2[1] + gNewClip.u[3] * p2[2] +
    gNewClip.a)
    / gNewClip.u[1];
  gClipCut[i].pos[2] := p2[1];
  gClipCut[i].pos[3] := p2[2];
end;

procedure RecordParamsImpl2(p1, p2: Vector2; k: Integer);
var
  i: Integer;
begin
  inc(gLastSegmCC);
  i := gLastSegmCC;
  if i >= High(gClipCut) then
    GetRoomCC;
  gClipCut[i].key := k;
  gClipCut[i].pos[2] := -(gNewClip.u[3] * p2[1] + gNewClip.u[1] * p2[2] +
    gNewClip.a)
    / gNewClip.u[2];
  gClipCut[i].pos[3] := p2[1];
  gClipCut[i].pos[1] := p2[2];
end;

procedure RecordParamsImpl3(p1, p2: Vector2; k: Integer);
var
  i: Integer;
begin
  inc(gLastSegmCC);
  i := gLastSegmCC;
  if i >= High(gClipCut) then
    GetRoomCC;
  gClipCut[i].key := k;
  gClipCut[i].pos[3] := -(gNewClip.u[1] * p2[1] + gNewClip.u[2] * p2[2] +
    gNewClip.a)
    / gNewClip.u[3];
  gClipCut[i].pos[1] := p2[1];
  gClipCut[i].pos[2] := p2[2];
end;

function CheckRangeClipParam(var depth: Double; p: Vector2): Boolean;
begin
  depth := min(gHalfEdges2D[1] - abs(p[1] - gCenter2D[1]),
    gHalfEdges2D[2] - abs(p[2] - gCenter2D[2]));
  result := (depth >= 0);
end;

function CheckRangeClipImpl1(var depth: Double; p: Vector2): Boolean;
begin
  depth := min(gHalfEdges3D[2] - abs(p[1] - gCenter3D[2]),
    gHalfEdges3D[3] - abs(p[2] - gCenter3D[3]));
  result := (depth >= 0);
end;

function CheckRangeClipImpl2(var depth: Double; p: Vector2): Boolean;
begin
  depth := min(gHalfEdges3D[3] - abs(p[1] - gCenter3D[3]),
    gHalfEdges3D[1] - abs(p[2] - gCenter3D[1]));
  result := (depth >= 0);
end;

function CheckRangeClipImpl3(var depth: Double; p: Vector2): Boolean;
begin
  depth := min(gHalfEdges3D[1] - abs(p[1] - gCenter3D[1]),
    gHalfEdges3D[2] - abs(p[2] - gCenter3D[2]));
  result := (depth >= 0);
end;

procedure ClipValueParam(var value: double; param: vector2);
var
  x: Vector3;
begin
  theFPar2(x, param);
  value := gNewClip.u[1] * x[1] + gNewClip.u[2] * x[2] + gNewClip.u[3] * x[3]
    + gNewClip.a;
end;

procedure ClipValueImpl1(var value: double; param: vector2);
var
  x: Vector3;
begin
  x[1] := -(gNewClip.u[2] * param[1] + gNewClip.u[3] * param[2] + gNewClip.a)
    / gNewClip.u[1];
  x[2] := param[1];
  x[3] := param[2];
  value := gTheF1.V(x);
end;

procedure ClipValueImpl2(var value: double; param: vector2);
var
  x: Vector3;
begin
  x[2] := -(gNewClip.u[3] * param[1] + gNewClip.u[1] * param[2] + gNewClip.a)
    / gNewClip.u[2];
  x[3] := param[1];
  x[1] := param[2];
  value := gTheF1.V(x);
end;

procedure ClipValueImpl3(var value: double; param: vector2);
var
  x: Vector3;
begin
  x[3] := -(gNewClip.u[1] * param[1] + gNewClip.u[2] * param[2] + gNewClip.a)
    / gNewClip.u[3];
  x[1] := param[1];
  x[2] := param[2];
  value := gTheF1.V(x);
end;


procedure FindClipCut;
var
  i: Integer;
  minCorner, maxCorner: Vector2;
begin
  ellipseCenter2D := gFirstPtDrawing2D;
  ellipseCenter3D := gFirstPtDrawing3D;
  InitializeClipCut;
  if gType = Implicit then begin
    if abs(gNewClip.u[1]) >= abs(gNewClip.u[2]) then begin
      if (abs(gNewClip.u[2]) >= abs(gNewClip.u[3])) or (abs(gNewClip.u[1])
        >= abs(gNewClip.u[3])) then
        i := 1
      else
        i := 3
    end
    else begin
      if (abs(gNewClip.u[1]) >= abs(gNewClip.u[3])) or (abs(gNewClip.u[2])
        >= abs(gNewClip.u[3])) then
        i := 2
      else
        i := 3
    end;
    case i of
      1: begin
          minCorner[1] := gMin3D[2];
          minCorner[2] := gMin3D[3];
          maxCorner[1] := gMax3D[2];
          maxCorner[2] := gMax3D[3];
          theImplicitCurve.ComputeImplicitCurve(ClipValueImpl1,
            CheckRangeClipImpl1,
            RecordParamsImpl1, gVerySmall, minCorner, maxCorner, kppLine2D, 4
            * kppLine2D,
            False)
        end;
      2: begin
          minCorner[1] := gMin3D[3];
          minCorner[2] := gMin3D[1];
          maxCorner[1] := gMax3D[3];
          maxCorner[2] := gMax3D[1];
          theImplicitCurve.ComputeImplicitCurve(ClipValueImpl2,
            CheckRangeClipImpl2,
            RecordParamsImpl2, gVerySmall, minCorner, maxCorner, kppLine2D, 4
            * kppLine2D,
            False)
        end;
      3: begin
          minCorner[1] := gMin3D[1];
          minCorner[2] := gMin3D[2];
          maxCorner[1] := gMax3D[1];
          maxCorner[2] := gMax3D[2];
          theImplicitCurve.ComputeImplicitCurve(ClipValueImpl3,
            CheckRangeClipImpl3,
            RecordParamsImpl3, gVerySmall, minCorner, maxCorner, kppLine2D, 4
            * kppLine2D,
            False)

        end;
    end;
  end
  else
    theImplicitCurve.ComputeImplicitCurve(ClipValueParam, CheckRangeClipParam,
      RecordParamsParam, gVerySmall, gMin2D, gMax2D, kppLine2D, 4 * kppLine2D,
      False);
end;

procedure DrawCutSegmentParamSurf2D(canv: TCanvas; par1, par2: vector2;
  vec1, vec2: Vector3Single);
var
  i: Integer;
  pixel1, pixel2: Tpoint;
  vis1, vis2: boolean;
  half, half1, half2: Vector3Single;
  parhalf, parhalf1, parhalf2: Vector2;
begin
  FromParametersToDrawing(par1, pixel1);
  FromParametersToDrawing(par2, pixel2);
  vis1 := True;
  vis2 := True;
  if gClips <> nil then
    for i := 0 to gClipTop do begin
      vis1 := vis1 and (HeightOverClipS(gClips[i], vec1) < 0);
      vis2 := vis2 and (HeightOverClipS(gClips[i], vec2) < 0);
    end;
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      canv.moveto(pixel1.x, pixel1.y);
      canv.lineto(pixel2.x, pixel2.y);
    end
    else if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then
    begin
      for i := 1 to 2 do begin
        half[i] := 0.5 * (vec1[i] + vec2[i]);
        parhalf[i] := 0.5 * (par1[i] + par2[i]);
      end;
      half[3] := 0.5 * (vec1[3] + vec2[3]);
      half1 := vec1;
      half2 := half;
      parhalf1 := par1;
      parhalf2 := parhalf;
      DrawCutSegmentParamSurf2D(canv, parhalf1, parhalf2, half1,
        half2);
      half1 := half;
      half2 := vec2;
      parhalf1 := parhalf;
      parhalf2 := par2;
      DrawCutSegmentParamSurf2D(canv, parhalf1, parhalf2, half1,
        half2);
    end;
  end;
end;

procedure DrawCutSegmentParamSurf3D(canv: TCanvas; vec1, vec2:
  Vector3Single);
var
  i: Integer;
  p1, p2: single;
  pixel1, pixel2: Tpoint;
  vis1, vis2: boolean;
  half, half1, half2: Vector3Single;
begin
  pixel1 := ProjectSToPaneIntAndDist(vec1, p1);
  pixel2 := ProjectSToPaneIntAndDist(vec2, p2);
  vis1 := True;
  vis2 := True;
  if gClips <> nil then
    for i := 0 to gClipTop do begin
      vis1 := vis1 and (HeightOverClipS(gClips[i], vec1) < 0);
      vis2 := vis2 and (HeightOverClipS(gClips[i], vec2) < 0);
    end;
  vis1 := vis1 and PtInRect(gRect, pixel1) and (p1 < 1.001 *
    gZBuffer^[pixel1.x,
    pixel1.y]);
  vis2 := vis2 and PtInRect(gRect, pixel2) and (p2 < 1.001 *
    gZBuffer^[pixel2.x,
    pixel2.y]);
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      canv.moveto(pixel1.x, pixel1.y);
      canv.lineto(pixel2.x, pixel2.y);
    end
    else if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then
    begin
      for i := 1 to 3 do
        half[i] := 0.5 * (vec1[i] + vec2[i]);
      half1 := vec1;
      half2 := half;
      DrawCutSegmentParamSurf3D(canv, half1, half2);
      half1 := half;
      half2 := vec2;
      DrawCutSegmentParamSurf3D(canv, half1, half2);
    end;
  end;
end;

procedure DrawCutSegmentImplicitSurf(canv: TCanvas; vec1, vec2:
  Vector3Single);
var
  i: Integer;
  pixel1, pixel2: Tpoint;
  vis1, vis2: boolean;
  half, half1, half2: Vector3Single;
  inside: double;
begin
  pixel1 := ProjectSToPaneInt(vec1);
  pixel2 := ProjectSToPaneInt(vec2);
  vis1 := PtInRect(gRect, pixel1) and IsVisibleImplicitS(gViewpoint, vec1,
    gVPCubeKey,
    true) and InRangeAndInClipsImplicitSurfaceS(inside, vec1);
  vis2 := PtInRect(gRect, pixel2) and IsVisibleImplicitS(gViewpoint, vec2,
    gVPCubeKey,
    true) and InRangeAndInClipsImplicitSurfaceS(inside, vec2);
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      canv.moveto(pixel1.x, pixel1.y);
      canv.lineto(pixel2.x, pixel2.y);
    end
    else if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then
    begin
      for i := 1 to 3 do
        half[i] := 0.5 * (vec1[i] + vec2[i]);
      half1 := vec1;
      half2 := half;
      DrawCutSegmentImplicitSurf(canv, half1, half2);
      half1 := half;
      half2 := vec2;
      DrawCutSegmentImplicitSurf(canv, half1, half2);
    end;
  end;
end;

procedure DrawCutSegmentWithoutClipsParamSurf(canv: TCanvas; vec1, vec2:
  Vector3);
var
  i: Integer;
  p1, p2: single;
  pixel1, pixel2: Tpoint;
  vis1, vis2: boolean;
  half, half1, half2: Vector3;
begin
  pixel1 := ProjectToPaneIntAndDist(vec1, p1);
  pixel2 := ProjectToPaneIntAndDist(vec2, p2);
  vis1 := PtInRect(gRect, pixel1) and (p1 < 1.001 * gZBuffer^[pixel1.x,
    pixel1.y]);
  vis2 := PtInRect(gRect, pixel2) and (p2 < 1.001 * gZBuffer^[pixel2.x,
    pixel2.y]);
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      canv.moveto(pixel1.x, pixel1.y);
      canv.lineto(pixel2.x, pixel2.y);
    end
    else if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then
    begin
      for i := 1 to 3 do
        half[i] := 0.5 * (vec1[i] + vec2[i]);
      half1 := vec1;
      half2 := half;
      DrawCutSegmentWithoutClipsParamSurf(canv, half1, half2);
      half1 := half;
      half2 := vec2;
      DrawCutSegmentWithoutClipsParamSurf(canv, half1, half2);
    end;
  end;
end;

procedure DrawCutSegmentWithoutClipsImplSurf(canv: TCanvas; vec1, vec2:
  Vector3);
var
  i: Integer;
  pixel1, pixel2: Tpoint;
  vis1, vis2: boolean;
  half, half1, half2: Vector3;
begin
  pixel1 := ProjectToPaneInt(vec1);
  pixel2 := ProjectToPaneInt(vec2);
  vis1 := PtInRect(gRect, pixel1) and IsVisibleImplicit(gViewpoint, vec1,
    gVPCubeKey,
    false);
  vis2 := PtInRect(gRect, pixel2) and IsVisibleImplicit(gViewpoint, vec2,
    gVPCubeKey,
    false);
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      canv.moveto(pixel1.x, pixel1.y);
      canv.lineto(pixel2.x, pixel2.y);
    end
    else if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then
    begin
      for i := 1 to 3 do
        half[i] := 0.5 * (vec1[i] + vec2[i]);
      half1 := vec1;
      half2 := half;
      DrawCutSegmentWithoutClipsImplSurf(canv, half1, half2);
      half1 := half;
      half2 := vec2;
      DrawCutSegmentWithoutClipsImplSurf(canv, half1, half2);
    end;
  end;
end;

procedure DrawCutParamSurface(fromDraw3D: Boolean);
var
  i: Integer;
  point1, point2: Vector3Single;
  canv3D, canv2D: TCanvas;
  param1, param2: vector2;
  goodptPanel, goodPtParam: Boolean;
  d: single;
begin
  if ((gGeomOption = NormalSection)) and (gLastSegmCC < 0) then
    Exit;
  canv3D := surface.Drawing3D.Canvas;
  canv3D.Draw(0, 0, gStdBmp);
  canv3D.pen.mode := pmCopy;
  canv3D.Pen.Color := clWhite;

  canv2D := surface.Drawing2D.Canvas;
  canv2D.FillRect(gRect);
  canv2D.pen.mode := pmCopy;
  canv2D.Pen.Color := clGreen;
  i := 0;
  point2 := gClipCut[i].pos;
  param2 := gCCParams[i];

  while i < gLastSegmCC do begin
    point1 := point2;
    param1 := param2;
    inc(i);
    point2 := gClipCut[i].pos;
    param2 := gCCParams[i];
    while (gClipCut[i].key = 1) and (i < gLastSegmCC) do begin
      inc(i);
      point1 := point2;
      param1 := param2;
      point2 := gClipCut[i].pos;
      param2 := gCCParams[i];
    end;
    if (gClipCut[i].key = 0) then begin
      DrawCutSegmentParamSurf3D(canv3D, point1, point2);
      DrawCutSegmentParamSurf2D(canv2D, param1, param2, point1,
        point2);
    end;
  end;
  if (gGeomOption = NormalSection) then begin
    if fromDraw3D then begin
      goodptPanel := PickedPtParametricS(ellipseCenter3D, param1, point1);
      goodPtParam := goodptPanel;
    end
    else begin
      goodPtParam := True;
      FromDrawingXYToParameters(ellipseCenter2D.x, ellipseCenter2D.y, param1);
      goodptPanel := InRangeAndClipsS(point1, param1);
      ProjectSToPaneIntAndDist(point1, d);
      goodptPanel := goodptPanel and (d < 1.001 * gZBuffer^[ellipseCenter3D.x,
        ellipseCenter3D.y]);
    end;
    if goodPtParam then begin
      canv2D.Pen.Color := clRed;
      canv2D.Ellipse(ellipseCenter2D.x - 3, ellipseCenter2D.y - 3,
        ellipseCenter2D.x + 3,
        ellipseCenter2D.y
        + 3);
    end;
    if goodptPanel then begin
      canv3D.Pen.Color := clRed;
      canv3D.Ellipse(ellipseCenter3D.x - 3, ellipseCenter3D.y - 3,
        ellipseCenter3D.x + 3,
        ellipseCenter3D.y
        + 3);
    end;
  end
  else
    DrawClipEllipse;
end;

procedure DrawCutImplicitSurface;
var
  i: Integer;
  point1, point2: Vector3Single;
  canv: TCanvas;
begin
  if (gGeomOption = NormalSection) and (gLastSegmCC < 0) then
    Exit;
  canv := surface.Drawing3D.Canvas;
  canv.Draw(0, 0, gStdBmp);
  canv.pen.mode := pmCopy;
  canv.Pen.Color := clWhite;
  i := 0;
  point2 := gClipCut[i].pos;
  while i < gLastSegmCC do begin
    point1 := point2;
    inc(i);
    point2 := gClipCut[i].pos;
    while (gClipCut[i].key = 1) and (i < gLastSegmCC) do begin
      inc(i);
      point1 := point2;
      point2 := gClipCut[i].pos;
    end;
    if (gClipCut[i].key = 0) then begin
      if gType = Implicit then
        DrawCutSegmentImplicitSurf(canv, point1, point2)
      else
        DrawCutSegmentParamSurf3D(canv, point1, point2);
    end;
  end;
  if (gGeomOption = NormalSection) then begin
    canv.Pen.Color := clRed;
    canv.Ellipse(ellipseCenter3D.x - 3, ellipseCenter3D.y - 3,
      ellipseCenter3D.x + 3, ellipseCenter3D.y
      + 3);
  end
  else
    DrawClipEllipse;
end;

procedure ComputeDataEllipseClip(fullData: Boolean);
var
  i: Integer;
  a: double;
begin
  uBE := -HeightOverClip(gNewClip, gCenter3D);
  uRE := uR2 - sqr(uBE);
  if urE > 0 then
    uRE := sqrt(uRE);
  if fullData then begin
    up := EuclCrossProduct(gNewClip.u, gN);
    a := EuclNorm3(up);
    if a > 1.0E-6 then begin
      uq := EuclCrossProduct(gNewClip.u, up);
      a := 1 / a;
      for i := 1 to 3 do begin
        up[i] := up[i] * a;
        uq[i] := uq[i] * a;
      end
    end
    else begin
      up := gH;
      uq := gV;
    end;
  end;
end;

procedure DrawSpecialSegment(canv: Tcanvas; p1, p2: Vector3);
var
  i: Integer;
  pixel1, pixel2: Tpoint;
  vis1, vis2, underClip1, underClip2: boolean;
  half, half1, half2: Vector3;
begin
  pixel1 := ProjectToPaneInt(p1);
  pixel2 := ProjectToPaneInt(p2);
  vis1 := PtInRect(gRect, pixel1);
  vis2 := PtInRect(gRect, pixel2);
  if (vis1 or vis2) then begin
    underClip1 := HeightOverClip(gNewClip, p1) < 0;
    underClip2 := HeightOverClip(gNewClip, p2) < 0;
    if vis1 and vis2 then begin
      if underClip1 and underClip2 then
        canv.pen.Color := clYellow
      else if underClip1 = underClip2 then
        canv.pen.Color := clSilver;
      if underClip1 = underClip2 then begin
        canv.moveto(pixel1.x, pixel1.y);
        canv.lineto(pixel2.x, pixel2.y);
      end;
    end;
    if ((vis1 <> vis2) or (underClip1 <> underClip2)) and (abs(pixel2.x -
      pixel1.x)
      + abs(pixel2.y - pixel1.y) >
      3) then begin
      for i := 1 to 3 do
        half[i] := 0.5 * (p1[i] + p2[i]);
      half1 := p1;
      half2 := half;
      DrawSpecialSegment(canv, half1, half2);
      half1 := half;
      half2 := p2;
      DrawSpecialSegment(canv, half1, half2);
    end;
  end;
end;

procedure DrawCircle;
var
  p1, p2: Vector3;
  h: double;
  i, j: Integer;
  canv: TCanvas;
begin
  canv := surface.Drawing3D.canvas;
  canv.Pen.Color := clSilver;
  h := 2 * pi / 50;
  for j := 1 to 3 do
    p2[j] := cContour[j] + rContour * gH[j];
  for i := 1 to 50 do begin
    p1 := p2;
    for j := 1 to 3 do
      p2[j] := cContour[j] + rContour * (gH[j] * cos(i * h) +
        gV[j] * sin(i * h));
    DrawSpecialSegment(canv, p1, p2);
  end;

end;

procedure DrawClipEllipse;
var
  i, j: Integer;
  v1, v2: Vector3;
  h: double;
  canv: TCanvas;
begin
  canv := surface.Drawing3D.canvas;
  canv.Pen.Color := clSilver;
  canv.Brush.Style := BSClear;
  DrawCircle;
  h := 2 * pi / 50;
  canv.Brush.Style := BSSolid;
  if uRE <= 0 then Exit;
  if toMove <> 0 then
    canv.Pen.Color := cOrange;
  for j := 1 to 3 do
    v2[j] := gCenter3D[j] + uBE * gNewClip.u[j] + uRE * up[j];
  for i := 1 to 50 do begin
    v1 := v2;
    for j := 1 to 3 do
      v2[j] := gCenter3D[j] + uBE * gNewClip.u[j] + uRE * (up[j] * cos(i
        * h) + uq[j] * sin(i * h));
    if gType = Implicit then
      DrawCutSegmentWithoutClipsImplSurf(canv, v1, v2)
    else
      DrawCutSegmentWithoutClipsParamSurf(canv, v1, v2);
  end;
end;

procedure PrepareToSetClip;
const
  epsilon = 2;
var
  j: Integer;
  co, si: Extended;
begin
  if gShowCube then
    surface.OnDrawCube(nil);
  gDist2 := sqr(gDist);
  uNear := abs(gDist * epsilon / gC0);
  c0D := gC0 * gDist;
  c0D2 := sqr(c0D);
  uR2 := sqr(gSpan3D / 2.2);
  dcContour := ur2 / gDist;
  rContour := sqrt(ur2 * (1 - ur2 / gDist2));
  for j := 1 to 3 do
    cContour[j] := gCenter3D[j] - dcContour * gN[j];
  ComputeDataEllipseClip(True);
  tDefault := (gDist2 - uR2) / c0D;
  toMove := 0;
  sincos(3.65, si, co);
  for j := 1 to 3 do
    gNewClip.u[j] := gV[j] * co + gN[j] * si;
  gNewClip.a := -EuclInnerProd3(gCenter3D, gNewClip.u);
end;

procedure FindWhatToMove(pt: TPoint);
var
  k: Vector3;
  a, HeightOverClip, discr, t: double;
begin
  toMove := 0;
  k := FindRay(pt);
  a := EuclInnerProd3(k, k);
  discr := c0D2 - a * (gDist2 - uR2);
  if discr < 0 then begin
    t := tDefault;
    HeightOverClip := (gViewpoint[1] + t * k[1]) * gNewClip.u[1] +
      (gViewpoint[2] + t * k[2])
      *
      gNewClip.u[2] +
      (gViewpoint[3] + t * k[3]) * gNewClip.u[3] + gNewClip.a;
    if abs(HeightOverClip) < uNear then
      toMove := 3;
  end
  else begin
    discr := sqrt(discr);
    t := (c0D - discr) / a;
    HeightOverClip := (gViewpoint[1] + t * k[1]) * gNewClip.u[1] +
      (gViewpoint[2] + t * k[2])
      *
      gNewClip.u[2] +
      (gViewpoint[3] + t * k[3]) * gNewClip.u[3] + gNewClip.a;
    if abs(HeightOverClip) < uNear then
      toMove := 1
    else begin
      t := (c0D + discr) / a;
      HeightOverClip := (gViewpoint[1] + t * k[1]) * gNewClip.u[1] +
        (gViewpoint[2] + t *
        k[2])
        *
        gNewClip.u[2] +
        (gViewpoint[3] + t * k[3]) * gNewClip.u[3] + gNewClip.a;
      if abs(HeightOverClip) < uNear then
        toMove := 2
    end
  end
end;

procedure PickedPtInSphere(pt: TPoint; var pe: Vector3);
var
  i: Integer;
  k: Vector3;
  a, discr, t: double;
begin
  k := FindRay(pt);
  a := EuclInnerProd3(k, k);
  discr := c0D2 - a * (gDist2 - uR2);
  if discr < 0 then
    t := tDefault
  else begin
    discr := sqrt(discr);
    if toMove = 1 then
      t := (c0D - discr) / a
    else
      t := (c0D + discr) / a;
  end;
  for i := 1 to 3 do
    pe[i] := gViewpoint[i] + t * k[i];
end;

procedure PickedUnitVector(pt: TPoint; var k: Vector3);
var
  i: Integer;
  dir: Vector3;
  a, discr, t: double;
begin
  dir := FindRay(pt);
  a := EuclInnerProd3(dir, dir);
  discr := c0D2 - a * (gDist2 - uR2);
  if discr < 0 then
    t := tDefault
  else begin
    discr := sqrt(discr);
    t := (c0D - discr) / a
  end;
  for i := 1 to 3 do
    k[i] := gViewpoint[i] + t * dir[i] - gCenter3D[i];
  a := 1 / EuclNorm3(k);
  for i := 1 to 3 do
    k[i] := a * k[i];
end;

procedure MoveClip(pt1, pt2: TPoint);
var
  i: Integer;
  k, e1, e2, b: Vector3;
  norm, e2b, twoVe2, vb: double;
begin
  if toMove = 0 then begin
    PickedUnitVector(pt1, e2);
    PickedUnitVector(pt2, e1);
    norm := sqrt(sqr(e1[1] + e2[1]) + sqr(e1[2] + e2[2]) + sqr(e1[3] +
      e2[3]));
    for i := 1 to 3 do
      b[i] := (e1[i] + e2[i]) / norm;
    e2b := e2[1] * b[1] + e2[2] * b[2] + e2[3] * b[3];
    twoVe2 := 2 * EuclInnerProd3(gNewClip.u, e2);
    vb := EuclInnerProd3(gNewClip.u, b);
    vb := 2 * (twoVe2 * e2b - vb);
    e2b := HeightOverClip(gNewClip, gCenter3D);
    for i := 1 to 3 do
      gNewClip.u[i] := vb * b[i] - twoVe2 * e2[i] + gNewClip.u[i];
    norm := 1 / EuclNorm3(gNewClip.u);
    for i := 1 to 3 do
      gNewClip.u[i] := gNewClip.u[i] * norm;
    gNewClip.a := e2b - EuclInnerProd3(gNewClip.u, gCenter3D);
  end
  else begin
    PickedPtInSphere(pt2, k);
    gNewClip.a := -EuclInnerProd3(gNewClip.u, k);
  end;
  FindClipCut;
  ComputeDataEllipseClip(toMove = 0);
  if gType = Implicit then
    DrawCutImplicitSurface
  else
    DrawCutParamSurface(false);
end;


end.
