

{-----------------------------------------------------------------------------
 Unit Name: ParametricUtils
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Utilities specific for surfaces given in parametric or tube form
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit ParametricUtils;
interface
uses
{$IFDEF MSWINDOWS}
  Windows, Graphics, stdctrls,
{$ENDIF}
{$IFDEF LINUX}
  Types,  QGraphics, Qstdctrls,
{$ENDIF}
  Globals;

function CheckRangeParametric(lastX, nextX: Vector3; findProlong: Boolean;
  var newX, newVel: Vector3): Integer;
function CheckCameBackParametricSurface(firstX, prevX, curX:
  Vector3; counter: Integer): Boolean;


function InRangeAndClips(var x: Vector3; param: vector2)
  : Boolean;
function InRangeAndClipsS(var x: Vector3Single; param: vector2)
  : Boolean;
procedure ComputeEFG(p: vector2; var E, F, G: double);
procedure ComputeEFGAndPartials(p: vector2; var E, F, G: double; var xu, xv:
  Vector3);
function ComputeEFGefg(p: vector2; var E, F, G, eII, fII, gII: double; var
  x, xu, xv: Vector3): Boolean;
function DirectionsParamSurf(var deriv: Vector3; prevDeriv, here: Vector3;
  axis: Integer): Boolean;
procedure DrawXParamSurface;
procedure ComputeCurvaturesParamSurface(p: Vector2; pos: Vector3);
procedure EraseCross2D;
procedure DrawCross2D;
procedure EraseCross3D;
procedure DrawCross3D;
function ComputeClipParamSurface(p1, p2: TPoint; FromDrawing3D: Boolean; var
  aClip: Clip):
  Boolean;
procedure AddClipParamSurface(param, paramVel: vector2);

function NextPoint(lastX, nextX: Vector3; var
  newX, newVel: Vector3): Boolean;
function PickedPtParametric(mousePt: TPoint;
  var paramPt: vector2;
  var pos: Vector3): Boolean;
function PickedPtParametricS(mousePt: TPoint;
  var paramPt: vector2;
  var pos: Vector3Single): Boolean;
function FindVelFor2OrderCurvesParamSurf(var paramVel: Vector3): Boolean;
procedure DrawCoordLinesParamSurface(param:
  Vector2);
procedure DrawCurvesParamSurface(fromHere: vector2);
procedure DrawMeshParamSurface;
procedure DrawSegmentParamSurface3D(vec1, vec2:
  Vector3; g: gamma16Ptr; toSave: Boolean);
procedure RecordAndDrawSegmentParamSurface23D(x1, x2: Vector3; toSave:
  Boolean);
function ComputeChristSymbParam(p: vector3; var metric, gammaU, gammaV:
  Vector3):
  BOOLEAN;
function GeodAccelParametricSurf(var accel: Vector3; vel, x: vector3; toRight:
  Integer):
  Boolean;
function ComputeFrameParamSurface(var u1, u2: Vector3; var phi, topT: Double;
  x: Vector3): double;
{Given the vector 'vel'in the parameter space (3th dimension don't count) at the
parameter point 'x', 'orthToVel' shall return a vector at x that is
orthogonal to 'vel', and with same length. The result is the determinant of the
first fundamental form at 'x'.}

procedure ComputeWireParametricSurface;

implementation
uses
  Math,  SysUtils,
{$IFDEF MSWINDOWS}
   Controls,
{$ENDIF}
{$IFDEF LINUX}
  QControls,
{$ENDIF}
  OcFunct, Utilities,  FSurface, FRunKut;

function CheckRangeParametric(lastX, nextX: Vector3; findProlong: Boolean;
  var newX, newVel: Vector3): Integer;
var
  i: Integer;
  pos: Vector3;
  inside, inClips: Boolean;
begin
  inside := Min(gHalfEdges2D[1] - abs(nextX[1] - gCenter2D[1]),
    gHalfEdges2D[2] - abs(nextX[2] - gCenter2D[2])) >= 0;
  theFPar(pos, nextX[1], nextX[2]);
  inClips := True;
  if inside and (gClips <> nil) then
    for i := 0 to gClipTop do
      inClips := inClips and (HeightOverClip(gClips[i], pos) < 0);
  result := ord(inside and inClips);
  if inClips and (not inside) and findProlong and
    NextPoint(lastX, nextX, newX, newVel) then begin
    result := 2;
  end;
end;

function CheckCameBackParametricSurface(firstX, prevX, curX:
  Vector3; counter: Integer): Boolean;
var
  bb, bc, cc, aux: Double;
begin
  result := False;
  if counter < 25 then
    Exit;
  bb := sqr(firstX[1] - prevX[1]) + sqr(firstX[2] - prevX[2]);
  if sqrt(bb) < gMaxSegmentLength2D then begin
    bc := (firstX[1] - prevX[1]) * (curX[1] - prevX[1]) + (firstX[2] -
      prevX[2]) * (curX[2] - prevX[2]);
    if bc > 0 then begin
      cc := sqr(curX[1] - prevX[1]) + sqr(curX[2] - prevX[2]);
      if cc > bc then begin
        aux := sqrt(abs(bb - sqr(bc) / cc));
        result := aux < 2 * gRadius2D;
      end;
    end;
  end;
end;

function InClips(var x: Vector3; param: vector2): Boolean;
var
  i: Integer;
begin
  result := True;
  theFPar2(x, param);
  if gClips <> nil then
    for i := 0 to gClipTop do
      Result := Result and (HeightOverClip(gClips[i], x) < 0);
end;

function InRangeAndClips(var x: Vector3; param: vector2)
  : Boolean;
var
  i: Integer;
  depth: Double;
begin
  depth := min(gHalfEdges2D[1] - abs(param[1] - gCenter2D[1]),
    gHalfEdges2D[2] - abs(param[2] - gCenter2D[2]));
  theFPar2(x, param);
  Result := (depth >= 0);
  if gClips <> nil then
    for i := 0 to gClipTop do
      Result := Result and (HeightOverClip(gClips[i], x) < 0);
end;

function InRangeAndClipsS(var x: Vector3Single; param: vector2)
  : Boolean;
var
  i: Integer;
  y: Vector3;
  depth: Double;
begin
  depth := min(gHalfEdges2D[1] - abs(param[1] - gCenter2D[1]),
    gHalfEdges2D[2] - abs(param[2] - gCenter2D[2]));
  theFPar2(y, param);
  Result := (depth >= 0);
  if gClips <> nil then
    for i := 0 to gClipTop do
      Result := Result and (HeightOverClip(gClips[i], y) < 0);
  for i := 1 to 3 do
    x[i] := y[i];
end;

procedure ComputeEFG(p: vector2; var E, F, G: double);
{Returns at 'E,F,G' the coefficients of the first fundamental form.}
var
  xMinus, xPlus, xu, xv: Vector3;
  i: Integer;
begin
  theFPar(xMinus, p[1] - gIncr2D[1], p[2]);
  theFPar(xPlus, p[1] + gIncr2D[1], p[2]);
  for i := 1 to 3 do
    xu[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[1]);
  theFPar(xMinus, p[1], p[2] - gIncr2D[2]);
  theFPar(xPlus, p[1], p[2] + gIncr2D[2]);
  for i := 1 to 3 do
    xv[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[2]);
  E := InnerProd3(xu, xu);
  F := InnerProd3(xu, xv);
  G := InnerProd3(xv, xv);
end;

procedure ComputePartials(p: vector2; var xu, xv: Vector3);
{Returns  'xu, xv' the partial derivatives of the map 'theFPar'.}
var
  xMinus, xPlus: Vector3;
  i: Integer;
begin
  theFPar(xMinus, p[1] - gIncr2D[1], p[2]);
  theFPar(xPlus, p[1] + gIncr2D[1], p[2]);
  for i := 1 to 3 do
    xu[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[1]);
  theFPar(xMinus, p[1], p[2] - gIncr2D[2]);
  theFPar(xPlus, p[1], p[2] + gIncr2D[2]);
  for i := 1 to 3 do
    xv[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[2]);
end;

procedure ComputeEFGAndPartials(p: vector2; var E, F, G: double; var xu, xv:
  Vector3);
var
  xMinus, xPlus: Vector3;
  i: Integer;
begin
  theFPar(xMinus, p[1] - gIncr2D[1], p[2]);
  theFPar(xPlus, p[1] + gIncr2D[1], p[2]);
  for i := 1 to 3 do
    xu[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[1]);
  theFPar(xMinus, p[1], p[2] - gIncr2D[2]);
  theFPar(xPlus, p[1], p[2] + gIncr2D[2]);
  for i := 1 to 3 do
    xv[i] := (xPlus[i] - xMinus[i]) / (2 * gIncr2D[2]);
  E := InnerProd3(xu, xu);
  F := InnerProd3(xu, xv);
  G := InnerProd3(xv, xv);
end;

function ComputeEFGefg(p: vector2; var E, F, G, eII, fII, gII: double; var
  x, xu, xv: Vector3): Boolean;
var
  xMinusU, xPlusU, xMinusV, xPlusV, xPlusUV, xMinusUV, xuu, xuv, xvv, normal:
  Vector3;
  i: Integer;
  q: double;
begin
  Result := False;
  theFPar(x, p[1], p[2]);
  theFPar(xMinusU, p[1] - gIncr2D[1], p[2]);
  theFPar(xPlusU, p[1] + gIncr2D[1], p[2]);
  theFPar(xMinusV, p[1], p[2] - gIncr2D[2]);
  theFPar(xPlusV, p[1], p[2] + gIncr2D[2]);
  theFPar(xMinusUV, p[1] - gIncr2D[1], p[2] - gIncr2D[2]);
  theFPar(xPlusUV, p[1] + gIncr2D[1], p[2] + gIncr2D[2]);
  for i := 1 to 3 do begin
    xu[i] := (xPlusU[i] - xMinusU[i]) / (2 * gIncr2D[1]);
    xuu[i] := (xPlusU[i] + xMinusU[i] - 2 * x[i]) / sqr(gIncr2D[1]);
    xv[i] := (xPlusV[i] - xMinusV[i]) / (2 * gIncr2D[2]);
    xvv[i] := (xPlusV[i] + xMinusV[i] - 2 * x[i]) / sqr(gIncr2D[2]);
    xuv[i] := 2 * (xPlusUV[i] + xMinusUV[i] - xMinusU[i] - xPlusU[i] -
      xMinusV[i]
      - xPlusV[i] + 2 * x[i]) / g4Incr1Incr2;
  end;
  normal := CrossProduct(xu, xv);
  q := sqrt(abs(InnerProd3(normal, normal)));
  if q < 1.0E-10 then
    Exit;
  for i := 1 to 3 do
    normal[i] := normal[i] / q;

  E := InnerProd3(xu, xu);
  F := InnerProd3(xu, xv);
  G := InnerProd3(xv, xv);
  eII := InnerProd3(normal, xuu);
  fII := InnerProd3(normal, xuv);
  gII := InnerProd3(normal, xvv);

  Result := True;
end;

procedure ComputeCurvaturesParamSurface(p: Vector2; pos: Vector3);
 {Gives in 'gKStr, gHStr, gK1Str, gK2Str' the Gauss curvature, mean curv.,}
 {and both principal curvatures at the param value 'p'.}
var
  E, F, G, eII, fII, gII, det, disc: double;
  x, xu, xv: Vector3;
  good: Boolean;
  gg, auxH, auxK: double;
begin
  gKStr := '';
  gHStr := '';
  gK1Str := '';
  gK2Str := '';
  good := ComputeEFGefg(p, E, F, G, eII, fII, gII, x, xu, xv);
  if not good then
    Exit;
  det := E * G - sqr(F);
  auxK := (eII * gII - sqr(fII)) / det;
  if det > 0 then
    gKStr := FloatToStrF(auxK, fffixed, 8, 4)
  else
    gKStr := kBullet + FloatToStrF(auxK, fffixed, 7, 4);
  auxH := (eII * G - 2 * fII * F + gII * E) / (2 * det);
  gHStr := FloatToStrF(auxH, fffixed, 8, 4);
  disc := sqrt(abs(sqr(auxH) - auxK));
  gg := auxH + disc;
  gK1Str := FloatToStrF(gg, fffixed, 8, 4);
  gg := auxH - disc;
  gK2Str := FloatToStrF(gg, fffixed, 8, 4);
end;

procedure EraseCross3D;
var
  can: TCanvas;
begin
  if gThereIsCross3D then begin
    can := surface.Drawing3D.canvas;
    can.Pen.Color := clWhite;
    can.pen.Mode := pmXor;
    can.moveto(gCross3D[1].x, gCross3D[1].y);
    can.lineto(gCross3D[2].x, gCross3D[2].y);
    can.moveto(gCross3D[3].x, gCross3D[3].y);
    can.lineto(gCross3D[4].x, gCross3D[4].y);
    can.pen.Mode := pmCopy;
    gThereIsCross3D := False;
  end;
end;

procedure DrawCross3D;
var
  i: Integer;
  p: single;
  xu, xv, pos1, pos2, pos3, pos4: Vector3;
  can: TCanvas;
  aux: double;
begin
  EraseCross3D;
  if not ptInRect(gRect, gCurPtDrawing3D) then
    Exit;
  ProjectToPaneIntAndDist(gCur3DPos, p);
  if (p < 1.001 * gZBuffer^[gCurPtDrawing3D.x, gCurPtDrawing3D.y]) then begin
    ComputePartials(gCur2DPos, xu, xv);
    aux := EuclNorm3(xu);
    if aux > 1.0E-7 then
      p := 2 * gMaxSegmentLength3D / aux
    else
      p := 0;
    for i := 1 to 3 do
      xu[i] := xu[i] * p;
    aux := EuclNorm3(xv);
    if aux > 1.0E-7 then
      p := 2 * gMaxSegmentLength3D / aux
    else
      p := 0;
    for i := 1 to 3 do
      xv[i] := xv[i] * p;
    for i := 1 to 3 do begin
      pos1[i] := gCur3DPos[i] - xu[i];
      pos2[i] := gCur3DPos[i] + xu[i];
      pos3[i] := gCur3DPos[i] - xv[i];
      pos4[i] := gCur3DPos[i] + xv[i];
    end;
    gCross3D[0] := gCurPtDrawing3D;
    gCross3D[1] := ProjectToPaneInt(pos1);
    gCross3D[2] := ProjectToPaneInt(pos2);
    gCross3D[3] := ProjectToPaneInt(pos3);
    gCross3D[4] := ProjectToPaneInt(pos4);

    can := surface.Drawing3D.canvas;
    can.Pen.Color := clWhite;
    can.pen.Mode := pmXor;
    can.moveto(gCross3D[1].x, gCross3D[1].y);
    can.lineto(gCross3D[2].x, gCross3D[2].y);
    can.moveto(gCross3D[3].x, gCross3D[3].y);
    can.lineto(gCross3D[4].x, gCross3D[4].y);
    can.pen.Mode := pmCopy;
    gThereIsCross3D := True;
  end;
end;

procedure EraseCross2D;
var
  can: TCanvas;
begin
  if gThereIsCross2D then begin
    can := surface.Drawing2D.canvas;
    can.Pen.Color := clGreen;
    can.pen.Mode := pmNotXoR;
    can.moveto(gCross2D[1].x, gCross2D[1].y);
    can.lineto(gCross2D[2].x, gCross2D[2].y);
    can.moveto(gCross2D[3].x, gCross2D[3].y);
    can.lineto(gCross2D[4].x, gCross2D[4].y);
    can.pen.Mode := pmCopy;
    gThereIsCross2D := False;
  end;
end;

procedure DrawCross2D;
var
  can: TCanvas;
begin
  EraseCross2D;
  if not ptInRect(gRect, gCurPtDrawing2D) then
    Exit;
  gCross2D[0] := gCurPtDrawing2D;
  gCross2D[1].x := gCurPtDrawing2D.x + 5;
  gCross2D[1].y := gCurPtDrawing2D.y;
  gCross2D[2].x := gCurPtDrawing2D.x - 5;
  gCross2D[2].y := gCurPtDrawing2D.y;
  gCross2D[3].x := gCurPtDrawing2D.x;
  gCross2D[3].y := gCurPtDrawing2D.y + 5;
  gCross2D[4].x := gCurPtDrawing2D.x;
  gCross2D[4].y := gCurPtDrawing2D.y - 5;

  can := surface.Drawing2D.canvas;
  can.Pen.Color := clGreen;
  can.pen.Mode := pmNotXoR;
  can.moveto(gCross2D[1].x, gCross2D[1].y);
  can.lineto(gCross2D[2].x, gCross2D[2].y);
  can.moveto(gCross2D[3].x, gCross2D[3].y);
  can.lineto(gCross2D[4].x, gCross2D[4].y);
  can.pen.Mode := pmCopy;
  gThereIsCross2D := True;
end;



procedure DrawXParamSurface;
var
  i, j: integer;
  auxs: Single;
  aux, eLow, fLow, gLow: double;
  E, F, G, det, disc, K, H, k1, k2: double;
  xu, xv, pos1, pos2, pos3, pos4: Vector3;
  p: Segment2;
  tang: Segment3;
  can: TCanvas;

  procedure Normalize;
  var
    i, h, v: Integer;
  begin
    gX2D[0]:= gCurPtDrawing2D;
    for i := 1 to 2 do begin
      if gGeomOption <> LightRays then begin
        aux := sqrt(abs(E * sqr(p[i, 1]) + 2 * F * p[i, 1] * p[i, 2] + G *
          sqr(p[i, 2])));
        aux := max(aux, 1.0E-7);
        p[i, 1] := p[i, 1] / aux;
        p[i, 2] := p[i, 2] / aux;
      end
      else begin
        aux := max(0.01, hypot(p[i, 1], p[i, 2]));
        p[i, 1] := p[i, 1] / aux;
        p[i, 2] := p[i, 2] / aux;
      end;
      h := round(p[i, 1] * kWidth / gEdges2D[1]);
      v := round(p[i, 2] * kHeight / gEdges2D[2]);
      gX2D[2 * i].x := gCurPtDrawing2D.x + h;
      gX2D[2 * i].y := gCurPtDrawing2D.y - v;
      gX2D[2 * i - 1].x := gCurPtDrawing2D.x - h;
      gX2D[2 * i - 1].y := gCurPtDrawing2D.y + v;
    end;
  end;

begin
  EraseX;
  if (gGeomOption = LightRays) and (gEpsilon = 1) then
    Exit;
  if not ComputeEFGefg(gCur2DPos, E, F, G, eLow, fLow, gLow, pos1, xu, xv)
    then
    Exit;

  det := E * G - sqr(F);
  K := (elow * glow - sqr(flow)) / det;
  H := (elow * G - 2 * flow * F + glow * E) / (2 * det);
  disc := sqrt(abs(sqr(H) - K));
  k1 := H + disc;
  k2 := H - disc;
  gKstr := FloatToStrF(K, ffFixed, 10, 3);
  gHstr := FloatToStrF(H, ffFixed, 10, 3);
  gk1Str := FloatToStrF(k1, ffFixed, 10, 3);
  gk2Str := FloatToStrF(k2, ffFixed, 10, 3);
  surface.ShowCurvatures(False);

  if gGeomOption = CurvatureLines then begin
    aux := eLow;
    disc := fLow;
    eLow := fLow * E - eLow * F;
    fLow := (gLow * E - aux * G) / 2;
    gLow := gLow * F - disc * G;
  end
  else if gGeomOption = LightRays then begin
    eLow := E;
    fLow := F;
    gLow := G;
  end;
  det := sqr(fLow) - eLow * gLow;
  if det < 1.0E-14 then
    Exit;
  det := sqrt(det);
  if (abs(eLow) > 1.0E-8) and (abs(gLow) > 1.0E-8) then begin
    p[1, 1] := -gLow;
    p[1, 2] := det + fLow;
    p[2, 1] := det + fLow;
    p[2, 2] := -eLow;
  end
  else if abs(eLow) > 1.0E-8 then begin
    p[1, 1] := 0;
    p[1, 2] := 1;
    p[2, 1] := 2 * fLow;
    p[2, 2] := -eLow;
  end
  else if abs(gLow) > 1.0E-8 then begin
    p[1, 1] := 1;
    p[1, 2] := 0;
    p[2, 1] := gLow;
    p[2, 2] := -2 * fLow;
  end
  else begin
    p[1, 1] := 1;
    p[1, 2] := 0;
    p[2, 1] := 0;
    p[2, 2] := 1;
  end;
  Normalize;
  can := surface.Drawing2D.canvas;
  can.Pen.Color := clTeal;
  can.Pen.Mode := pmNotXoR;
  can.moveto(gX2D[1].x, gX2D[1].y);
  can.lineto(gX2D[2].x, gX2D[2].y);
  can.moveto(gX2D[3].x, gX2D[3].y);
  can.lineto(gX2D[4].x, gX2D[4].y);
  gThereIsX2D := True;

  if ptInRect(gRect, gCurPtDrawing3D) then begin
    ProjectToPaneIntAndDist(gCur3DPos, auxs);
    if (auxs < 1.001 * gZBuffer^[gCurPtDrawing3D.x, gCurPtDrawing3D.y]) then
    begin
      for i := 1 to 2 do
        for j := 1 to 3 do
          tang[i, j] := p[i, 1] * xu[j] + p[i, 2] * xv[j];
      for i := 1 to 2 do begin
        p[i, 1] := 4 * gMaxSegmentLength3D / (1.0E-12 + EuclNorm3(tang[i]));
        for j := 1 to 3 do
          tang[i][j] := tang[i][j] * p[i, 1];
      end;
      for i := 1 to 3 do begin
        pos1[i] := gCur3DPos[i] - tang[1][i];
        pos2[i] := gCur3DPos[i] + tang[1][i];
        pos3[i] := gCur3DPos[i] - tang[2][i];
        pos4[i] := gCur3DPos[i] + tang[2][i];
      end;
      gX3D[0] := gCurPtDrawing3D;
      gX3D[1] := ProjectToPaneInt(pos1);
      gX3D[2] := ProjectToPaneInt(pos2);
      gX3D[3] := ProjectToPaneInt(pos3);
      gX3D[4] := ProjectToPaneInt(pos4);
      can := surface.Drawing3D.canvas;
      can.Pen.Color := clBlack;
      can.Pen.Mode := pmNotXor;
      can.moveto(gX3D[1].x, gX3D[1].y);
      can.lineto(gX3D[2].x, gX3D[2].y);
      can.moveto(gX3D[3].x, gX3D[3].y);
      can.lineto(gX3D[4].x, gX3D[4].y);
      can.Pen.Mode := pmCopy;
      gThereIsX3D := True;
    end;
  end;
end;

function DirectionsParamSurf(var deriv: Vector3; prevDeriv, here: Vector3;
  axis: Integer): Boolean;
var
  i: integer;
  aux, eLow, fLow, gLow, l1, l2, p, q: double;
  good: Boolean;
  E, F, G: double;
  x, xu, xv: Vector3;
  param: Vector2;
begin
  Result := False;
  param[1] := here[1];
  param[2] := here[2];
  good := ComputeEFGefg(param, E, F, G, eLow, fLow, gLow, x, xu, xv);
  if not good then
    Exit;
  if gGeomOption = CurvatureLines then begin
    aux := eLow;
    q := fLow;
    eLow := fLow * E - eLow * F;
    fLow := (gLow * E - aux * G) / 2;
    gLow := gLow * F - q * G;
  end
  else if gGeomOption = LightRays then begin
    eLow := E;
    fLow := F;
    gLow := G;
  end;
  aux := sqr(fLow) - eLow * gLow;
  if aux < 1.0E-12 then
    Exit;
  aux := (1 - 2 * axis) * sqrt(aux);
  l1 := sqr(-fLow + aux) + sqr(eLow);
  l2 := sqr(fLow + aux) + sqr(gLow);
  if l1 < l2 then begin
    l2 := 1 / sqrt(l2);
    p := l2 * gLow;
    q := -(fLow + aux) * l2;
  end
  else begin
    l1 := 1 / sqrt(l1);
    p := (aux - fLow) * l1;
    q := eLow * l1;
  end;
  deriv[1] := p;
  deriv[2] := q;
  aux := deriv[1] * prevDeriv[1] + deriv[2] * prevDeriv[2];
  if aux < 0 then
    for i := 1 to 2 do
      deriv[i] := -deriv[i];
  Result := True;
end;


function GeodAccelParametricSurf(var accel: Vector3; vel, x: vector3; toRight:
  Integer):
  Boolean;
{Gives the acceleration  'accel' of a curve of constant geodesic curvature at 'x'}
{when it goes with velocity 'vel'. If  toRight = 0, that curve will be a geodesic;}
{else, its curvature radius will be proportional to the length of 'vel' and turn to}
{right if 'toRight =1' or left if it is -1. If cannot be computed, result will be}
{False.}
var
  metric, gammaU, gammaV: Vector3;
  externalAccel: vector2;
  det: double;
begin
  result := FALSE;
  if ComputeChristSymbParam(x, metric, gammaU, gammaV) then begin
    result := TRUE;
    accel[1] := -(gammaU[1] * sqr(vel[1]) + 2 * gammaU[2] * vel[1] *
      vel[2] + gammaU[3] * sqr(vel[2]));
    accel[2] := -(gammaV[1] * sqr(vel[1]) + 2 * gammaV[2] * vel[1] *
      vel[2] + gammaV[3] * sqr(vel[2]));
    if toRight <> 0 then begin
      det := sqrt(abs(metric[1] * metric[3] - sqr(metric[2])));
      externalAccel[1] := toRight * (vel[1] * metric[2] + vel[2] *
        metric[3]) / det;
      externalAccel[2] := -toRight * (vel[1] * metric[1] + vel[2] *
        metric[2]) / det;
      accel[1] := accel[1] + externalAccel[1];
      accel[2] := accel[2] + externalAccel[2];
    end;
    accel[3] := 0;
  end;
end;


function FindTarget(target, startingFrom: Vector3;
  var param: vector2): double;
const
  minDenom: double = 1.0E-10;
var
  moving, shift: vector2;
  p: Vector3;
  twoF: Single;
  f1, f2, f3, f4, f5, f6, a, b, c, r, q, det, step, tol: double;
  badPoint, good: Boolean;
  i, iterations: Integer;

  function distToTarget2(y: Vector3): double;
  begin
    distToTarget2 := sqr(y[1] - target[1]) + sqr(y[2] - target[2]) +
      sqr(y[3] - target[3]);
  end;
begin
  FindTarget := -1;
  moving := param;
  p := startingFrom;
  twoF := 2 * (sqr(p[1] - target[1]) + sqr(p[2] - target[2]) + sqr(p[3] -
    target[3]));
  if twoF < gVerySmall then begin
    FindTarget := min(gHalfEdges2D[1] - abs(param[1] -
      gCenter2D[1]), gHalfEdges2D[2] - abs(param[2] -
      gCenter2D[2]));
    exit;
  end;
  iterations := 0;
  repeat
    theFPar(p, moving[1] + gIncr2D[1], moving[2]);
    f1 := distToTarget2(p);
    theFPar(p, moving[1] - gIncr2D[1], moving[2]);
    f3 := distToTarget2(p);
    theFPar(p, moving[1], moving[2] + gIncr2D[2]);
    f2 := distToTarget2(p);
    theFPar(p, moving[1], moving[2] - gIncr2D[2]);
    f4 := distToTarget2(p);
    theFPar(p, moving[1] + gIncr2D[1], moving[2] + gIncr2D[2]);
    f5 := distToTarget2(p);
    theFPar(p, moving[1] - gIncr2D[1], moving[2] - gIncr2D[2]);
    f6 := distToTarget2(p);
    a := (f1 + f3 - twoF) / g2SqrIncr2D[1];
    c := (f2 + f4 - twoF) / g2SqrIncr2D[2];
    b := (f5 + f6 - f1 - f2 - f3 - f4 + twoF) / g4Incr1Incr2;
    det := a * c - sqr(b);
    if abs(det) < 1.0E-9 then
      Exit;
    r := (f1 - f3) / g4Incr2D[1];
    q := (f2 - f4) / g4Incr2D[2];
    shift[1] := (r * c - q * b) / det;
    shift[2] := (a * q - b * r) / det;
    for i := 1 to 2 do
      moving[i] := moving[i] - shift[i];
    badPoint := min(gHalfEdges2D[1] - abs(moving[1] -
      gCenter2D[1]), gHalfEdges2D[2] - abs(moving[2] -
      gCenter2D[2])) < 0;
    if not badPoint then begin
      step := abs(shift[1]) + abs(shift[2]);
      theFPar(p, moving[1], moving[2]);
      twoF := 2 * distToTarget2(p);
      iterations := iterations + 1;
      badPoint := (iterations > 10) or (twoF > gTolDistance) or (step <
        gRadius2D);
    end;
  until (twoF < gVerySmall) or badPoint;
  if (twoF < gVerySmall) then begin
    if InRangeAndClips(p, moving) then
      FindTarget := min(gHalfEdges2D[1] - abs(moving[1] -
        gCenter2D[1]), gHalfEdges2D[2] - abs(moving[2] -
        gCenter2D[2]));
    param := moving;
  end;
end;

function PickedPtParametric(mousePt: TPoint;
  var paramPt: vector2;
  var pos: Vector3): Boolean;
var
  c, ch, cv: vector2;
  hh, hv, vh, vv: double;
  det: Single;
  v3: Vector3;
begin
  Result := (gParamBuffer^[mousePt.x, mousePt.y, 1] >= 0) and
    (gParamBuffer^[mousePt.x, mousePt.y, 1] >= 0);
  if Result then begin
    paramPt[1] := gMin2D[1] + gStep2D[1] * gParamBuffer^[mousePt.x,
      mousePt.y, 1];
    paramPt[2] := gMin2D[2] + gStep2D[2] * gParamBuffer^[mousePt.x,
      mousePt.y, 2];
    theFPar(pos, paramPt[1], paramPt[2]);
    c := ProjectToPaneReal(pos);
    theFPar(v3, paramPt[1] + gRadius2D, paramPt[2]);
    ch := ProjectToPaneReal(v3);
    theFPar(v3, paramPt[1], paramPt[2] + gRadius2D);
    cv := ProjectToPaneReal(v3);
    hh := (ch[1] - c[1]) / gRadius2D;
    hv := (cv[1] - c[1]) / gRadius2D;
    vh := (ch[2] - c[2]) / gRadius2D;
    vv := (cv[2] - c[2]) / gRadius2D;
    det := hh * vv - hv * vh;
    if abs(det) < 1.0E-8 then Exit;
    paramPt[1] := paramPt[1] + ((mousePt.x - c[1]) * vv - (mousePt.y -
      c[2]) * hv) / det;
    paramPt[2] := paramPt[2] + ((mousePt.y - c[2]) * hh - (mousePt.x -
      c[1]) * vh) / det;
    if paramPt[1] < gMin2D[1] then
      paramPt[1] := gMin2D[1] + 1.0E-10;
    if paramPt[2] < gMin2D[2] then
      paramPt[2] := gMin2D[2] + 1.0E-10;
    if paramPt[1] > gMax2D[1] then
      paramPt[1] := gMax2D[1] - 1.0E-10;
    if paramPt[2] > gMax2D[2] then
      paramPt[2] := gMax2D[2] - 1.0E-10;
    Result := InRangeAndClips(pos, paramPt);
  end;
end;

function PickedPtParametricS(mousePt: TPoint;
  var paramPt: vector2;
  var pos: Vector3Single): Boolean;
var
  v: Vector3;
  i: Integer;
begin
  for i := 1 to 3 do
    v[i] := pos[i];
  result := PickedPtParametric(mousePt, paramPt, v);
  for i := 1 to 3 do
    pos[i] := v[i];
end;

function NextPoint(lastX, nextX: Vector3; var
  newX, newVel: Vector3): Boolean;
{Only the first two coordinates of 'lastX, nextX, newX, newVel are of use. These
vectors represent points ('newVel' a velocity) in the parameter space.
'lastX' is the last point in the domain and with image inside the clips.
'nextX' is another point whose image perhaps belongs also to the surface,
NextPoint strives to find a value 'newX' whose image by theFPar(newX[1],newX[2])
is the same as that of 'nextX'. Also NextPoint will give, in 'newVel' the}
{velocity in 2D that,after mapping it to 3D, should correspond to the step from
the image of 'lastX' to the inmage of 'nextX'.}
var
  i, k: integer;
  param, paramVel: vector2;
  E, F, G, ah, av, det, aux: double;
  x, xu, xv, vel3, target: Vector3;
  step2D: vector2;

  function Find: Boolean;
  begin
    theFPar(x, param[1], param[2]);
    result := (FindTarget(target, x, param) > 0) and
      (hypot(param[1] - lastX[1], param[2] - lastX[2]) > 2 * gRadius2D);
  end;
begin
  theFPar(target, nextX[1], nextX[2]);
  step2D[1] := gEdges2D[1] / kppLine2D;
  step2D[2] := gEdges2D[2] / kppLine2D;
  param[2] := gMin2D[2] + 0.5 * step2D[2];
  for i := 0 to kppLine2D - 1 do begin
    param[1] := gMin2D[1] + (i + 0.5) * step2D[1];
    result := Find;
    if result then Break;
  end;
  if not result then begin
    param[2] := gMax2D[2] - 0.5 * step2D[2];
    for i := 0 to kppLine2D - 1 do begin
      param[1] := gMin2D[1] + (i + 0.5) * step2D[1];
      result := Find;
      if result then Break;
    end;
  end;
  if not result then begin
    param[1] := gMin2D[1] + 0.5 * step2D[1];
    for i := 1 to kppLine2D - 2 do begin
      param[2] := gMin2D[2] + (i + 0.5) * step2D[2];
      result := Find;
      if result then Break;
    end;
  end;
  if not result then begin
    param[1] := gMax2D[1] - 0.5 * step2D[1];
    for i := 1 to kppLine2D - 2 do begin
      param[2] := gMin2D[2] + (i + 0.5) * step2D[2];
      result := Find;
      if result then Break;
    end;
  end;
  if result then begin
    newX[1] := param[1];
    newX[2] := param[2];
    ComputeEFGAndPartials(param, E, F, G, xu, xv);
    det := E * G - sqr(F);
    if abs(det) > 1.0E-8 then begin
      theFPar(x, lastX[1], lastX[2]);
      for k := 1 to 3 do
        vel3[k] := target[k] - x[k];
      ah := InnerProd3(vel3, xu);
      av := InnerProd3(vel3, xv);
      paramVel[1] := (ah * G - av * F) / det;
      paramVel[2] := (av * E - ah * F) / det;
      param[1] := lastX[1]; param[2] := lastX[2];
      aux := abs(E * sqr(paramVel[1]) + 2 * F * paramVel[1] * paramVel[2] + G
        * sqr(paramVel[2]));
      ComputeEFG(param, E, F, G);
      aux := sqrt(abs((E * sqr(paramVel[1]) + 2 * F * paramVel[1] * paramVel[2] + G
        * sqr(paramVel[2]))) / aux);
      newVel[1] := paramVel[1] * aux; newVel[2] := paramVel[2] * aux;
    end;
    Exit;
  end;
end;

function FindVelFor2OrderCurvesParamSurf(var paramVel: Vector3): Boolean;
var
  i: Integer;
  E, F, G, ah, av, det, dg, t: double;
  xu, xv, director, vel, normal: Vector3;
begin
  result := False;
  for i := 1 to 3 do
    paramVel[i] := 0;
  if (gFirstPtDrawing3D.x = gCurPtDrawing3D.x) and (gFirstPtDrawing3D.y =
    gCurPtDrawing3D.y) then
    Exit;
  if not gFirstPtGood then
    Exit;
  director := FindRay(gCurPtDrawing3D);
  ComputeEFGAndPartials(gFirst2DPos, E, F, G, xu, xv);
  det := E * G - sqr(F);
  if abs(det) < 1.0E-8 then
    Exit;
  normal := EuclCrossProduct(xu, xv);
  dg := EuclInnerProd3(director, normal);
  if abs(dg) > 1.0E-7 then begin
    t := (gFirst3DPos[1] - gViewPoint[1]) * normal[1] + (gFirst3DPos[2] -
      gViewPoint[2]) *
      normal[2] + (gFirst3DPos[3] - gViewPoint[3]) * normal[3];
    t := t / dg;
    for i := 1 to 3 do
      vel[i] := gViewPoint[i] - gFirst3DPos[i] + t * director[i];
    ah := InnerProd3(vel, xu);
    av := InnerProd3(vel, xv);
    paramVel[1] := (ah * G - av * F) / det;
    paramVel[2] := (av * E - ah * F) / det;
  end;
  Result := True;
end;

function FindParamsPosAndVelParamSurface(normalized: Boolean; lastPt, curPt:
  TPoint;
  var paramPt, velocity: vector2;
  var pos: Vector3): Boolean;
var
  i: Integer;
  E, F, G, ah, av, an, det, dg, t: double;
  xu, xv, director, vel, normal: Vector3;
begin
  Result := False;
  for i := 1 to 2 do
    velocity[i] := 0;
  if (lastPt.x = curPt.x) and (lastPt.y = curPt.y) then
    Exit;
  if not PickedPtParametric(lastPt, paramPt, pos) then
    Exit;
  director := FindRay(curPt);
  ComputeEFGAndPartials(paramPt, E, F, G, xu, xv);
  det := E * G - sqr(F);
  if abs(det) < 1.0E-8 then
    Exit;
  normal := EuclCrossProduct(xu, xv);
  dg := EuclInnerProd3(director, normal);
  if abs(dg) > 1.0E-7 then begin
    t := (pos[1] - gViewPoint[1]) * normal[1] + (pos[2] - gViewPoint[2]) *
      normal[2] + (pos[3] - gViewPoint[3]) * normal[3];
    t := t / dg;
    for i := 1 to 3 do
      vel[i] := gViewPoint[i] - pos[i] + t * director[i];
    if normalized then begin
      an := sqrt(abs(InnerProd3(vel, vel)));
      if an < 1.0E-8 then
        Exit;
      for i := 1 to 3 do
        vel[i] := vel[i] / an;
    end;
    ah := InnerProd3(vel, xu);
    av := InnerProd3(vel, xv);
    velocity[1] := (ah * G - av * F) / det;
    velocity[2] := (av * E - ah * F) / det;
  end;
  Result := True;
end;


procedure DrawCoordLinesParamSurface(param:
  Vector2);
var
  i: Integer;
  vec1, vec2: Vector3;
  u, v, jumpU, jumpV: double;
  pt: Vector2Single;
  par1, par2: vector2;
  can: TCanvas;
  g: gamma16Ptr;
begin
  g := gLinesData[kCoordinates].gammaP;
  jumpU := gEdges2D[1] / gGridU;
  jumpV := gEdges2D[2] / gGridV;
  u := param[1];
  v := gMin2D[2];
  theFPar(vec2, u, v);
  for i := 1 to gGridV do begin
    vec1 := vec2;
    v := v + jumpV;
    theFPar(vec2, u, v);
    DrawSegmentParamSurface3D(vec1, vec2, g, True);
  end;
  u := gMin2D[1];
  v := param[2];
  theFPar(vec2, u, v);
  for i := 1 to gGridU do begin
    vec1 := vec2;
    u := u + jumpU;
    theFPar(vec2, u, v);
    DrawSegmentParamSurface3D(vec1, vec2, g, True);
  end;
  FromParametersToDrawingSingle(param, pt);
  can := Surface.Drawing2D.canvas;
  can.Pen.Color := gLinesData[kCoordinates].pen2DColor;
  can.Pen.Width := 1;
  can.Pen.Mode := pmCopy;

  par1[1] := 0;
  par1[2] := pt[2];
  can.moveto(0, round(pt[2]));
  par2[1] := kWidth;
  par2[2] := pt[2];
  can.lineto(kWidth, round(pt[2]));
  AddSegment(par1, par2, 2, 0);

  par1[1] := pt[1];
  par1[2] := 0;
  can.moveto(round(pt[1]), 0);
  par2[1] := pt[1];
  par2[2] := kHeight;
  can.lineto(round(pt[1]), kHeight);
  AddSegment(par1, par2, 2, 0);
end;

procedure DrawMeshParamSurface;
var
  i, j, activeW: Integer;
  vec1, vec2: Vector3;
  u, v, jumpU, jumpV: double;
  g: gamma16Ptr;
begin
  activeW := gActiveLine;
  gActiveLine := kCoordinates;
  g := gLinesData[kCoordinates].gammaP;
  surface.Drawing3D.canvas.Pen.Mode := pmCopy;
  jumpU := gEdges2D[1] / gMeshNumber;
  jumpV := gEdges2D[2] / gGridV;
  u := gMin2D[1];
  v := gMin2D[2];
  theFPar(vec2, u, v);
  for i := 0 to gMeshNumber do begin
    u := gMin2D[1] + i * jumpU;
    v := gMin2D[2];
    theFPar(vec2, u, v);
    for j := 1 to gGridV do begin
      vec1 := vec2;
      v := v + jumpV;
      theFPar(vec2, u, v);
      DrawSegmentParamSurface3D(vec1, vec2, g, True);
    end
  end;
  jumpU := (gMax2D[1] - gMin2D[1]) / gGridU;
  jumpV := (gMax2D[2] - gMin2D[2]) / gMeshNumber;

  v := gMin2D[2];
  u := gMin2D[1];
  theFPar(vec2, u, v);
  for i := 0 to gMeshNumber do begin
    v := gMin2D[2] + i * jumpV;
    u := gMin2D[1];
    theFPar(vec2, u, v);
    for j := 1 to gGridU do begin
      vec1 := vec2;
      u := u + jumpU;
      theFPar(vec2, u, v);
      DrawSegmentParamSurface3D(vec1, vec2, g, True);
    end
  end;
  gActiveLine := activeW;
end;

procedure DrawFirstOrderLines(fromHere: Vector2);
var
  axis, j: Integer;
  initvel: array[0..1] of Vector3;
  vel, here: Vector3;
  key: Integer;
begin
  Randomize;
  surface.Drawing2D.canvas.Pen.Mode := pmCopy;
  surface.Drawing3D.canvas.Pen.Mode := pmCopy;
  if (gGeomOption = LightRays) and (gEpsilon = 1) then
    exit;
  repeat
    for j := 1 to 2 do
      initVel[0, j] := Random;
  until (initVel[0, 1] <> 0) or (initVel[0, 2] <> 0);
  initVel[0, 3] := 0;
  here[1] := fromHere[1]; here[2] := fromHere[2];
  if not DirectionsParamSurf(vel, initVel[0], here, 0) or not
    DirectionsParamSurf(initVel[1], initVel[0], here, 1) then
    Exit;
  initVel[0] := vel;
  theRunKut.SetEnvironment(2, 1, kTol, DirectionsParamSurf,
    CheckRangeParametric,
    RecordAndDrawSegmentParamSurface23D, CheckCameBackParametricSurface,
    nil);
  for axis := 0 to 1 do begin
    vel := initVel[axis];
    here[1] := fromHere[1]; here[2] := fromHere[2];
    key := theRunKut.ComputeIntegralCurve(here, vel, axis);
    if (key and kDoubled) = 0 then begin
      here[1] := fromHere[1]; here[2] := fromHere[2];
      vel[1] := -initVel[axis, 1];
      vel[2] := -initVel[axis, 2];
      theRunKut.ComputeIntegralCurve(here, vel, axis);
    end;
  end;
end;


procedure DrawCurvesParamSurface(fromHere: vector2);
begin
  if (not gAlreadyDrawn) or (gMode <> Standard) then
    Exit;
  surface.cursor := crHourGlass;
  if gGeomOption = CoordLines then begin
    EraseCross2D;
    EraseCross3D;
    DrawCoordLinesParamSurface(fromHere)
  end
  else begin
    EraseX;
    DrawFirstOrderLines(fromHere);
  end;
  surface.cursor := crDefault;
end;

function ComputeClipParamSurface(p1, p2: TPoint; FromDrawing3D: Boolean; var
  aClip: Clip):
  Boolean;
var
  i: Integer;
  param, v: vector2;
  xu, xv, vel, normal, pos: Vector3;
  a: Double;
begin
  result := False;
  if FromDrawing3D then begin
    if not FindParamsPosAndVelParamSurface(True, p1, p2, param, v, pos) then
      Exit;
  end
  else begin
    FromDrawingToParameters(p1, param);
    FromDrawingToParameters(p2, v);
    for i := 1 to 2 do
      v[i] := v[i] - param[i];
    theFPar2(pos, param);
  end;
  ComputePartials(param, xu, xv);
  for i := 1 to 3 do
    vel[i] := xu[i] * v[1] + xv[i] * v[2];
  normal := EuclCrossProduct(xu, xv);
  xu := EuclCrossProduct(normal, vel);
  a := EuclNorm3(xu);
  if a > 1.0E-9 then begin
    a := 1 / a;
    for i := 1 to 3 do
      aClip.u[i] := xu[i] * a;
    aClip.a := -EuclInnerProd3(xu, pos) * a;
    result := True;
  end;
end;

procedure AddClipParamSurface(param, paramVel: vector2);
var
  i: Integer;
  xu, xv, v, normal: Vector3;
  a: Double;
begin
  ComputePartials(param, xu, xv);
  for i := 1 to 3 do
    v[i] := xu[i] * paramVel[1] + xv[i] * paramVel[2];
  normal := EuclCrossProduct(xu, xv);
  xu := EuclCrossProduct(normal, v);
  a := EuclNorm3(xu);
  if a > 1.0E-9 then begin
    if gClips <> nil then
      i := Length(gClips)
    else
      i := 0;
    SetLength(gClips, i + 1);
    gClipTop := i;
    theFPar2(xv, param);
    if normal[1] * (gViewPoint[1] - gCenter3D[1]) + normal[2] * (gViewPoint[2]
      -
      gCenter3D[2]) + normal[3] * (gViewPoint[3] - gCenter3D[3]) < 0 then
      a := -a;
    for i := 1 to 3 do
      gClips[gClipTop].u[i] := xu[i] / a;
    gClips[gClipTop].a := -EuclInnerProd3(xu, xv) / a;
  end;
end;

procedure DrawSegmentParamSurface3D(vec1, vec2:
  Vector3; g: gamma16Ptr; toSave: Boolean);
var
  i, colorSegm, anver: Integer;
  brightness: double;
  d1, d2: Double;
  col: byte;
  pixR1, pixR2: Vector2;
  pixel1, pixel2, cent: Tpoint;
  vis1, vis2: boolean;
  mid, mid1, mid2: Vector3;
  can: TCanvas;
begin
  can := surface.drawing3D.canvas;
  pixR1 := ProjectToPaneRealAndDist(vec1, d1);
  pixel1.x := round(pixR1[1]);
  pixel1.Y := round(pixR1[2]);
  pixR2 := ProjectToPaneRealAndDist(vec2, d2);
  pixel2.x := round(pixR2[1]);
  pixel2.Y := round(pixR2[2]);
  vis1 := True;
  vis2 := True;
  if gClips <> nil then
    for i := 0 to gClipTop do begin
      vis1 := vis1 and (HeightOverClip(gClips[i], vec1) < 0);
      vis2 := vis2 and (HeightOverClip(gClips[i], vec2) < 0);
    end;
  vis1 := vis1 and PtInRect(gRect, pixel1) and (d1 < 1.001 *
    gZBuffer^[pixel1.x,
    pixel1.y]);
  vis2 := vis2 and PtInRect(gRect, pixel2) and (d2 < 1.001 *
    gZBuffer^[pixel2.x,
    pixel2.y]);
  if (vis1 or vis2) then begin
    if vis1 and vis2 then begin
      cent.x := (pixel1.x + pixel2.x) div 2;
      cent.y := (pixel1.y + pixel2.y) div 2;
      if gMap <> nil then begin
        col := gMap^[cent.x, cent.y];
        if col > 100 then begin
          brightness := (col - 100) / 100;
          anver := 1;
        end
        else if col > 0 then begin
          brightness := col / 100;
          anver := 0;
        end
        else begin
          brightness := 0;
          anver := -1;
        end;
        colorSegm := round(1 + 15 * brightness);
        can.Pen.Color := g^[colorSegm];
        if ((gSideToDrawOn = 2) or (anver = gSideToDrawOn)) then begin
          can.Moveto(pixel1.x, pixel1.y);
          can.Lineto(pixel2.x, pixel2.y);
          AddSegment(pixR1, pixR2, 3, colorSegm);
        end;
      end
    end
    else if (abs(pixel2.x - pixel1.x) + abs(pixel2.y - pixel1.y) > 3) then
    begin
      for i := 1 to 3 do
        mid[i] := 0.5 * (vec1[i] + vec2[i]);
      mid1 := vec1;
      mid2 := mid;
      DrawSegmentParamSurface3D(mid1, mid2, g, toSave);
      mid1 := mid;
      mid2 := vec2;
      DrawSegmentParamSurface3D(mid1, mid2, g, toSave);
    end;
  end;
end;

procedure RecordAndDrawSegmentParamSurface23D(x1, x2: Vector3; toSave:
  Boolean);
var
  p1Window, p2Window: Vector2;
  can: TCanvas;
  pos1, pos2: Vector3;
  g: gamma16Ptr;
begin
  can := surface.Drawing2D.canvas;
  can.pen.mode := pmCopy;
  can.Pen.Color := gLinesData[gActiveLine].pen2DColor;
  FromParametersToDrawingReal(x1, p1Window);
  FromParametersToDrawingReal(x2, p2Window);
  AddSegment(p1Window, p2Window, 2, 0);
  can.moveto(round(p1Window[1]), round(p1Window[2]));
  can.lineto(round(p2Window[1]), round(p2Window[2]));
  theFPar(pos1, x1[1], x1[2]);
  theFPar(pos2, x2[1], x2[2]);
  g := gLinesData[gActiveLine].gammaP;
  DrawSegmentParamSurface3D(pos1, pos2, g, toSave);
end;

function ComputeChristSymbParam(p: vector3; var metric, gammaU, gammaV:
  Vector3):
  BOOLEAN;
{Computes Christoffel symbols at p. If abs(det(EG-F*F)) < 1.0e-8}
{at p, returns False. Else, True.}
{gammaU[1] corresponds to uuu, gammaU[2] to uuv, gammaU[3] to uvv.}
{On 'metric' are given E,F,G at p}

var
  E, F, G, E1, E2, F1, F2, G1, G2, Ex, Ey, Fx, Fy, Gx, Gy, det, inc: double;
  pos: vector2;
begin
  pos[1] := p[1]; pos[2] := p[2];
  Result := FALSE;
  ComputeEFG(pos, E, F, G);
  metric[1] := E;
  metric[2] := F;
  metric[3] := G;

  pos[1] := p[1] - gIncr2D[1];
  pos[2] := p[2];
  ComputeEFG(pos, E1, F1, G1);
  pos[1] := p[1] + gIncr2D[1];
  ComputeEFG(pos, E2, F2, G2);
  inc := 2 * gIncr2D[1];
  Ex := (E2 - E1) / inc;
  Fx := (F2 - F1) / inc;
  Gx := (G2 - G1) / inc;

  pos[1] := p[1];
  pos[2] := p[2] - gIncr2D[2];
  ComputeEFG(pos, E1, F1, G1);
  pos[2] := p[2] + gIncr2D[2];
  ComputeEFG(pos, E2, F2, G2);
  inc := 2 * gIncr2D[2];
  Ey := (E2 - E1) / inc;
  Fy := (F2 - F1) / inc;
  Gy := (G2 - G1) / inc;

  det := 2 * (E * G - sqr(F));
  if abs(det) > 1.0E-8 then begin
    Result := True;
    gammaU[1] := (G * Ex - 2 * F * Fx + F * Ey) / det;
    gammaU[2] := (G * Ey - F * Gx) / det;
    gammaU[3] := (2 * G * Fy - G * Gx - F * Gy) / det;
    gammaV[1] := (2 * E * Fx - E * Ey - F * Ex) / det;
    gammaV[2] := (E * Gx - F * Ey) / det;
    gammaV[3] := (E * Gy - 2 * F * Fy + F * Gx) / det;
  end;
end;

function ComputeFrameParamSurface(var u1, u2: Vector3; var phi, topT: Double;
  x: Vector3): double;
var
  E, F, G, a: Double;
  e1, e2, p: Vector2;
begin
  p[1] := x[1]; p[2] := x[2]; u1[3] := 0; u2[3] := 0;
  ComputeEFG(p, E, F, G);
  result := E * G - sqr(F);
  topT := sqrt(abs(E * sqr(u1[1]) + 2 * F * u1[1] * u1[2] + G * sqr(u1[2])));

  if result > 1.0E-12 then begin
    u2[1] := -(F * u1[1] + G * u1[2]);
    u2[2] := E * u1[1] + F * u1[2];
    a := sqrt(abs(E * sqr(u2[1]) + 2 * F * u2[1] * u2[2] + G
      * sqr(u2[2])));
    if a < 1.0E-7 then
      result := 0
    else begin
      a := topT / a;
      u2[1] := a * u2[1];
      u2[2] := a * u2[2];
      u2[3] := 0;
    end;
  end
  else if result < -1.0E-7 then begin
    a := sqrt(-result);
    if abs(E) > 1.0E-7 then begin
      e1[1] := -F + a; e1[2] := E;
      e2[1] := -F - a; e2[2] := E
    end
    else begin
      e1[1] := 1; e1[2] := 0;
      e2[1] := -G; e2[2] := 2 * F;
    end;
    a := hypot(e1[1], e1[2]);
    e1[1] := e1[1] / a; e1[2] := e1[2] / a;
    a := hypot(e2[1], e2[2]);
    e2[1] := e2[1] / a; e2[2] := e2[2] / a;
    a := e1[1] * e2[1] + e1[2] * e2[2];
    if a < 0 then begin
      e2[1] := -e2[1]; e2[2] := -e2[2];
    end;
    a := abs(a);
    phi := arccos(a) / 2;
    u1[1] := (e1[1] + e2[1]) / sqrt(2 * (1 + a));
    u1[2] := (e1[2] + e2[2]) / sqrt(2 * (1 + a));
    u2[1] := (e1[1] - e2[1]) / sqrt(2 * (1 - a));
    u2[2] := (e1[2] - e2[2]) / sqrt(2 * (1 - a));
  end
  else
    result := 0;
end;

procedure ComputeWireParametricSurface;
var
  i, j: Integer;
  vec1, vec2: Vector3;
  param: Vector2;
  jumpU, jumpV: double;
  ended: Boolean;
begin
  InitWire;
  jumpU := gEdges2D[1] / kWiresPerLine;
  jumpV := gEdges2D[2] / kSegmentsAlongLine;
  for i := 0 to kWiresPerLine do begin
    param[1] := gMin2D[1] + i * jumpU;
    param[2] := gMin2D[2];
    theFPar2(vec2, param);
    AddSegmentToWire(vec2, 1);
    ended := False;
    for j := 1 to kSegmentsAlongLine do begin
      vec1 := vec2;
      param[2] := param[2] + jumpV;
      if InClips(vec2, param) then
        AddSegmentToWire(vec2, ord(ended))
      else
        ended := True;
    end;
  end;
  jumpU := gEdges2D[1] / kSegmentsAlongLine;
  jumpV := gEdges2D[2] / kWiresPerLine;
  for i := 0 to kWiresPerLine do begin
    param[2] := gMin2D[2] + i * jumpV;
    param[1] := gMin2D[1];
    theFPar2(vec2, param);
    AddSegmentToWire(vec2, 1);
    ended := False;
    for j := 1 to kSegmentsAlongLine do begin
      vec1 := vec2;
      param[1] := param[1] + jumpU;
      if InClips(vec2, param) then
        AddSegmentToWire(vec2, ord(ended))
      else
        ended := True;
    end;
  end;
end;

end.
