

{-----------------------------------------------------------------------------
 Unit Name: ReadForms
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: Routines for registering and verifying the data entered by the user
    for defining a surface.
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit ReadForms;

interface

function ReadForm: Integer;

implementation

uses
  Globals, FSurface, OCFunct, ResStrings, SysUtils, Classes;

type
  EReadFormError = class(Exception)
    ErrorCode: Integer;
  end;

var
  newF1, newF2, newF3, newF4: TOpCodeFunction;
  uMin, uMax, vMin, vMax, wMIn, wMax: Double;
  grid: Integer;
  readStrings: array[1..kNumOfSavedToDiskStrings] of AnsiString;

procedure createFunctions;
begin
  newF1 := TOpCodeFunction.create;
  newF2 := TOpCodeFunction.create;
  newF3 := TOpCodeFunction.create;
  newF4 := TOpCodeFunction.create;
end;

procedure freeFunctions;
begin
  newF1.free;
  newF2.free;
  newF3.free;
  newF4.free;
end;

procedure AssignFunctions;
begin
  gTheF1.Free;
  gTheF1 := newF1;
  if gType <> Implicit then begin
    gTheF2.Free;
    gTheF2 := newF2;
    gTheF3.Free;
    gTheF3 := newF3;
  end;
  if gType = Tube then begin
    gTheF4.Free;
    gTheF4 := newF4;
  end;
end;



function ReadFormParametric: Integer;
var
  warning, vars: AnsiString;
  separator: char;

  procedure HandleFormReadingException(mess: AnsiString);
  begin
    ReadFormParametric := 2;
    surface.Inform(mess);
    beep;
  end;

begin
  if gType = Parametric then
    vars := 'uv'
  else
    vars := 'xy';
  result := 0;
  try
    readStrings[1] := Trim(surface.f1Edit.Text);
    warning := newF1.ParseNewText(readStrings[1], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[1] <> gSurfaceStr[1] then
      Result := 1;

    readStrings[2] := Trim(surface.f2Edit.Text);
    warning := newF2.ParseNewText(readStrings[2], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[2] <> gSurfaceStr[2] then
      Result := 1;

    readStrings[3] := Trim(surface.f3Edit.Text);
    warning := newF3.ParseNewText(readStrings[3], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[3] <> gSurfaceStr[3] then
      Result := 1;

    readStrings[5] := Trim(surface.uMinEdit.Text);
    warning := ReadConstantField(readStrings[5], uMin);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[5] <> gSurfaceStr[5] then
      Result := 1;

    readStrings[6] := Trim(surface.uMaxEdit.Text);
    warning := ReadConstantField(readStrings[6], uMax);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[6] <> gSurfaceStr[6] then
      Result := 1;

    readStrings[7] := Trim(surface.vMinEdit.Text);
    warning := ReadConstantField(readStrings[7], vMin);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[7] <> gSurfaceStr[7] then
      Result := 1;

    readStrings[8] := Trim(surface.vMaxEdit.Text);
    warning := ReadConstantField(readStrings[8], vMax);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[8] <> gSurfaceStr[8] then
      Result := 1;
    if gType = Parametric then begin
      try
        separator := DecimalSeparator;
        DecimalSeparator := '.';
        grid := Trunc(Abs(StrToFloat(Trim(surface.wMaxEdit.Text))));
        DecimalSeparator := separator;
      except
        on EConvertError do begin
          grid := 60;
          DecimalSeparator := separator;
        end;
      end;
      if (grid < 10) then grid := 10;
      if (grid > 127) then grid := 127;
      surface.wMaxEdit.Text := IntToStr(grid);
      readStrings[10] := surface.wMaxEdit.Text;
      if readStrings[10] <> gSurfaceStr[10] then
        Result := 1;
    end;

    if (uMin >= uMax) then begin
      warning := eUMinLessThanUMax;
      raise EReadFormError.Create(warning);
    end;
    if (vMin >= vMax) then begin
      warning := eVMinLessThanVMax;
      raise EReadFormError.Create(warning);
    end;
    gMin2D[1] := uMin;
    gMin2D[2] := vMin;
    gMax2D[1] := uMax;
    gMax2D[2] := vMax;
    if gType = Parametric then begin
      gGridU := grid;
      gGridV := grid;
    end;
  except
    on EReadFormError do
      HandleFormReadingException(warning);
  end;
end;

function ReadFormImplicit: Integer;
var
  warning, vars: AnsiString;

  procedure HandleFormReadingException(mess: AnsiString);
  begin
    ReadFormImplicit := 2;
    surface.Inform(mess);
    beep;
  end;

begin
  vars := 'xyz';
  result := 0;
  try
    readStrings[1] := Trim(surface.f1Edit.Text);
    warning := newF1.ParseNewText(readStrings[1], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[1] <> gSurfaceStr[1] then
      Result := 1;

    readStrings[5] := Trim(surface.uMinEdit.Text);
    warning := ReadConstantField(readStrings[5], uMin);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[5] <> gSurfaceStr[5] then
      Result := 1;

    readStrings[6] := Trim(surface.uMaxEdit.Text);
    warning := ReadConstantField(readStrings[6], uMax);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[6] <> gSurfaceStr[6] then
      Result := 1;

    readStrings[7] := Trim(surface.vMinEdit.Text);
    warning := ReadConstantField(readStrings[7], vMin);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[7] <> gSurfaceStr[7] then
      Result := 1;

    readStrings[8] := Trim(surface.vMaxEdit.Text);
    warning := ReadConstantField(readStrings[8], vMax);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[8] <> gSurfaceStr[8] then
      Result := 1;

    readStrings[9] := Trim(surface.wMinEdit.Text);
    warning := ReadConstantField(readStrings[9], wMin);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[9] <> gSurfaceStr[9] then
      Result := 1;

    readStrings[10] := Trim(surface.wMaxEdit.Text);
    warning := ReadConstantField(readStrings[10], wMax);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[10] <> gSurfaceStr[10] then
      Result := 1;

    if (uMin >= uMax) then begin
      warning := eXMinLessThanXMax;
      raise EReadFormError.Create(warning);
    end;
    if (vMin >= vMax) then begin
      warning := eYMinLessThanYMax;
      raise EReadFormError.Create(warning);
    end;
    if (wMin >= wMax) then begin
      warning := eZMinLessThanZMax;
      raise EReadFormError.Create(warning);
    end;
    gMin3D[1] := uMin;
    gMin3D[2] := vMin;
    gMin3D[3] := wMin;
    gMax3D[1] := uMax;
    gMax3D[2] := vMax;
    gMax3D[3] := wMax;
  except
    on EReadFormError do
      HandleFormReadingException(warning);
  end;
end;

function ReadFormTube: Integer;
var
  warning, vars: AnsiString;

  procedure HandleFormReadingException(mess: AnsiString);
  begin
    ReadFormTube := 2;
    surface.Inform(mess);
    beep;
  end;

begin
  vars := 't';
  result := 0;
  try
    readStrings[1] := Trim(surface.f1Edit.Text);
    warning := newF1.ParseNewText(readStrings[1], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[1] <> gSurfaceStr[1] then
      Result := 1;

    readStrings[2] := Trim(surface.f2Edit.Text);
    warning := newF2.ParseNewText(readStrings[2], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[2] <> gSurfaceStr[2] then
      Result := 1;

    readStrings[3] := Trim(surface.f3Edit.Text);
    warning := newF3.ParseNewText(readStrings[3], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[3] <> gSurfaceStr[3] then
      Result := 1;

    readStrings[4] := Trim(surface.f4Edit.Text);
    warning := newF4.ParseNewText(readStrings[4], vars);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[4] <> gSurfaceStr[4] then
      Result := 1;

    readStrings[5] := Trim(surface.uMinEdit.Text);
    warning := ReadConstantField(readStrings[5], uMin);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[5] <> gSurfaceStr[5] then
      Result := 1;

    readStrings[6] := Trim(surface.uMaxEdit.Text);
    warning := ReadConstantField(readStrings[6], uMax);
    if warning <> '' then
      raise EReadFormError.Create(warning);
    if readStrings[6] <> gSurfaceStr[6] then
      Result := 1;

    if (uMin >= uMax) then begin
      warning := eUMinLessThanUMax;
      raise EReadFormError.Create(warning);
    end;

    gMin2D[1] := uMin;
    gMin2D[2] := 0;
    gMax2D[1] := uMax;
    gMax2D[2] := 2 * pi;
    gGridU := kTubeGridU;
    gGridV := kTubeGridV;
  except
    on EReadFormError do
      HandleFormReadingException(warning);
  end;
end;

function ReadForm: Integer;
var
  i: Integer;
begin
  createFunctions;
  surface.Inform('');
  for i := 1 to kNumOfSavedToDiskStrings do
    readStrings[i] := '';
  case gType of
    Parametric, PseudoRiemannian:
      result := ReadFormParametric;
    Implicit:
      result := ReadFormImplicit;
    Tube:
      result := ReadFormTube;
  else
    ;
  end;
  if result < 2 then begin
    AssignFunctions;
    for i := 1 to kNumOfSavedToDiskStrings do begin
      gSurfaceStr[i] := readStrings[i];
      gSurfaceStrLengths[i] := Length(readStrings[i]);
    end;
  end
  else
    freeFunctions;
end;

end.
