

{-----------------------------------------------------------------------------
 Unit Name: Utilities
 Project Name : Superficies
 Author:    Angel Montesinos Amilibia
 05/02/02
 Purpose: General utils used everywhere in the project
-----------------------------------------------------------------------------}
{

Superficies, workshop for the geometry of surfaces
Copyright (C) 2002 Angel Montesinos Amilibia

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

The source code, documentation, and examples may be freely downloaded from
     < ftp://topologia.geomet.uv.es/pub/montesin/Superficies_W32 >

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Dr. Moliner, 50
     46100 Burjasot (Valencia)
     Spain

}


unit Utilities;

interface
uses
  Classes,
  {$IFDEF MSWINDOWS}
   Windows, Forms, Graphics,
{$ENDIF}
{$IFDEF LINUX}
   Types, QGraphics,  QForms, QTypes,
{$ENDIF}
   Globals;

function InnerProd3(x, y: Vector3): Double;
function InnerProd2(x, y: Vector2): Double;
function EuclInnerProd3(x, y: Vector3): Double;
function CrossProduct(x, y: Vector3): Vector3;
function EuclCrossProduct(x, y: Vector3): Vector3;
function EuclDistance3(pos1, pos2: Vector3): Double;
function EuclNorm3(x: Vector3): Double;
function EuclNorm3S(x: Vector3Single): Double;

procedure FromParametersToDrawing(param: Vector2; var p: TPoint);
procedure FromParametersToDrawingAux(param: Vector3; var p: TPoint);
procedure FromParametersToDrawingReal(param: Vector3; var p: Vector2);
procedure FromParametersToDrawingSingle(param: Vector2; var p: Vector2Single);
procedure FromDrawingToParameters(p: TPoint; var param: Vector2);
procedure FromDrawingXYToParameters(x, y: Integer; var param: Vector2);
function ComputeParamVelFromDrawing2DPoints(first, last: TPoint): Vector3;

function ProjectToPaneInt(pos: Vector3): TPoint;
function ProjectSToPaneInt(pos: Vector3Single): TPoint;
function ProjectToPaneReal(pos: Vector3): Vector2;
function ProjectToFocusPlaneInt(pos: Vector3): TPoint;
function ProjectToFocusPlaneReal(pos: Vector3): Vector2;
function ProjectToFocusPlaneRealAndDist(x: Vector3Single): Vector3Single;
function ProjectSToPaneRealAndDist(x: Vector3Single): Vector3Single;
function ProjectToPaneRealAndDist(x: Vector3; var d: Double): Vector2;
function ProjectToPaneIntAndDist(x: Vector3; var z: single): TPoint;
function ProjectSToPaneIntAndDist(x: Vector3Single; var z: single): TPoint;
function ProjectToPaneRealInt(x: Vector3; var projR: Vector2): TPoint;
function FindRay(pix: TPoint): Vector3;

function HeightOverClip(c: Clip; x: Vector3): single;
function HeightOverClipS(c: Clip; x: Vector3Single): single;
function InClips(x: Vector3): Boolean;
procedure AddNewClip;

function IsVisibleImplicit(fromHere, point: Vector3; theKey: CubeKey;
  pointIsInCube:
  Boolean): Boolean;
function IsVisibleImplicitS(fromHere: Vector3; point: Vector3Single; theKey:
  CubeKey;
  pointIsInCube:
  Boolean): Boolean;
procedure SegmentInCube(ori, dir: Vector3; keyOri: CubeKey; var a, b:
  double);
function FirstZero(foc, dir: Vector3; a, b: double; var root: double):
  Boolean;

procedure ReceiveColorMap(var bm: TBitmap; aMap: ColorMapPtr);

procedure InitSavedLines;
procedure GetLinesRoom(dim: Integer);
procedure AddSegment(v1, v2: Vector2; dim, color: Integer);
procedure DrawSavedLines(dim: Integer);

procedure InitWire;
procedure AddSegmentToWire(currentX: Vector3; key: Integer);

procedure DrawCube(can: TCanvas);

procedure EraseX;
procedure DrawUnitConic;
procedure EraseUnitConic;

procedure DrawWire(can: TCanvas; nLeft, hLeft: Vector3);
procedure PreparegForPrintingBmp;
procedure SaveParams;
procedure RetrieveParams;
procedure MoveSurface(prevPt,
  curPt: TPoint; Shift: TShiftState);
procedure PrintAsDesired;
function ExtracTitle(pathAndName: AnsiString): AnsiString;

implementation
uses
  Math, SysUtils,
  {$IFDEF MSWINDOWS}
  Printers,
{$ENDIF}
{$IFDEF LINUX}
  QPrinters,
{$ENDIF}
  OCFunct, FSurface, ResStrings, ParametricUtils, DrawParSurfNoShadows;

const
  EySin: double = 1 / 15;
  EyCos: double = 0.9977753031;

var
  d, dSqr, virtualSphRad2, c0abs, c0sqr, coefRadius: double;
  svp, sH, sV, sN: Vector3;
  sc0, sc1, sc2, sc3: Double;
  spcx, spcy: Integer;
  r, rOr: TRect;

function InnerProd2(x, y: Vector2): Double;
begin
  result := x[1] * y[1] + x[2] * y[2];
end;

function EuclInnerProd3(x, y: Vector3): Double;
begin
  result := x[1] * y[1] + x[2] * y[2] + x[3] * y[3];
end;

function InnerProd3(x, y: Vector3): Double;
begin
  result := x[1] * y[1] + x[2] * y[2] + gEpsilon * x[3] * y[3];
end;

function CrossProduct(x, y: Vector3): Vector3;
begin
  result[1] := x[2] * y[3] - x[3] * y[2];
  result[2] := x[3] * y[1] - x[1] * y[3];
  result[3] := gEpsilon * (x[1] * y[2] - x[2] * y[1]);
end;

function EuclCrossProduct(x, y: Vector3): Vector3;
begin
  result[1] := x[2] * y[3] - x[3] * y[2];
  result[2] := x[3] * y[1] - x[1] * y[3];
  result[3] := x[1] * y[2] - x[2] * y[1];
end;

function EuclDistance3(pos1, pos2: Vector3): Double;
begin
  result := sqrt(sqr(pos2[1] - pos1[1]) + sqr(pos2[2] - pos1[2]) + sqr(pos2[3]
    - pos1[3]));
end;

function EuclNorm3(x: Vector3): Double;
begin
  result := sqrt(sqr(x[1]) + sqr(x[2]) + sqr(x[3]));
end;

function EuclNorm3S(x: Vector3Single): Double;
begin
  result := sqrt(sqr(x[1]) + sqr(x[2]) + sqr(x[3]));
end;

procedure FromParametersToDrawing(param: Vector2; var p: TPoint);
begin
  p.x := Round(pd1 * param[1] + pd2);
  p.y := Round(pd3 * param[2] + pd4);
end;

procedure FromParametersToDrawingAux(param: Vector3; var p: TPoint);
begin
  p.x := Round(pd1 * param[1] + pd2);
  p.y := Round(pd3 * param[2] + pd4);
end;

procedure FromParametersToDrawingReal(param: Vector3; var p: Vector2);
begin
  p[1] := pd1 * param[1] + pd2;
  p[2] := pd3 * param[2] + pd4;
end;

procedure FromParametersToDrawingSingle(param: Vector2; var p: Vector2Single);
begin
  p[1] := pd1 * param[1] + pd2;
  p[2] := pd3 * param[2] + pd4;
end;

procedure FromDrawingToParameters(p: TPoint; var param: Vector2);
begin
  param[1] := dp1 * p.x + dp2;
  param[2] := dp3 * p.y + dp4;
end;

procedure FromDrawingXYToParameters(x, y: Integer; var param: Vector2);
begin
  param[1] := dp1 * x + dp2;
  param[2] := dp3 * y + dp4;
end;

function ComputeParamVelFromDrawing2DPoints(first, last: TPoint): Vector3;
begin
  result[1] := dp1 * (last.x - first.x);
  result[2] := dp3 * (last.y - first.y);
  result[3] := 0;
end;

function ProjectToPaneInt(pos: Vector3): TPoint;
var
  lambda: Double;
begin
  lambda := gC0 / (EuclInnerProd3(gN, pos) - gC3);
  result.x := round(lambda * (EuclInnerProd3(gH, pos) - gC1)) + gpcx;
  result.y := round(lambda * (EuclInnerProd3(gV, pos) - gC2)) + gpcy;
end;

function ProjectSToPaneInt(pos: Vector3Single): TPoint;
var
  lambda: Double;
begin
  lambda := gC0 / (gN[1] * pos[1] + gN[2] * pos[2] + gN[3] * pos[3] - gC3);
  result.x := round(lambda * (gH[1] * pos[1] + gH[2] * pos[2] + gH[3] * pos[3]
    - gC1)) + gpcx;
  result.y := round(lambda * (gV[1] * pos[1] + gV[2] * pos[2] + gV[3] * pos[3]
    - gC2)) + gpcy;
end;

function ProjectToPaneReal(pos: Vector3): Vector2;
var
  lambda: Double;
begin
  lambda := gC0 / (EuclInnerProd3(gN, pos) - gC3);
  result[1] := lambda * (EuclInnerProd3(gH, pos) - gC1) + gpcx;
  result[2] := lambda * (EuclInnerProd3(gV, pos) - gC2) + gpcy;
end;

function ProjectToFocusPlaneInt(pos: Vector3): TPoint;
var
  lambda: Double;
begin
  lambda := gFC0 / (EuclInnerProd3(gFN, pos) - gFC3);
  result.x := round(lambda * (EuclInnerProd3(gFH, pos) - gFC1)) + kHalfWidth;
  result.y := round(lambda * (EuclInnerProd3(gFV, pos) - gFC2)) + kHalfHeight;
end;

function ProjectToFocusPlaneReal(pos: Vector3): Vector2;
var
  lambda: Double;
begin
  lambda := gFC0 / (EuclInnerProd3(gFN, pos) - gFC3);
  result[1] := lambda * (EuclInnerProd3(gFH, pos) - gFC1) + kHalfWidth;
  result[2] := lambda * (EuclInnerProd3(gFV, pos) - gFC2) + kHalfHeight;
end;

function ProjectToFocusPlaneRealAndDist(x: Vector3Single): Vector3Single;
begin
  result[3] := gFN[1] * x[1] + gFN[2] * x[2] + gFN[3] * x[3] - gFC3;
  result[1] := (gFH[1] * x[1] + gFH[2] * x[2] + gFH[3] * x[3] - gFC1) * gFC0 /
    result[3]
    + kHalfWidth;
  result[2] := (gFV[1] * x[1] + gFV[2] * x[2] + gFV[3] * x[3] - gFC2) * gFC0 /
    result[3]
    + kHalfHeight;
end;


function ProjectSToPaneRealAndDist(x: Vector3Single): Vector3Single;
begin
  result[3] := (gN[1] * x[1] + gN[2] * x[2] + gN[3] * x[3] - gC3);
  result[1] := (gH[1] * x[1] + gH[2] * x[2] + gH[3] * x[3] - gC1) * gC0 /
    result[3]
    +
    gpcx;
  result[2] := (gV[1] * x[1] + gV[2] * x[2] + gV[3] * x[3] - gC2) * gC0 /
    result[3]
    +
    gpcy;
end;

function ProjectToPaneRealAndDist(x: Vector3; var d: Double): Vector2;
begin
  d := (gN[1] * x[1] + gN[2] * x[2] + gN[3] * x[3] - gC3);
  result[1] := (gH[1] * x[1] + gH[2] * x[2] + gH[3] * x[3] - gC1) * gC0 / d +
    gpcx;
  result[2] := (gV[1] * x[1] + gV[2] * x[2] + gV[3] * x[3] - gC2) * gC0 / d +
    gpcy;
end;

function ProjectToPaneIntAndDist(x: Vector3; var z: single): TPoint;
begin
  z := (gN[1] * x[1] + gN[2] * x[2] + gN[3] * x[3] - gC3);
  result.x := round((gH[1] * x[1] + gH[2] * x[2] + gH[3] * x[3] - gC1) * gC0 /
    z) +
    gpcx;
  result.y := round((gV[1] * x[1] + gV[2] * x[2] + gV[3] * x[3] - gC2) * gC0 /
    z) +
    gpcy;
end;

function ProjectSToPaneIntAndDist(x: Vector3Single; var z: single): TPoint;
begin
  z := (gN[1] * x[1] + gN[2] * x[2] + gN[3] * x[3] - gC3);
  result.x := round((gH[1] * x[1] + gH[2] * x[2] + gH[3] * x[3] - gC1) * gC0 /
    z) +
    gpcx;
  result.y := round((gV[1] * x[1] + gV[2] * x[2] + gV[3] * x[3] - gC2) * gC0 /
    z) +
    gpcy;
end;

function ProjectToPaneRealInt(x: Vector3; var projR: Vector2): TPoint;
var
  z: Double;
begin
  z := (gN[1] * x[1] + gN[2] * x[2] + gN[3] * x[3] - gC3);
  projR[1] := (gH[1] * x[1] + gH[2] * x[2] + gH[3] * x[3] - gC1) * gC0 / z +
    gpcx;
  result.x := round(projR[1]);
  projR[2] := (gV[1] * x[1] + gV[2] * x[2] + gV[3] * x[3] - gC2) * gC0 / z +
    gpcy;
  result.y := round(projR[2]);
end;

function FindRay(pix: TPoint): Vector3;
var
  k: Integer;
  ah, av: Double;
begin
  ah := pix.x - gpcx;
  av := pix.y - gpcy;
  for k := 1 to 3 do
    result[k] := ah * gH[k] + av * gV[k] + gC0 * gN[k];
end;

function HeightOverClip(c: Clip; x: Vector3): single;
begin
  result := c.u[1] * x[1] + c.u[2] * x[2] + c.u[3] * x[3] + c.a;
end;

function HeightOverClipS(c: Clip; x: Vector3Single): single;
begin
  result := c.u[1] * x[1] + c.u[2] * x[2] + c.u[3] * x[3] + c.a;
end;

function InClips(x: Vector3): Boolean;
var
  i: Integer;
begin
  Result := True;
  if gClips <> nil then
    for i := 0 to gClipTop do
      Result := Result and (HeightOverClip(gClips[i], x) < 0);
end;

procedure AddNewClip;
var
  i: Integer;
  norm: Double;
begin
  norm := EuclNorm3(gNewClip.u);
  if norm > 0 then begin
    if gClips <> nil then
      i := Length(gClips)
    else
      i := 0;
    SetLength(gClips, i + 1);
    gClipTop := i;
    gClips[i] := gNewClip;
    surface.StopButton.Caption := eNoClips;
  end;
end;

function FirstZero(foc, dir: Vector3; a, b: double; var root: double):
  Boolean;
 {On start, it is assumed that a < b. Finds and gives in 'root' the first}
 {value of  t between 'a' and 'b' for which abs( theFImpl(foc + t*dir) ) is less}
 {than 'gVerySmall', and FirstZero is set to True. If there is none, FirstZero}
 {is set to FALSE.}
const
  maxStepNum = 30;
var
  stepNum, counter, i: Integer;
  step, delta, minAdvance, advance, c, f0, f1, prevF0, ab, a0, prevA0, dif, t:
  double;
  pos, incrStep, incrDelta, movedPos: Vector3;
  found: Boolean;

  procedure RegulaFalsi(var Result: double; a, b, fa, fb: double);
  {It is assumed that fa=theFImpl(foc+a*dir), fb=theFImpl(foc+b*dir), sign(fa)
  <> sign(fb),a < b. Gives a value of  t, a<= t <= b, such that
  |theFImpl(foc+t*v)| < gVerySmall.}
  var
    t, ft, a1, b1, fa1, fb1: double;
    i: Integer;
    posi: Vector3;
  begin
    a1 := a;
    b1 := b;
    fa1 := fa;
    fb1 := fb;
    t := a;
    while (abs(b1 - a1) > minAdvance) do begin
      t := (a1 * fb1 - b1 * fa1) / (fb1 - fa1);
      for i := 1 to 3 do
        posi[i] := foc[i] + t * dir[i];
      ft := gTheF1.V(posi);
      if abs(ft) < gVerySmall then
        break;
      if ft * fa1 > 0 then begin
        a1 := t;
        fa1 := ft;
      end
      else begin
        b1 := t;
        fb1 := ft;
      end;
    end; {del WHILE}
    Result := t;
  end; {de RegulaFalsi}

  function FindZero(t: double; var c: double): Boolean;
  var
    thePoint, theOtherPoint: Vector3;
    param, val0, val1, jump, prevJump, differ: double;
    badPoint: Boolean;
    i: Integer;
  begin
    Result := FALSE;
    param := t;
    for i := 1 to 3 do
      thePoint[i] := foc[i] + param * dir[i];
    val0 := gTheF1.V(thePoint);
    jump := abs(step);
    repeat
      prevJump := 0.9 * abs(jump);
      for i := 1 to 3 do
        theOtherPoint[i] := thePoint[i] + incrDelta[i];
      val1 := gTheF1.V(theOtherPoint);
      differ := val0 - val1;
      if abs(differ) < 1.0E-13 then
        exit;
      jump := val0 * delta / differ;
      badPoint := abs(jump) > prevJump;
      if badPoint then
        exit;
      param := param + jump;
      for i := 1 to 3 do
        thePoint[i] := foc[i] + param * dir[i];
      val0 := gTheF1.V(thePoint);
    until (abs(val0) < gVerySmall) or (prevJump < minAdvance);
    Result := TRUE;
    c := param;
  end; {of FindZero}



begin
  c := EuclNorm3(dir);
  delta := gRadius3D / (20 * c);
  minAdvance := 5 * delta;
  ab := b - a;
  c := ab * c;
  FirstZero := FALSE;
  root := 0;
  if c < gRadius3D then begin
    for i := 1 to 3 do begin
      pos[i] := foc[i] + a * dir[i];
      movedPos[i] := pos[i] + delta * dir[i];
    end;
    f0 := gTheF1.V(pos);
    f1 := gTheF1.V(movedPos);
    dif := f0 - f1;
    if abs(dif) > 1.0E-11 then begin
      advance := f0 * delta / dif;
      if (advance > 0) and (advance < ab) then begin
        FirstZero := TRUE;
        root := a + advance;
      end;
    end; {of IF abs(dif) > 1.0e-13}
  end {of IF c < gData.radio3D}
  else begin
    stepNum := 2 + round(maxStepNum * c / gSpan3D);
    step := ab / stepNum;
    for i := 1 to 3 do begin
      pos[i] := foc[i] + (a - step) * dir[i];
      incrDelta[i] := delta * dir[i];
      movedPos[i] := pos[i] + incrDelta[i];
      incrStep[i] := step * dir[i];
    end;
    counter := 0;
    found := FALSE;
    repeat
      for i := 1 to 3 do begin
        pos[i] := pos[i] + incrStep[i];
        movedPos[i] := movedPos[i] + incrStep[i];
      end;
      f0 := gTheF1.V(pos);
      f1 := gTheF1.V(movedPos);
      dif := f0 - f1;
      if abs(dif) > 1.0E-11 then begin
        if (counter > 0) and (f0 * prevF0 < 0) then begin
          FirstZero := TRUE;
          a0 := a + counter * step;
          prevA0 := a0 - step;
          RegulaFalsi(root, prevA0, a0, prevF0, f0);
          found := TRUE;
        end
        else begin
          advance := f0 * delta / dif;
          if (advance > 0) and (advance < 2 * step) then begin
            t := a + counter * step + 0.4 * advance;
            found := FindZero(t, c);
          end;
          if found and (c > a) and ((c - t) < step) and (c <= b) then begin
            FirstZero := TRUE;
            root := c;
          end
          else
            found := FALSE;
        end; {of ELSE}
      end; {of IF abs(dif) > 1.0e-13}
      inc(counter);
      prevF0 := f0;
    until found or (counter = stepNum);
  end;
end; {of FirstZero}


procedure EntersInCube(var value: double; ori, dir: Vector3; keyOri:
  CubeKey);
{Gives the value t < 1 such that ori + t*dir belongs to the surface of the cube}
var
  t, denom: double;
  i: Integer;
begin
  value := 0;
  for i := 1 to 3 do
    if keyOri[i] < 2 then begin
      t := (gMin3D[i] + keyOri[i] * gEdges3D[i] - ori[i]) / dir[i];
      if t > value then
        value := t;
    end;
  if gClips <> nil then
    for i := 0 to High(gClips) do begin
      denom := EuclInnerProd3(gClips[i].u, dir);
      if denom < 0 then begin
        t := -HeightOverClip(gClips[i], ori) / denom;
        if (t > value) then
          value := t;
      end
    end;
end; {of EntersInCube}

procedure SegmentInCube(ori, dir: Vector3; keyOri: CubeKey; var a, b:
  double);
const
  veryBig: double = 1.0E+30;
  yetBigger: double = 2.0E+30;
var
  ta, tb, denom: double;
  i: Integer;
begin
  a := -veryBig;
  b := veryBig;
  tb := yetBigger;
  for i := 1 to 3 do
    if keyOri[i] = 2 then begin
      if dir[i] > 0 then
        tb := (gMax3D[i] - ori[i]) / dir[i]
      else if dir[i] < 0 then
        tb := (gMin3D[i] - ori[i]) / dir[i];
      if tb < b then
        b := tb;
    end
    else begin
      ta := (gMin3D[i] + keyOri[i] * gEdges3D[i] - ori[i]) /
        dir[i];
      tb := (gMin3D[i] + (1 - keyOri[i]) * gEdges3D[i] - ori[i]) /
        dir[i];
      if ta > a then
        a := ta;
      if tb < b then
        b := tb;
    end;
  if gClips <> nil then
    for i := 0 to High(gClips) do begin
      denom := EuclInnerProd3(gClips[i].u, dir);
      if denom <> 0 then begin
        ta := -HeightOverClip(gClips[i], ori) / denom;
        if (denom > 0) then begin
          if (ta < b) then
            b := ta
        end
        else if (ta > a) then
          a := ta;
      end
    end;
end;

function IsZeroBetween(vp, dir: Vector3; a, b: double): Boolean;
const
  maxProbes = 20;
var
  values: array[1..maxProbes] of double;
  i, j, iOfSmallest, numOfProbes: Integer;
  pos, displacement, displacedPos: Vector3;
  step, smallestValue, val0, val1, minAdvance, c, delta, jump, prevJump,
    differ,
    param, ab
    : double;
begin
  result := FALSE;
  ab := b - a;
  c := ab * EuclNorm3(dir);
  if c > gRadius3D then begin
    numOfProbes := 2 + round(maxProbes * c / gSpan3D);
    step := ab / numOfProbes;
    for i := 1 to 3 do begin
      pos[i] := vp[i] + a * dir[i];
      displacement[i] := step * dir[i];
    end;
    values[1] := gTheF1.V(pos);
    smallestValue := abs(values[1]);
    iOfSmallest := 1;
    for i := 2 to numOfProbes do begin
      for j := 1 to 3 do
        pos[j] := pos[j] + displacement[j];
      values[i] := gTheF1.V(pos);
      if values[i] * values[i - 1] < 0 then begin
        result := TRUE;
        Exit;
      end;
      if abs(values[i]) < smallestValue then begin
        smallestValue := abs(values[i]);
        iOfSmallest := i;
      end;
    end; {de FOR i:=2 TO numOfProbes}
    param := a + (iOfSmallest - 1) * step;
    delta := gRadius3D * ab / (20 * c);
    minAdvance := 4 * delta;
    jump := abs(step);
    val0 := values[iOfSmallest];
    for j := 1 to 3 do
      pos[j] := vp[j] + param * dir[j];
    repeat
      prevJump := 0.9 * abs(jump);
      for j := 1 to 3 do
        displacedPos[j] := pos[j] + delta * dir[j];
      val1 := gTheF1.V(displacedPos);
      differ := val0 - val1;
      if abs(differ) < 1.0E-11 then
        Exit;
      jump := val0 * delta / differ;
      if abs(jump) > prevJump then
        Exit;
      param := param + jump;
      for j := 1 to 3 do
        pos[j] := vp[j] + param * dir[j];
      val0 := gTheF1.V(pos);
    until (abs(val0) < gVerySmall) or (prevJump < minAdvance);
    result := (param > a) and (param < 0.999 * b);
  end;
end; {of IsZeroBetween}


function IsVisibleImplicit(fromHere, point: Vector3; theKey: CubeKey;
  pointIsInCube:
  Boolean): Boolean;
const
  cut: double = 0.999;
var
  a, b: double;
  director: Vector3;
  i: Integer;
begin
  for i := 1 to 3 do
    director[i] := point[i] - fromHere[i];
  b := 1;
  if pointIsInCube then
    EntersInCube(a, fromHere, director, theKey)
  else
    SegmentInCube(fromHere, director, theKey, a, b);
  if b > 1 then b := 1;
  if (a > cut) or (b < a) then
    result := TRUE
  else
    result := not IsZeroBetween(fromHere, director, a, b);
end;

function IsVisibleImplicitS(fromHere: Vector3; point: Vector3Single; theKey:
  CubeKey;
  pointIsInCube:
  Boolean): Boolean;
const
  cut: double = 0.999;
var
  a, b: double;
  director: Vector3;
  i: Integer;
begin
  for i := 1 to 3 do
    director[i] := point[i] - fromHere[i];
  b := 1;
  if pointIsInCube then
    EntersInCube(a, fromHere, director, theKey)
  else
    SegmentInCube(fromHere, director, theKey, a, b);
  if b > 1 then b := 1;
  if (a > cut) or (b < a) then
    result := TRUE
  else
    result := not IsZeroBetween(fromHere, director, a, b);
end;

procedure ReceiveColorMap(var bm: TBitmap; aMap: ColorMapPtr);
var
  i, j: Integer;
  line: pTColorArray;
  thePal: Pal200Ptr;
begin
  if gType <> PseudoRiemannian then
    thePal := gPalette3D
  else
    thePal := gPalette2D;
  for j := 0 to kHeightMinus1 do begin
    line := bm.ScanLine[j];
    for i := 0 to kWidthMinus1 do
      line^[i] := thePal^[aMap^[i, j]];
  end;
end;

procedure InitSavedLines;
var
  i: Integer;
begin
  for i := kCoordinates to kArrows do begin
    gLines2D[i] := nil;
    SetLength(gLines2D[i], 50);
    gLines3D[i] := nil;
    SetLength(gLines3D[i], 50);
    gLinesData[i].numSegm2D := 0;
    gLinesData[i].numSegm3D := 0;
    gLinesData[i].noMoreRoom := False;
  end;
end;

procedure GetLinesRoom(dim: Integer);
var
  size: Integer;
begin
  try
    if dim = 2 then begin
      size := Length(gLines2D[gActiveLine]);
      if size > gLinesData[gActiveLine].maxSegm then begin
        beep;
        surface.Inform(eExcessCurves);
        gLinesData[gActiveLine].noMoreRoom := True;
      end
      else if size <= gLinesData[gActiveLine].numSegm2D then
        SetLength(gLines2D[gActiveLine], size + 50);
    end
    else begin
      size := Length(gLines3D[gActiveLine]);
      if size > gLinesData[gActiveLine].maxSegm then
        gLinesData[gActiveLine].noMoreRoom := True
      else if size <= gLinesData[gActiveLine].numSegm3D then
        SetLength(gLines3D[gActiveLine], size + 50);
    end;

  except on EOutOfMemory do begin
      surface.Inform(eExcessCurves);
      gLinesData[gActiveLine].noMoreRoom := True;
      if dim = 2 then
        gLinesData[gActiveLine].numSegm2D := Length(gLines2D)
      else
        gLinesData[gActiveLine].numSegm3D := Length(gLines3D);
    end;
  end
end;

procedure AddSegment(v1, v2: Vector2; dim, color: Integer);
var
  index: Integer;
begin
  if not gLinesData[gActiveLine].noMoreRoom then begin
    GetLinesRoom(dim);
    if dim = 2 then begin
      index := gLinesData[gActiveLine].numSegm2D;
      gLinesData[gActiveLine].numSegm2D := index + 1;
      gLines2D[gActiveLine, index, 1, 1] := v1[1];
      gLines2D[gActiveLine, index, 1, 2] := v1[2];
      gLines2D[gActiveLine, index, 2, 1] := v2[1];
      gLines2D[gActiveLine, index, 2, 2] := v2[2];
    end
    else begin
      index := gLinesData[gActiveLine].numSegm3D;
      gLinesData[gActiveLine].numSegm3D := index + 1;
      gLines3D[gActiveLine, index].seg[1, 1] := v1[1];
      gLines3D[gActiveLine, index].seg[1, 2] := v1[2];
      gLines3D[gActiveLine, index].seg[2, 1] := v2[1];
      gLines3D[gActiveLine, index].seg[2, 2] := v2[2];
      gLines3D[gActiveLine, index].color := color;
    end;
  end;
end;

procedure DrawSavedLines(dim: Integer);
var
  i, j: Integer;
  can: TCanvas;
  g: gamma16Ptr;
begin
  if odd(dim) then begin
    if dim = 1 then
      can := gForPrintingBmp.Canvas
    else
      can := surface.Drawing3D.canvas;
    can.Pen.Mode := pmCopy;
    can.Pen.Width := 1;
    if gType <> PseudoRiemannian then
      for i := kCoordinates to kArrows do
        if gLinesData[i].numSegm3D > 0 then begin
          g := gLinesData[i].gammaP;
          for j := 0 to gLinesData[i].numSegm3D - 1 do begin
            can.Pen.Color := g^[gLines3D[i, j].color];
            can.moveto(round(gLines3D[i, j].seg[1, 1]), round(gLines3D[i,
              j].seg[1, 2]));
            can.lineto(round(gLines3D[i, j].seg[2, 1]), round(gLines3D[i,
              j].seg[2, 2]));
          end;
        end;
  end
  else begin
    if dim = 2 then
      can := surface.Drawing2D.canvas
    else
      can := gForPrintingBmp.Canvas;
    can.Pen.Mode := pmCopy;
    can.Pen.Width := 1;
    if gType <> Implicit then
      for i := kCoordinates to kArrows do
        if gLinesData[i].numSegm2D > 0 then begin
          can.Pen.Color := gLinesData[i].pen2DColor;
          for j := 0 to gLinesData[i].numSegm2D - 1 do begin
            can.moveto(round(gLines2D[i, j, 1, 1]), round(gLines2D[i, j, 1,
              2]));
            can.lineto(round(gLines2D[i, j, 2, 1]), round(gLines2D[i, j, 2,
              2]));
          end;
        end;
  end

end;

procedure InitWire;
begin
  gWire := nil;
  gNumNodes3D := 0;
  gNoMemoryForWire := False;
end;

procedure GetWireRoom;
var
  size: Integer;
begin
  if gNoMemoryForWire then
    surface.Inform(eExcessWire)
  else try
    size := Length(gWire);
    if size <= gNumNodes3D then
      SetLength(gWire, gNumNodes3D + 50);
  except on EOutOfMemory do begin
      surface.Inform(eExcessWire);
      gNoMemoryForWire := True;
    end
  end;
end;

procedure AddSegmentToWire(currentX: Vector3; key: Integer);
begin
  if not gNoMemoryForWire then begin
    GetWireRoom;
    inc(gNumNodes3D);
    gWire[gNumNodes3D - 1].pos[1] := currentX[1];
    gWire[gNumNodes3D - 1].pos[2] := currentX[2];
    gWire[gNumNodes3D - 1].pos[3] := currentX[3];
    gWire[gNumNodes3D - 1].key := key;
  end;
end;



procedure DrawCube(can: TCanvas);
var
  i, x, y: Integer;
begin
  if gCube = nil then
    Exit;
  can.Pen.Color := gammaGrey[7];
  for i := 0 to gCubeLast do begin
    x := round(gCube[i, 1, 1]);
    y := round(gCube[i, 1, 2]);
    can.moveto(x, y);
    x := round(gCube[i, 2, 1]);
    y := round(gCube[i, 2, 2]);
    can.lineto(x, y);
  end;
end;

procedure EraseX;
var
  can: TCanvas;
begin
  if gThereIsX3D then begin
    can := surface.Drawing3D.canvas;
    can.Pen.Color := clBlack;
    can.Pen.Mode := pmNotXoR;
    can.moveto(gX3D[1].x, gX3D[1].y);
    can.lineto(gX3D[2].x, gX3D[2].y);
    can.moveto(gX3D[3].x, gX3D[3].y);
    can.lineto(gX3D[4].x, gX3D[4].y);
    can.pen.Mode := pmCopy;
    gThereIsX3D := False;
  end;
  if gThereIsX2D then begin
    can := surface.Drawing2D.canvas;
    can.Pen.Color := clTeal;
    can.Pen.Mode := pmNotXoR;
    can.moveto(gX2D[1].x, gX2D[1].y);
    can.lineto(gX2D[2].x, gX2D[2].y);
    can.moveto(gX2D[3].x, gX2D[3].y);
    can.lineto(gX2D[4].x, gX2D[4].y);
    can.pen.Mode := pmCopy;
    gThereIsX2D := False;
  end;
end;

procedure EraseUnitConic;
var
  can: TCanvas;
  i, j, k: Integer;
  nearly: Boolean;
begin
  if gThereIsUnitConic > 0 then begin
    can := surface.drawing2D.canvas;
    can.Pen.Color := gammaGrey^[2];
    can.Pen.Mode := pmNotXor;
    nearly := (abs(gUnitConic[1].x) < 15000) and (abs(gUnitConic[1].y) <
      15000);
    if nearly then
      can.moveto(gUnitConic[1].x, gUnitConic[1].y);
    if (gThereIsUnitConic = 1) then begin
      for i := 2 to 31 do
        if (abs(gUnitConic[i].x) < 5000) and (abs(gUnitConic[i].y) < 5000) then
        begin
          if nearly then begin
            can.lineto(gUnitConic[i].x, gUnitConic[i].y);
            can.lineto(gUnitConic[i].x, gUnitConic[i].y);
          end
          else begin
            can.moveto(gUnitConic[i].x, gUnitConic[i].y);
            nearly := true;
          end
        end;
    end
    else begin
      i := 0;
      for j := 1 to 4 do begin
        inc(i);
        nearly := (abs(gUnitConic[i].x) < 15000) and (abs(gUnitConic[i].y) <
          15000);
        can.MoveTo(gUnitConic[i].x, gUnitConic[i].y);
        for k := 1 to 20 do begin
          inc(i);
          if (abs(gUnitConic[i].x) < 5000) and (abs(gUnitConic[i].y) < 5000)
            then begin
            if nearly then begin
              can.lineto(gUnitConic[i].x, gUnitConic[i].y);
              can.lineto(gUnitConic[i].x, gUnitConic[i].y);
            end
            else begin
              can.moveto(gUnitConic[i].x, gUnitConic[i].y);
              nearly := true;
            end
          end;
        end;
      end;
      can.Pen.Mode := pmCopy;
    end;
    gThereIsUnitConic := 0;
  end;
end;

function ComputeOrthonormalBase(var e1, e2: Vector2; E, F, G, det: Double):
  Boolean;
const
  maxLejos = 5 * kWidth;
var
  discr, p, q, twoF: double;
  i: Integer;
  temp: vector2;
begin
  Result := FALSE;
  if abs(det) > 1.0E-20 then begin
    twoF := 2 * F;
    discr := hypot(E - G, twoF);
    if F = 0 then begin
      e1[1] := 1;
      e1[2] := 0;
      e2[1] := 0;
      e2[2] := 1;
    end
    else begin
      e1[1] := E - G + discr;
      e1[2] := twoF;
      e2[1] := E - G - discr;
      e2[2] := twoF;
    end;
    p := E * sqr(e1[1]) + twoF * e1[1] * e1[2] + G * sqr(e1[2]);
    q := E * sqr(e2[1]) + twoF * e2[1] * e2[2] + G * sqr(e2[2]);
    if p < 0 then begin
      temp := e1;
      e1 := e2;
      e2 := temp;
      discr := p;
      p := q;
      q := discr;
    end;
    p := 1 / sqrt(abs(p));
    q := 1 / sqrt(abs(q));
    for i := 1 to 2 do begin
      e1[i] := e1[i] * p;
      e2[i] := e2[i] * q;
      if gType = PseudoRiemannian then begin
        if abs(pd1 * e1[i]) > maxLejos then
          Exit;
        if abs(pd1 * e2[i]) > maxLejos then
          Exit;
      end
      else begin
        if abs(e1[i] / gEdges2D[i]) > 5 then
          Exit;
        if abs(e2[i] / gEdges2D[i]) > 5 then
          Exit;
      end
    end;
    Result := True;
  end
end;

procedure DrawUnitConic;
var
  i, j, k: Integer;
  alpha, beta, det, E, F, G: double;
  co, si: Extended;
  pos, e1, e2: vector2;
  nearly: Boolean;
  can: TCanvas;
begin
  EraseUnitConic;
  if gType = PseudoRiemannian then begin
    E := gTheF1.V(gCur2DPos);
    F := gTheF2.V(gCur2DPos);
    G := gTheF3.V(gCur2DPos);
  end
  else
    ParametricUtils.ComputeEFG(gCur2DPos, E, F, G);
  det := E * G - sqr(F);
  if not ComputeOrthonormalBase(e1, e2, E, F, G, det) then
    Exit;
  can := surface.drawing2D.canvas;
  can.Pen.Color := gammaGrey^[2];
  can.Pen.Mode := pmNotXor;
  if (det > 0) then begin
    alpha := 0;
    beta := pi / 15;
    pos[1] := gCur2DPos[1] + e1[1];
    pos[2] := gCur2DPos[2] + e1[2];
    FromParametersToDrawing(pos, gUnitConic[1]);
    nearly := (abs(gUnitConic[1].x) < 15000) and (abs(gUnitConic[1].y) <
      15000);
    if nearly then
      can.moveto(gUnitConic[1].x, gUnitConic[1].y);
    gUnitConic[0] := gCurPtDrawing2D;
    for i := 2 to 31 do begin
      alpha := alpha + beta;
      sincos(alpha, si, co);
      pos[1] := gCur2DPos[1] + e1[1] * co + e2[1] * si;
      pos[2] := gCur2DPos[2] + e1[2] * co + e2[2] * si;
      FromParametersToDrawing(pos, gUnitConic[i]);
      if (abs(gUnitConic[i].x) < 5000) and (abs(gUnitConic[i].y) < 5000) then
      begin
        if nearly then begin
          can.lineto(gUnitConic[i].x, gUnitConic[i].y);
          can.lineto(gUnitConic[i].x, gUnitConic[i].y);
        end
        else begin
          can.moveto(gUnitConic[i].x, gUnitConic[i].y);
          nearly := true;
        end
      end;
    end;
    gThereIsUnitConic := 1;
  end
  else begin
    i := 0;
    for j := 0 to 3 do begin
      inc(i);
      alpha := -1;
      beta := 1 / 10;
      if j < 2 then begin
        co := (1 - 2 * j) * cosh(alpha);
        si := sinh(alpha);
      end
      else begin
        si := (5 - 2 * j) * cosh(alpha);
        co := sinh(alpha);
      end;
      pos[1] := gCur2DPos[1] + e1[1] * co + e2[1] * si;
      pos[2] := gCur2DPos[2] + e1[2] * co + e2[2] * si;
      FromParametersToDrawing(pos, gUnitConic[i]);
      nearly := (abs(gUnitConic[i].x) < 5000) and (abs(gUnitConic[i].y) <
        5000);
      if nearly then
        can.moveto(gUnitConic[i].x, gUnitConic[i].y);
      for k := 1 to 20 do begin
        inc(i);
        alpha := alpha + beta;
        if j < 2 then begin
          co := (1 - 2 * j) * cosh(alpha);
          si := sinh(alpha);
        end
        else begin
          si := (5 - 2 * j) * cosh(alpha);
          co := sinh(alpha);
        end;
        pos[1] := gCur2DPos[1] + e1[1] * co + e2[1] * si;
        pos[2] := gCur2DPos[2] + e1[2] * co + e2[2] * si;
        FromParametersToDrawing(pos, gUnitConic[i]);
        if (abs(gUnitConic[i].x) < 5000) and (abs(gUnitConic[i].y) < 5000) then
        begin
          if nearly then begin
            can.lineto(gUnitConic[i].x, gUnitConic[i].y);
            can.lineto(gUnitConic[i].x, gUnitConic[i].y);
          end
          else begin
            can.moveto(gUnitConic[i].x, gUnitConic[i].y);
            nearly := true;
          end;
        end;
      end;
    end;
    gThereIsUnitConic := 2;
  end;
  can.Pen.Mode := pmCopy;
end;

procedure BasicDrawWire(can: TCanvas; color: TColor);
var
  i, j, k: Integer;
  place1, place2: vector3Single;
  pixel1, pixel2: TPoint;
  vertices: array[0..1, 0..1, 0..1] of TPoint;
begin
  can.Pen.Color := color;
  for i := 0 to 1 do
    for j := 0 to 1 do
      for k := 0 to 1 do begin
        place1[1] := gMin3D[1] + i * gEdges3D[1];
        place1[2] := gMin3D[2] + j * gEdges3D[2];
        place1[3] := gMin3D[3] + k * gEdges3D[3];
        vertices[i, j, k] := ProjectSToPaneInt(place1);
      end;
  for i := 0 to 1 do
    for j := 0 to 1 do begin
      can.moveto(vertices[i, j, 0].x, vertices[i, j, 0].y);
      can.lineto(vertices[i, j, 1].x, vertices[i, j, 1].y);
      can.moveto(vertices[i, 0, j].x, vertices[i, 0, j].y);
      can.lineto(vertices[i, 1, j].x, vertices[i, 1, j].y);
      can.moveto(vertices[0, i, j].x, vertices[0, i, j].y);
      can.lineto(vertices[1, i, j].x, vertices[1, i, j].y);
    end;
  i := 0;
  if gNumNodes3D < 1 then
    Exit;
  place2 := gWire[0].pos;
  pixel2 := ProjectSToPaneInt(place2);
  can.moveto(pixel2.x, pixel2.y);
  while i < gNumNodes3D - 1 do begin
    i := i + 1;
    while (gWire[i].key = 1) do begin
      place2 := gWire[i].pos;
      pixel2 := ProjectSToPaneInt(place2);
      can.moveto(pixel2.x, pixel2.y);
      i := i + 1;
    end;
    place1 := place2;
    pixel1 := pixel2;
    place2 := gWire[i].pos;
    pixel2 := ProjectSToPaneInt(place2);
    can.lineto(pixel2.x, pixel2.y)
  end;
end;



procedure DrawWire(can: TCanvas; nLeft, hLeft: Vector3);
var
  pcxPrev: Integer;
  c0Prev: double;
  prevH, prevN: Vector3;
begin
  gMovingBmp.canvas.Brush.Color := clBlack;
  if gMode = WireframeMono then begin
    gMovingBmp.canvas.Pen.Mode := PmCopy;
    gMovingBmp.canvas.FillRect(gRect);
    BasicDrawWire(gMovingBmp.canvas, gammaBlue[12]);
  end
  else if gMode = WireframeStereo then begin
    gMovingBmp.canvas.FillRect(gRect);
    BasicDrawWire(gMovingBmp.canvas, clGreen);
    prevH := gH;
    prevN := gN;
    gH := hLeft;
    gN := nLeft;
    gMovingBmp.canvas.Pen.Mode := PmMerge;
    BasicDrawWire(gMovingBmp.canvas, cEyeRed);
    gH := prevH;
    gN := prevN;
  end
  else if gMode = WireframeStereoPair then begin
    gMovingBmp.canvas.FillRect(gRect);
    pcxPrev := gpcx;
    c0Prev := gC0;
    gC0 := c0Prev / 2;
    gpcx := gpcx - kWidth div 4;
    gMovingBmp.canvas.Pen.Mode := pmCopy;
    BasicDrawWire(gMovingBmp.canvas, gammaBlue[12]);
    gpcx := pcxPrev + kWidth div 4;
    prevH := gH;
    prevN := gN;
    gH := hLeft;
    gN := nLeft;
    BasicDrawWire(gMovingBmp.canvas, gammaBlue[12]);
    gH := prevH;
    gN := prevN;
    gpcx := pcxPrev;
    gC0 := c0Prev;
  end;
  can.draw(0, 0, gMovingBmp);
end;

procedure PrintWire(can: TCanvas; n, h: Vector3);
var
  pcxPrev: Integer;
  c0Prev: double;
  prevH, prevN: Vector3;
begin
  can.Pen.Mode := pmCopy;
  can.Pen.Width := 2;
  if gMode = WireframeMono then
    BasicDrawWire(can, gammaBlue[12])
  else if gMode = WireframeStereo then begin
    BasicDrawWire(can, gammaBlue[15]);
    prevH := gH;
    prevN := gN;
    gH := h;
    gN := n;
    BasicDrawWire(can, gammaRed[12]);
    gH := prevH;
    gN := prevN;
  end
  else if gMode = WireframeStereoPair then begin
    pcxPrev := gpcx;
    c0Prev := gC0;
    gC0 := c0Prev / 2;
    gpcx := gpcx - (gResolution*kWidth) div 4;
    BasicDrawWire(can, gammaBlue[12]);
    gpcx := pcxPrev + (gResolution*kWidth) div 4;
    prevH := gH;
    prevN := gN;
    gH := h;
    gN := n;
    BasicDrawWire(can, gammaBlue[12]);
    gH := prevH;
    gN := prevN;
    gpcx := pcxPrev;
    gC0 := c0Prev;
  end;
end;

function DefinedUnitVector(place: TPoint; var theVector: Vector3):
  Boolean;
var
  u2, t, discr: double;
  i: Integer;
begin
  DefinedUnitVector := TRUE;
  u2 := sqr(place.x - gpcx) + sqr(place.y - gpcy) + c0sqr;
  discr := dSqr * c0sqr - u2 * (dSqr - virtualSphRad2);
  if discr < 0 then
    DefinedUnitVector := FALSE
  else begin
    discr := discr + virtualSphRad2;
    t := (d * c0abs - sqrt(discr)) / u2;
    for i := 1 to 3 do
      theVector[i] := t * ((place.x - gpcx) * gH[i] + (place.y - gpcy) *
        gV[i] + gc0 * gN[i]) + gViewpoint[i] - gCenter3D[i];
    discr := sqrt(EuclInnerProd3(theVector, theVector));
    for i := 1 to 3 do
      theVector[i] := theVector[i] / discr;
  end;
end;

procedure SaveParams;
begin
  svp := gViewpoint;
  sH := gH;
  sV := gV;
  sN := gN;
  sC0 := gC0;
  sC1 := gC1;
  sC2 := gC2;
  sC3 := gC3;
  spcx := gpcx;
  spcy := gpcy;
end;

procedure RetrieveParams;
begin
  gViewpoint := svp;
  gH := sH;
  gV := sV;
  gN := sN;
  gC0 := sC0;
  gC1 := sC1;
  gC2 := sC2;
  gC3 := sC3;
  gpcx := spcx;
  gpcy := spcy;
end;


procedure MoveSurface(prevPt,
  curPt: TPoint; Shift: TShiftState);
var
  nLeft, hLeft: Vector3;
  e1, e2, b: Vector3;
  norm, twoVe2, e2b, vb, scale: double;
  i, mode: Integer;
  good: Boolean;
begin
  dSqr := sqr(gViewpoint[1] - gCenter3D[1]) + sqr(gViewpoint[2] - gCenter3D[2])
    + sqr(gViewpoint[3] - gCenter3D[3]);
  d := sqrt(dSqr);
  virtualSphRad2 := 0.175 * sqr(gSpan3D);
  c0abs := abs(gc0);
  c0sqr := sqr(c0abs);
  coefRadius := sqrt(virtualSphRad2 / (dSqr - virtualSphRad2));
  good := DefinedUnitVector(prevPt, e1);

  if ssCtrl in Shift then
    mode := 1
  else if ssShift in Shift then
    mode := 2
  else
    mode := 0;

  case mode of
    1: begin
        gpcx := gpcx - prevPt.x + curPt.x;
        gpcy := gpcy - prevPt.y + curPt.y;
      end;
    2: begin
        scale := (1 + (prevPt.y - curPt.y) / 500);
        if ((c0abs > 1) or (scale > 1)) and ((c0abs < 10000) or (scale <
          1)) then
          gc0 := gc0 * scale;
        c0abs := abs(gc0);
        c0sqr := sqr(c0abs);
      end;
  else begin
      good := good and DefinedUnitVector(curPt, e2);
      if good then begin
        for i := 1 to 3 do
          b[i] := e1[i] + e2[i];
        norm := EuclNorm3(b);
        for i := 1 to 3 do
          b[i] := b[i] / norm;
        e2b := EuclInnerProd3(e2, b);
        twoVe2 := 2 * EuclInnerProd3(e2, gH);
        vb := EuclInnerProd3(b, gH);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          gH[i] := vb * b[i] - twoVe2 * e2[i] + gH[i];
        twoVe2 := 2 * EuclInnerProd3(e2, gV);
        vb := EuclInnerProd3(b, gV);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          gV[i] := vb * b[i] - twoVe2 * e2[i] + gV[i];
        twoVe2 := 2 * EuclInnerProd3(e2, gN);
        vb := EuclInnerProd3(b, gN);
        vb := 2 * (twoVe2 * e2b - vb);
        for i := 1 to 3 do
          gN[i] := vb * b[i] - twoVe2 * e2[i] + gN[i];
        gc3 := EuclInnerProd3(gCenter3D, gN) - d;
        gc1 := EuclInnerProd3(gCenter3D, gH);
        gc2 := EuclInnerProd3(gCenter3D, gV);
        for i := 1 to 3 do
          gViewpoint[i] := gc3 * gN[i] + gc1 * gH[i] + gc2 * gV[i];
      end; {if good}
    end; {else}
  end; {case}
  if gMode = Movable then begin
    gMovingBmp.canvas.FillRect(gRect);
    DrawSurfaceWithoutShadows;
    surface.drawing3D.canvas.Draw(0, 0, gMovingBmp);
  end
  else if gMode in gWireModes then begin
    for i := 1 to 3 do begin
      nLeft[i] := EyCos * gN[i] + EySin * gH[i];
      hLeft[i] := -EySin * gN[i] + EyCos * gH[i];
    end;
    DrawWire(surface.Drawing3D.canvas, nLeft, hLeft);
  end;
end;

procedure Draw3DLinesForPrinter(magnif: Double);
var
  i, j: Integer;
  savedPen: TPen;
  p1, p2: TPoint;
  g: gamma16Ptr;
  can: TCanvas;
begin
  can := gForPrintingBmp.canvas;
  savedPen := can.Pen;
  can.copyMode := cmsrcCopy;
  can.brush.color := clWhite;
  can.brush.style := bsSolid;
  can.Pen.Mode := pmCopy;
  if gResolution > 2 then
    can.Pen.Width := 2
  else
    can.Pen.Width := 1;
  if gType <> PseudoRiemannian then begin
    for i := kCoordinates to kArrows do
      if gLinesData[i].toPrint and (gLinesData[i].numSegm3D > 0) then begin
        g := gLinesData[i].gammaP;
        for j := 0 to gLinesData[i].numSegm3D - 1 do begin
          can.Pen.Color := g^[gLines3D[i, j].color];
          p1.x := round(magnif * gLines3D[i, j].seg[1, 1]);
          p1.y :=round(magnif * gLines3D[i, j].seg[1, 2]);
          p2.x := round(magnif * gLines3D[i, j].seg[2, 1]);
          p2.y := round(magnif * gLines3D[i, j].seg[2, 2]);
          can.moveto(p1.x, p1.y);
          can.lineto(p2.x, p2.y);
        end;
      end;
    if gShowCube and (gCube <> nil) then begin
      can.Pen.Color := gammaGrey[7];
      for i := 0 to gCubeLast do begin
        p1.x := round(magnif * gCube[i, 1, 1]);
        p1.y :=  round(magnif * gCube[i, 1, 2]);
        can.moveto(p1.x, p1.y);
        p2.x := round(magnif * gCube[i, 2, 1]);
        p2.y := round(magnif * gCube[i, 2, 2]);
        can.lineto(p2.x, p2.y);
      end;
    end;
  end;
  can.pen := savedPen;
end;

procedure Draw2DLinesForPrinter(magnif: double);
var
  i, j: Integer;
  savedPen: TPen;
  can: TCanvas;
begin
  can := gForPrintingBmp.canvas;
  savedPen := can.Pen;
  can.Pen.Mode := pmCopy;
  if gResolution > 2 then
    can.Pen.Width := 2
  else
    can.Pen.Width := 1;
  if gType <> Implicit then
    for i := kCoordinates to kArrows do
      if (gLinesData[i].numSegm2D > 0) and gLinesData[i].toPrint then begin
        can.Pen.Color := gLinesData[i].pen2DColor;
        for j := 0 to gLinesData[i].numSegm2D - 1 do begin
          can.moveto(round(magnif * gLines2D[i, j, 1, 1]), r.top + round(magnif * gLines2D[i, j, 1,
            2]));
          can.lineto(round(magnif * gLines2D[i, j, 2, 1]), r.top + round(magnif * gLines2D[i, j, 2,
            2]));
        end;
      end;
  can.Pen := savedPen;
end;

procedure BitmapToPrinterCanvas(destRect: TRect; source: TBitmap);
var
{$IFDEF MSWINDOWS}
  BitmapHeader: pBitmapInfo;
  BitmapImage: POINTER;
  HeaderSize: DWORD;
  ImageSize: DWORD;
  w, h: Integer;
{$ENDIF}
  can: TCanvas;
begin
  can := Printer.Canvas;
{$IFDEF MSWINDOWS}
  GetDIBSizes(source.Handle, HeaderSize, ImageSize);
  GetMem(BitmapHeader, HeaderSize);
  GetMem(BitmapImage, ImageSize);
  w := destRect.right - destRect.left;
  h := destRect.bottom - destRect.top;
  try
    GetDIB(source.Handle, source.Palette, BitmapHeader^, BitmapImage^);
    StretchDIBits(can.Handle,
      destRect.left, destRect.top, // Destination Origin
      w, h, // Destination Width and Height
      0, 0, // Source Origin
      source.Width, source.Height, // source Width and Height
      BitmapImage,
      TBitmapInfo(BitmapHeader^),
      DIB_RGB_COLORS,
      SRCCOPY);
  finally
    FreeMem(BitmapHeader);
    FreeMem(BitmapImage)
  end;
{$ENDIF}
{$IFDEF LINUX}
  can.StretchDraw(destRect,source);
{$ENDIF}
end;

procedure DrawLongFormula(theText: AnsiString; var y: Integer; lineHeight: Integer);
var
  head, theCurrentHead, theBareText, theLine: AnsiString;
  theChar: char;
  isFirstLine: Boolean;
  index, textLength, lineEnd, previousLineEnd, d, d1, d2, d3, b1, b2, b3,
    headWidth, lineWidth, x: Integer;
  can: TCAnvas;
begin
  can := Printer.Canvas;
  textLength := Length(theText);
  index := pos('=', theText);
  head := Copy(theText, 1, index + 1);
  if Length(head) > 0 then
    headWidth := can.TextWidth(head)
  else
    headWidth := 0;
  theBareText := '';
  if textLength > index + 1 then begin
    for index := index + 1 to textLength do begin
      theChar := theText[index];
      if not (theChar in [kReturn, kTab, kLineFeed]) then
        theBareText := theBareText + theChar;
    end;
    textLength := Length(theBareText);
  end;
  if textLength = 0 then
  begin
    x := (Printer.PageWidth - headWidth) div 2;
    can.TextOut(x, y, head);
    y := y + lineHeight;
    Exit;
  end;
  isFirstLine := True;
  previousLineEnd := 0;
  d := 0;
  b1 := 0;
  b2 := 0;
  b3 := 0;
  theCurrentHead := head;
  index := 0;
  while index < textLength do
  begin
    index := index + 1;
    theChar := theBareText[index];
    case theChar of
      '+', '-':
        if d = 0 then
        begin
          b1 := index - 1;
          d1 := d;
        end
        else
        begin
          b2 := index - 1;
          d2 := d;
        end;
      ',', '.', ';', ':':
        if d = 0 then
        begin
          b1 := index;
          d1 := d;
        end
        else
        begin
          b2 := index;
          d2 := d;
        end;
      '(', '[', '{':
        d := d + 1;
      ')', ']', '}':
        begin
          d := d - 1;
          if d = 0 then
          begin
            b2 := index;
            d2 := d;
          end;
        end;
      '*', '/':
        if d = 0 then
        begin
          b2 := index;
          d2 := d;
        end;
      ' ':
        if d = 0 then
        begin
          b3 := index - 1;
          d3 := d;
        end
        else
          ;
    end; { of CASE}
    lineWidth := r.left + headWidth + can.TextWidth(Copy(theBareText,
      previousLineEnd + 1, index - previousLineEnd));
    if lineWidth > (r.Right - r.left) then
    begin
      if b1 > previousLineEnd + 1 then
      begin
        index := b1;
        d := d1;
      end
      else if b2 > previousLineEnd + 1 then
      begin
        index := b2;
        d := d2;
      end
      else if b3 > previousLineEnd + 1 then
      begin
        index := b3;
        d := d3;
      end
      else
        index := index - 1;
      LineEnd := index;
      theLine := Concat(theCurrentHead, Copy(theBareText, previousLineEnd +
        1, LineEnd - PreviousLineEnd));
      theCurrentHead := '';
      if isFirstLine then
      begin
        isFirstLine := False;
        x := r.left;
      end
      else
        x := r.left + headWidth;
      can.TextOut(x, y, theLine);
      y := y + lineHeight;
      b1 := 0;
      b2 := 0;
      b3 := 0;
      while (index < textLength) and (theBareText[index + 1] = ' ') do
        index := index + 1;
      previousLineEnd := index;
    end {of  IF lineWidth > theWidth}
    else if (index = textLength) then
    begin
      theLine := Concat(theCurrentHead, Copy(theBareText, previousLineEnd +
        1, textLength - PreviousLineEnd));
      if isFirstLine then
        x := r.left
      else
        x := r.left + headWidth;
      can.TextOut(x, y, theLine);
      y := y + lineHeight;
    end;
  end; {del While}
  y := y - lineHeight;
end; {of  DrawLongFormula}

procedure PreparegForPrintingBmp;
  var
    rBmp : TRect;
begin
  gForPrintingBmp.Height := gResolution * kHeight;
  gForPrintingBmp.Width := gResolution * kWidth;
  gForPrintingBmp.PixelFormat := pf32bit;
  gForPrintingBmp.Canvas.Brush.Color := clWhite;
  gForPrintingBmp.Canvas.Brush.Style := bsSolid;
  rBmp:= Rect(0, 0, gForPrintingBmp.Width, gForPrintingBmp.Height);
  gForPrintingBmp.Canvas.FillRect(rBmp);
end;

procedure GetPrinterParameters;
var
  w: Integer;
begin
  r.left := Printer.PageWidth div 10;
  r.top := Printer.PageHeight div 10;
  r.right := Printer.PageWidth - r.left;
  w := r.right - r.left;
  r.bottom := r.top + MulDiv(w, kHeight, kWidth);
  rOr := r;
  InflateRect(rOr, 8, 8);
  SaveParams;
  if (gMode in gWireModes) or not (gPrintSurface) then
    gResolution := 4;
  gc0 := sc0 * gResolution;
  gpcx := r.left + round(gpcx * gResolution);
  gpcy := r.top + round(gpcy * gResolution);
end;

procedure PrintText(t: AnsiString; var y: Integer; th: Integer; fnt: TFontName; sty: TFontStyles);
var
  textW, x: Integer;
begin
  with Printer.canvas do begin
    Font.Name := fnt;
    Font.Height := MulDiv(Printer.PageHeight, th, 1000);
    Font.Style := sty;
    textW := TextWidth(t);
    if (textW < r.Right - r.left) then begin
      x := (Printer.PageWidth - textW) div 2;
      TextOut(x, y, t);
    end
    else
      DrawLongFormula(t, y, Font.Height);
  end;
end;

procedure PrintAsDesired;
var
  num, y, shy, i, j: Integer;
  theText: AnsiString;
  nLeft, hLeft: Vector3;
  can : TCanvas;
begin
  num := Printer.copies;
  Printer.copies := 1;
  if num < 1 then
    Exit;
  Printer.orientation := poPortrait;
  GetPrinterParameters;
  shy := MulDiv(Printer.PageHeight, 20, 1000);
  can := gForPrintingBmp.Canvas;
  if (gMode in gWireModes) or not gPrintSurface then
    PreparegForPrintingBmp;
  if (gMode in gWireModes) then begin
    for j := 1 to 3 do begin
      nLeft[j] := EyCos * gN[j] + EySin * gH[j];
      hLeft[j] := -EySin * gN[j] + EyCos * gH[j];
    end;
    PrintWire(can, nLeft, hLeft);
  end
  else begin
    if (gResolution = 1) then begin
      PreparegForPrintingBmp;
      if gPrintSurface  then begin
        if gMode = ColoredStereoPair then
          can.Draw(0, 0, gMovingBmp)
        else
          can.Draw(0, 0, gStdBmp);
      end;
    end;
    if gPrintLines then begin
      if (gType = PseudoRiemannian) or ((not gPrintSurface) and (gMode = Standard)) then
        Draw2DLinesForPrinter(gResolution)
      else if (gMode = Standard) then
        Draw3DLinesForPrinter(gResolution);
    end;
  end;
  try
    Printer.BeginDoc;
    for i := 1 to num do begin
      y := r.Top - 3 * shy;
      Printer.canvas.FillRect(rOr);
      BitmapToPrinterCanvas(r, gForPrintingBmp);
      PrintText(surface.caption, y, 20, 'Arial', [fsBold]);
      if not (gMode in gWireModes) and (gMode <> ColoredStereoPair) then begin
          {$IFDEF MSWINDOWS}
          Printer.Canvas.Pen.Width := 1;
          {$ENDIF}
          {$IFDEF LINUX}
          Printer.Canvas.Pen.Width := 0;
          {$ENDIF}
          Printer.canvas.Pen.Color := clBlack;
          Printer.canvas.moveto(rOr.Left, rOr.Top);
          Printer.canvas.lineto(rOr.Left, rOr.Bottom);
          Printer.canvas.lineto(rOr.Right, rOr.Bottom);
          Printer.canvas.lineto(rOr.Right, rOr.Top);
          Printer.canvas.lineto(rOr.Left, rOr.Top);
      end;

      Printer.Canvas.Font.Style := [];
      y := r.Bottom + 2 * shy;
      case gType of
        Implicit: theText := '0 = ' + gSurfaceStr[1];
        Parametric: theText := 'x(u,v) = ' + gSurfaceStr[1];
        Tube: theText := 'x(t) = ' + gSurfaceStr[1];
        PseudoRiemannian: theText := 'E(x,y) = ' + gSurfaceStr[1];
      else ;
      end;
      PrintText(theText, y, 16, 'Arial', []);
      y := y + (shy * 3) div 2;
      case gType of
        Parametric: theText := 'y(u,v) = ' + gSurfaceStr[2];
        Tube: theText := 'y(t) = ' + gSurfaceStr[2];
        PseudoRiemannian: theText := 'F(x,y) = ' + gSurfaceStr[2];
      else ;
      end;
      if (gType <> Implicit) then begin
        PrintText(theText, y, 16, 'Arial', []);
        y := y + (shy * 3) div 2
      end;
      case gType of
        Parametric: theText := 'z(u,v) = ' + gSurfaceStr[3];
        Tube: theText := 'z(t) = ' + gSurfaceStr[3];
        PseudoRiemannian: theText := 'G(x,y) = ' + gSurfaceStr[3];
      else ;
      end;
      if (gType <> Implicit) then begin
        PrintText(theText, y, 16, 'Arial', []);
        y := y + (shy * 3) div 2
      end;
      if gType = Tube then begin
        theText := 'r(t) = ' + gSurfaceStr[4];
        PrintText(theText, y, 16, 'Arial', []);
        y := y + (shy * 3) div 2
      end;

      case gType of
        Implicit, PseudoRiemannian: theText := gSurfaceStr[5] + ' < x < ' + gSurfaceStr[6];
        Parametric: theText := gSurfaceStr[5] + ' < u < ' + gSurfaceStr[6];
        Tube: theText := gSurfaceStr[5] + ' < t < ' + gSurfaceStr[6];
      else ;
      end;
      PrintText(theText, y, 16, 'Arial', []);
      y := y + (shy * 3) div 2;
      case gType of
        Implicit, PseudoRiemannian: theText := gSurfaceStr[7] + ' < y < ' + gSurfaceStr[8];
        Parametric: theText := gSurfaceStr[7] + ' < v < ' + gSurfaceStr[8];
      else ;
      end;
      if gType <> Tube then begin
        PrintText(theText, y, 16, 'Arial', []);
        y := y + (shy * 3) div 2
      end;
      case gType of
        Implicit: theText := gSurfaceStr[9] + ' < z < ' + gSurfaceStr[10];
        Parametric: theText := eGridDensity + ' = ' + gSurfaceStr[10];
      else ;
      end;
      if (gType in [Implicit, Parametric]) then begin
        PrintText(theText, y, 16, 'Arial', []);
        y := y + (shy * 3) div 2
      end;
      if num > 1 then Printer.NewPage;
    end;
  finally
    Printer.EndDoc;
  end;
  RetrieveParams;
end;



function ExtracTitle(pathAndName: AnsiString): AnsiString;
var
  ext: AnsiString;
  len: Integer;
begin
  result := ExtractFilename(pathAndName);
  len := Length(result);
  if len > 4 then begin
    ext := lowerCase(Copy(result, len - 3, 4));
    if ext = '.gtv' then
      result := Copy(result, 1, len - 4);
  end;
end;

end.
